package com.jthomas.mailfaces;

/*
 * Face.java
 * Copyright (c) 1996 John Thomas  jthomas@cruzio.com
 *      All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for non-commercial purposes
 * is hereby granted provided that this copyright notice
 * appears in all copies.
 *
 */
  
import java.awt.*;
import java.io.*;
import java.util.*;
import java.net.*;

import com.jthomas.util.*;
/**
 * Class to handle the display of a "face" that represents one or more e-mail messages.
 * 
 * This class maintains a hashtable of Mailmsgs that are related to this face and 
 * a reference to a FaceButton object that contains the Picon.  The actual display of 
 * the Face is handled by FaceBoard.
 * <p>
 *
 * @see FaceButton FaceBoard Mailmsg 
 *
 * @author	John Thomas
 * <a href="mailto:jthomas@cruzio.com">jthomas@cruzio.com</a>
 * @version	1.0 ( Sept 1996 )
 *

 */

public 
class Face  {

// state variables

private	String		_URLstr;

private Mailmsg 	_Mailmsg;

private FaceButton  _FaceButton;            // Face Button for this Face

private FaceBoard  _FaceBoard;				// Face board that we belong to

private Hashtable 	_Mailmsgs;              // list of MailMsg objects represented

private int			_Number;				// Number of mailmsgs
    
 
    
/**
 * Face Constructor
 * 
 * <pre>
 *   Saves the passed reference to the container
 *   Creates a Hashtable to store related Mailmsgs
 * @param FaceBoard faceboard - Containing Frame object
 */

Face(FaceBoard faceboard) {
	
	//Debug.out("Face()");
	_Mailmsgs = new Hashtable(10);
	_URLstr = null;
	_FaceBoard = faceboard;  // save reference to container

}

//-------------------------------------------
/**
 * reset() is called when the containing FaceBoard is reset.
 * <p>
 * This will reflect the reset action to the associated
 * FaceButton.
 *
 * @return void
 */
//--------------------------------------------------------

public void
reset() {
	if ( _FaceButton != null)
		_FaceButton.reset();
}

//------------------------------------------
/**
 * update(mailmsg,imageurl) will update a new 
 * Face object with a mailmsg and a URL.
 *
 *@param Mailmsg mailmsg - Associated Mailmsg object
 *@param String imageurl - URL pointing to Face Icon
 *@return void
 */
//----------------------------------------------

public synchronized void
update(Mailmsg mailmsg, String imageurl) {
	//Debug.out("Face.update: " + mailmsg.getKey());
	_Mailmsgs.put(mailmsg.getKey(),mailmsg);
	_Number++;
	_URLstr = imageurl;
}

//------------------------------------------
/**
 * update(mailmsg,imageurl) will update an existing 
 * Face object with a mailmsg.
 *
 *@param Mailmsg mailmsg - Associated Mailmsg object
 *@return void
 */
//---------------------------------------------

public synchronized void
update(Mailmsg mailmsg) {
	if (mailmsg.getAction() == Mailmsg.ActionRemove)  { // remove a mailmsg
		_Number--;
		_Mailmsgs.remove(mailmsg.getKey());
		
	} else {
		_Mailmsgs.put(mailmsg.getKey(),mailmsg);
		_Number++;
	}
	if (_Number != 0 )
		if (_FaceButton == null) // should not happen
			Debug.out("Face.update() Internal error");
		else
   			_FaceButton.setCount(_Number);	// update the button with number of mailmsgs
   		
}




/**
** access the count of mailmsgs that are represented by this face.
*/
public int
getNumber() {
	return _Number;
}


/**
** save a reference to the FaceButton that contains the Picon 
*/
public void
setButton(FaceButton button) {

	_FaceButton = button;
}


/**
** Get a reference to the FaceButton that contains the Picon.
*/
public FaceButton
getButton()				{
	return _FaceButton;
}
	
//-------------------------------------------
/**
 * getMailmsgs() will return an array of Mailmsgs 
 * associated with this Face.
 *
 *@return Mailmsg[] - An array of Mailmsg objects.
 */
//---------------------------------------------
public synchronized Mailmsg[]
getMailmsgs()  {
	int n = _Mailmsgs.size();
	Mailmsg[] mailmsgs = new Mailmsg[ n ];
	Enumeration e = _Mailmsgs.elements();
	for (int i=0; e.hasMoreElements(); i++ ) {
		mailmsgs[i] = (Mailmsg)e.nextElement();
	}
	//Debug.out("getMailmsgs: " + n );
	return mailmsgs;
}

  
}  // end of class Face