package com.jthomas.mailfaces;

/*
 * FaceBoard.java
 * Copyright (c) 1996 John Thomas  jthomas@cruzio.com
 *      All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for non-commercial purposes
 * is hereby granted provided that this copyright notice
 * appears in all copies.
 *
 */

import java.awt.*;
import java.awt.event.*;
import java.net.*;
import java.util.*;
import java.io.*;
import com.jthomas.util.*;

/**
 *
 * Class to handle the display of a bunch of faces representing Mail senders.
 * 
 * This class sets up a panel that will contain
 * a number of Face objects.  There is one Face object for each unique
 * domain/user. For each Face object there is a FaceButton.
 * Each FaceButton is an ImageButton can contains the icon for the 
 * associated mailmsg objects and when clicked
 * will display the list of those mailmsgs.
 * <p>
 
 * @see MailFaces  
 * @see Face  
 * @see FaceButton  
 * @see FaceFrame 
 * @see FaceScrollablePanel 
 * @author	John Thomas
 * <a href="mailto:jthomas@cruzio.com>jthomas@cruzio.com</a>
 * @version	2.0 ( May 1998 )
 *
 */

public class FaceBoard extends Panel implements Runnable {

  // Class variables
  static final int ICONSIZE = 52; // 48x48 plus 2 pixel border	
  static final int VGAP = 4; // Vertical gap between components
  static final int HGAP = 4; // Horz gap between components

  // state variables

  private FaceScrollablePanel _ScrollPane;
  private Panel _FacePanel; // Panel that will hold Picon buttons
  //private Panel 		_MainPanel;				// Panel that holds scrollable panel and bar.
  private FaceStatusLabel _Status; // Label that will hold status info
  private MenuBar _MBar;

  private MailFaces _MailFaces;
  private FaceOptions _Options;
  private Queue _Queue;

  private Hashtable _Faces; // Hashtable that will hold Face objects

  private boolean _NoMailIcon; // true if NoMailIcon present
  private int _Number = 0; // Number of Face objects
  private boolean _FaceUpdate; // true if number of Face images is updated
  private int _ShowMax = 10; // Max to be displayed
  private int _ShowMin = 4; // Min to be displayed
  private int _Width; // Width of Main window
  private int _WidthOld; // Width of Main window before entry to paint()
  private int _HeightOld; // Height of Main window before entry to paint()
  private int _Height; // Height of Main window
  private Point _Location;
  private FaceProperties _Properties;

  //-------------------------------
  /**
  ** FaceBoard Constructor
  ** It only initializes a few references to other objects.
  ** The real initialization happens in the setup() method
  */
  //-------------------------------
  FaceBoard() {

    // get a reference to the current Options object and Properties Object
    _Options = FaceOptions.getRef();

    _MailFaces = MailFaces.getRef();
    _Queue = new Queue(100);
    _Properties = FaceProperties.getRef();

    // Process the Options
    refreshOptions();
    // ask to be informed if the Options get changed
    _MailFaces.addFaceEventListener(new FaceEventListener() {
      public void actionPerformed(FaceEvent e) {
        if (e.getID() == FaceEvent.ID_OPTIONS_CHANGED)
          FaceBoard.this.refreshOptions();
      }
    });

  }
  //-----------------------------------------------------
  /**
   *Method: refreshOptions()
   *@return void
   *
   *   Invoked at startup and when user refreshes the Option file
   */
  //-----------------------------------------------------

  public void refreshOptions() {
    Debug.out("FaceBoard:refreshOptions()");
    _ShowMax = _Options.getInt("ShowMaxPIcons", _ShowMax);
    _ShowMin = _Options.getInt("ShowMinPIcons", _ShowMin);
    if (_ScrollPane != null) { // Only on updates
      _ScrollPane.setMax(_ShowMax);
      _FaceUpdate = true;
      Graphics g = getGraphics();
      this.paint(g);
    }

  }
  //------------------------------------------------------
  /**
  ** setup
  **    Called by applet or application to setup the
  **    frame that will contain the face images that
  **    represent the mail in the users inbox
  */
  //------------------------------------------------------
  public void setup() {

    this.setLayout(new BorderLayout(0, 0));

    //Debug.out(this.toString());
    //Debug.out(this.getInsets().toString());
    // Get width from Properties file
    _ShowMax = Integer.parseInt(_Properties.getProperty("FaceFrameShowMax", "" + _ShowMax));

    _Faces = new Hashtable(20);

    _FacePanel = new Panel(); // Panel that will contain the faces

    // set the layout that we need
    _FacePanel.setLayout(new FlowLayout(FlowLayout.LEFT, 0, 0));
    _FacePanel.setSize(ICONSIZE * _ShowMin, ICONSIZE);
    //_FacePanel.setBackground(Color.yellow);  //<<<<<<<<<<<<<<<<<<<<<<

    _ScrollPane = new FaceScrollablePanel(_FacePanel, ICONSIZE, ICONSIZE, _ShowMin, _ShowMax);

    //_ScrollPane.setBackground(Color.pink);

    this.add("Center", _ScrollPane);

    Dimension d = _ScrollPane.getPreferredSize();
    _ScrollPane.setSize(d);
    _Width = d.width;
    _Height = d.height;

    _HeightOld = _Height;
    _WidthOld = this.getSize().width; // Used by paint()
    _Location = this.getLocation();
    //Debug.out(this.getSize());

    Graphics g = getGraphics();
    this.paint(g);
  }

  //-----------------------------------------------------------------
  /**
  ** paint:
  **    This method is called when the system needs to repaint the window.
  **    including when it has been resized and when we call it ourselves
  **    It will recognize when the scrollbar needs to be added or deleted
  **    because the user resized the screen or we have added more icons.
  */
  //-----------------------------------------------------------------   
  public synchronized void paint(Graphics g) {

    // Check if window has been resized
    Dimension d = _ScrollPane.getSize();
    int width = d.width;
    int height = d.height;

    Debug.out("FaceBoard.paint(): " + d);

    // return if no change in width, height or faces
    if ((width == 0 || width == _WidthOld) && (height == 0 || height == _HeightOld) && !_FaceUpdate) {
      // No change in width, height  number of icons

      super.paint(g); // Handle the components

      return;
    }

    // If resized horizontally, adjust the _ShowMax value for setting scrollbars
    if (width != _WidthOld && _Number > 0) {
      Debug.out("w=" + width + " _W=" + _WidthOld + " " + _Number);
      Insets in = this.getInsets();
      int mainwidth = width - in.left + in.right;
      if (width < _WidthOld)
        _ShowMax = Math.max(mainwidth / ICONSIZE, _ShowMin);
      else {
        int n = mainwidth / ICONSIZE;
        if (n > _Number)
          _ShowMax = Math.max(_Number, _ShowMin);
        else
          _ShowMax = Math.max(n, _ShowMin);
      }
      _ScrollPane.setMax(_ShowMax);
      _FaceUpdate = true;
      _Properties.put("FaceFrameShowMax", "" + _ShowMax);
      _Properties.save();
    }

    Debug.out("_Number=" + _Number + " _ShowMax=" + _ShowMax + " _ShowMin=" + _ShowMin);
    if (_FaceUpdate) {
      _FaceUpdate = false;

      // This is real funny.  No matter what I do the 
      // scroll position is one unit less than what it is supposed to be
      // It appears to be a known bug (4008152)
      // I also have code in FaceScrollablePanel that tries to set this
      // By having similar code in both places, it seems to bypass the 
      // problem  (brute force and persistance always pays off.
      int n = _Number - _ShowMax;
      if (n > 0) {
        Adjustable adjust = _ScrollPane.getHAdjustable();
        n = n * adjust.getUnitIncrement();
        _ScrollPane.setScrollPosition(n, 0);
      }
      // Now cause the window layout to reflect any new status
      this.invalidate();
      // Pack the outer Frame if it has been added to a frame 
      Container c = this;
      while (c != null && !(c instanceof Frame))
        c = c.getParent();
      if (c != null)
         ((Frame) c).pack();

    }
    d = _ScrollPane.getSize();
    _WidthOld = d.width;
    _HeightOld = d.height;

    super.paint(g); // Handle components 
  }

  //----------------------------------------------------------------------------
  /**
  ** reset()
  ** 	Reset the panel so that no FaceButtons are present
  */
  //----------------------------------------------------------------------------
  private void reset() {

    Debug.out("FaceBoard.reset()");
    // Issue reset to all the Faces
    if (_Faces != null) {
      Enumeration e = _Faces.elements();
      while (e.hasMoreElements())
         ((Face) e.nextElement()).reset();
    }
    _FacePanel.removeAll();
    _FacePanel.setVisible(false);
    _Faces = new Hashtable(20); // new Hashtable

    _Number = 0;
    //Graphics g = getGraphics(); 	
    _FaceUpdate = true;
    //this.paint(g);	
  }

  //-------------------------------------------------------------------------
  /**
   * run() watches the queue of GUI related work to do and parcels it out.
   *
   *   This will read either Mailmsgs from the queue or special commands
   *   If queued object is a Mailmsg, then it is sent to the update method.
   *   If the queued object is a RESET command, then the reset method is called.
   *   If the queued object is a NOMAIL command, then the nomail method is called.
   * @return void
   */
  //-------------------------------------------------------------------------
  public void run() {

    Mailmsg mailmsg;

    while (true) {
      Object o = _Queue.get();
      if (o instanceof Mailmsg) {
        mailmsg = (Mailmsg) o;
        Debug.out("From Queue: Mailmsg " + mailmsg.getKey() + " " + mailmsg.getImageURL());
        update(mailmsg);
      } else {
        if (o instanceof String) {
          String s = (String) o;
          Debug.out("From Queue: String " + s);
          if (s.equals("NOMAIL")) {
            reset();
            nomail();
          } else if (s.equals("RESET"))
            reset();
        }
      }

    }
  }

  //---------------------------------------------------------------------
  /**
   * queuePut will place the specified object into the FaceBoard Work Queue
   *
   */
  //---------------------------------------------------------------------- 
  public void queuePut(Object obj) {
    _Queue.put(obj);
  }

  //----------------------------------------------------------------------------
  /**
  ** 
  **	Update the FaceBoard so that a new mailmsg is represented
  **  It will be represented either by adding a new FaceButton or by adding
  **  to the count for an existing FaceButton
  */
  //----------------------------------------------------------------------------
  public void update(Mailmsg mailmsg) {
    String imageurl;
    String buildFrom = null;
    Face face;
    FaceButton button = null;
    // Remove MNoMailIcon if real mail has arrived.
    if (_NoMailIcon & (mailmsg.getAction() != Mailmsg.ActionDummy)) {
      reset();
      _NoMailIcon = false;
    }
    if (mailmsg.getAction() == Mailmsg.ActionRemove) { // remove face
      face = mailmsg.getFace();
      if (face != null) {
        face.update(mailmsg);
        if (face.getNumber() == 0) { // no mailmsgs left for icon
          button = face.getButton();
          _FacePanel.remove(button);
          _Faces.remove(mailmsg.getImageURL());
          _Number--;
          _ScrollPane.setNumber(_Number);
          // Handle a special case where it is not completely erasing the 
          // last icon. There has got to be a better way.  ????
          if ((_Number - _ShowMin) < 0) {
            Debug.out("Clear Panel - Special handling");
            _FaceUpdate = true;
            _ScrollPane.setVisible(false);
            _ScrollPane.invalidate();
            _ScrollPane.doLayout();
            Graphics g = getGraphics();
            this.paint(g);
            _ScrollPane.setVisible(true);

          }
        }
      }
    } else {

      imageurl = mailmsg.getImageURL();
      if (Debug.ON) Debug.out(0,"FaceBoard.update: imageURL="+imageurl);
      if (imageurl == null) { // possible if autodelete msg
        Debug.out("FaceBoard.update: No image to display");
        return;
      }
      if (mailmsg.isBuildIcon()) {
        buildFrom = mailmsg.getHeader("reply-to");
        if (buildFrom.length() == 0) 
          buildFrom = mailmsg.getHeader("from");
        face = (Face) _Faces.get(buildFrom);
      } else {
        face = (Face) _Faces.get(imageurl); // Check Hashtable for Face with this URL
      }
      if (face != null) { // We already have a Face for this guy
        face.update(mailmsg); // Run update for Face object        
      } else { // This is a new image URL
        _Number++;
        face = new Face(this); // Construct new Face object
        face.update(mailmsg, imageurl); // Run update for Face object 
        button = this.displayImage(mailmsg, imageurl); // create a FaceButton    	
        if (button == null) { // image could not be loaded
          imageurl = _Options.get("DefaultImageURL"); // Get default image

          button = this.displayImage(mailmsg, imageurl); // create a FaceButton
          if (button == null) { // If failed again
            Debug.out(0,"FaceBoard.update: Unable to create a Face Button from image " + imageurl);
            _Number--;
            return; // Just return with no record of face
          }
        }
        if (mailmsg.isBuildIcon())
          _Faces.put(buildFrom, face);
        else
          _Faces.put(imageurl, face); // Store Face object in hashtable  
        face.setButton(button); // save reference to Button in Face 
        button.setFace(face); // save reference to Face in button 
        _ScrollPane.setNumber(_Number);
      }
      mailmsg.setFace(face);
    }

    _FaceUpdate = true;
    Graphics g = getGraphics();
    _FacePanel.paintComponents(g);

    this.paint(g);
  }

 

  //-------------------------------------------------
  /** 
  ** Obtain a list of all Mailmsg objects for all Face 
  ** objects.
  */
  //-------------------------------------------------
  public Mailmsg[] getAllMailmsgs() {

    Mailmsg[] answer = null;
    Vector temp = new Vector(50);

    Enumeration enum = _Faces.elements();
    while (enum.hasMoreElements()) {
      Face face = (Face) enum.nextElement();
      Mailmsg[] msgs = face.getMailmsgs();
      for (int i = 0; i < msgs.length; i++)
        temp.addElement(msgs[i]);
    }
    enum = temp.elements();
    answer = new Mailmsg[temp.size()];
    for (int i = 0; enum.hasMoreElements(); i++)
      answer[i] = (Mailmsg) enum.nextElement();
    return answer;
  }

  //-------------------------------------------------
  /** 
  ** Handle the NoMail case
  */
  //-------------------------------------------------
  private void nomail() {
    String imageurl;

    _Number = 1;
    imageurl = _Options.get("NoMailURL");

    this.displayImage(null, imageurl);
    _NoMailIcon = true;
    _ScrollPane.setNumber(_Number);
    Graphics g = getGraphics();
    _FacePanel.paintComponents(g);
    _FaceUpdate = true;

    this.paint(g);
  }

  //----------------------------------------------------------------------------
  /**
  ** displayImage will
  ** create a FaceButton(subclass of ImageButton) 
  ** and cause it to be displayed
  */
  //----------------------------------------------------------------------------
  private FaceButton displayImage(Mailmsg mailmsg, String imageurl) {

    Image image = null;
    FaceButton ibutton;
    // fixup relative URLs
    imageurl = checkURL(imageurl);
    if (imageurl == null) // checkURL handled this err
      return null; // No Face button will be built

    Toolkit t = Toolkit.getDefaultToolkit();
    try {
      image = t.getImage(new URL(imageurl));
    } catch (MalformedURLException ex) {
      Debug.out("FaceBoard.displayImage: Invalid URL for Image\n" + imageurl);
      MsgDialog.create("Invalid URL for Image\n" + imageurl);
      return null;
    }
    MediaTracker tracker = new MediaTracker(this);
    tracker.addImage(image, 0);
    try {
      tracker.waitForAll();
    } catch (InterruptedException e) {
      Debug.out("MediaTracker Failed to get image " + imageurl);
      MsgDialog.create("MediaTracker failed to get image:\n  " + imageurl);
      return null;
    }
    if (tracker.isErrorAny()) {
      image.flush();
      Debug.out("Error loading image for " + imageurl);
      MsgDialog.create("Error loading image:\n  " + imageurl);
      return null;
    }

    ibutton = new FaceButton(image); // Create the ImageButton
    ibutton.setBorderWidth(2); // So 48x48 image will need 52x52 size
    ibutton.setSize(ICONSIZE, ICONSIZE);
    if (mailmsg != null && mailmsg.isBuildIcon()) {
      
      String buildFrom = mailmsg.getHeader("reply-to");
      if (buildFrom.length() == 0) 
        buildFrom = mailmsg.getHeader("from");
      String[] overlay = parseEmailAddr(buildFrom);
      ibutton.setText(overlay);
    }
    _FacePanel.add(ibutton); // Add to the FaceBoard

    // resize the Face Panel so that the image can be seen
    _FacePanel.setSize(new Dimension(ICONSIZE * _Number, ICONSIZE));

    _FacePanel.setVisible(true); // Make sure that the user can see it.
    
    Graphics g = getGraphics();    
    this.paint(g);
    
    return ibutton;
  }

  //----------------------------------------------------------------------------
  /**
  ** parseEmailAddr will take the input string and parse it as 
  ** an RFC822 Email address and return an array of 3 values.
  ** <br>1 - Userid
  ** <br>2 - Host name qualifier n-1
  ** <br>3 - Host name qualifier n
  **
  ** so "John Q Public<jqpublic@somename.somewhere.com> would return
  **   -> {"jqpublic","somewhere","com"}
  **
  */
  //----------------------------------------------------------------------------
  private String[] parseEmailAddr(String from) {
    String[] answer = { "Unknown", "?", "?" };

    String[] temp = Mailmsg.parseNetworkAddress(from);
    // above returns  "userid"  "host"  "name"
    if (temp[0].length() > 0)
      answer[0] = temp[0];
    StringTokenizer st = new StringTokenizer(temp[1], ".");
    int n = st.countTokens();
    for (int i = 0; st.hasMoreElements(); i++) {
      String str = st.nextToken();
      if (i == n - 2)
        answer[1] = str;
      else if (i == n - 1)
        answer[2] = str;
    }

    return answer;
  }

  //----------------------------------------------------------------------------
  /**
  ** checkURL
  **	Verify that a local "file:" URL exists.
  **
  **  It also will convert a relative file URL 
  **  like "file:./ddd/ssss.gif" which is invalid into a 
  **  valid URL with the full pathname.
  */
  //----------------------------------------------------------------------------

  private static String checkURL(String imageURL) {
    String newURL = null;
    String filename = null;

    if (imageURL.startsWith("file:")) {
      filename = imageURL.substring(5);
      // check for a file: url with a relative name. 	
      if (filename.startsWith("./"))
        filename = filename.substring(2); // Remove the "file:./" from front
      else {
        if (filename.startsWith("//"))
          filename = filename.substring(2);
        //filename = filename.substring(filename.indexOf('/'));	
      }

      File file = new File(filename);
      if (file.exists())
        newURL = "file:" + file.getAbsolutePath().replace(File.separatorChar, '/');
      else {
        MsgDialog.create("Selected local image " + imageURL + " does not exist.");
        Debug.out(0, "url=" + imageURL + " filename=" + filename);
        return null;
      }
      return newURL;
    } else {
      return imageURL;
    }
  }

  /**
  ** Calculates the FaceBoard dimensions
  */

  public Dimension getMinimumSize() {

    return _ScrollPane.getMinimumSize();

  }

  /**
  ** Calculates the FaceBoard dimensions
  */

  public Dimension getPreferredSize() {
    //Debug.out("PreferredSize");
    return getMinimumSize();
  }

} //------- End of class FaceBoard