package com.jthomas.mailfaces;

/*
 * FaceButton.java
 *
 *
 */

import java.io.*;
import java.util.*;
import java.awt.*;
import com.jthomas.util.*;

/**
 * FaceButton is a subclass of ImageButton that adds the ability to 
 * add a number that indicates the number of mailmsgs represented by 
 * this FaceButton.
 *
 * <pre>
 *
 * SubClass of ImageButton(
 *            	Marty Hall,
 *				hall@apl.jhu.edu,
 *				href="http://www.apl.jhu.edu/~hall/java"
 * This adds the ability to override the image with a number that represents the
 * number of mailmsgs represented by this FaceButton
 *
 * It also adds the action method to do something when the user presses the button
 *
 * Note that the ImageButton uses the old event model so FaceButton 
 * also needs to use the old event model.  At some point we will move
 * this to the JFC ImageButton (Swing 1.0) 
 * <p>
 *
 * </pre>
 *
 * @author	John Thomas
 * <a href="mailto:jthomas@cruzio.com">jthomas@cruzio.com</a>
 * @version	1.0 ( Sept 1996 )
 
*/

public class FaceButton extends ImageButton {

	// state variables

	private int _Count = 0;
	private Face _Face; // Reference to Face object
	private boolean _Paint1 = false;
	// represented by this button
	private MailList _MailList;

	private String[] _OverlayText;

	/**
	** if true then display a "grayed-out button"
	*/
	private boolean _GrayOut = false;
	//------------------------------------------------------------------    
	/**
	** FaceButton Constructor
	**	 Create an ImageButton using the image specified.
	*/
	//------------------------------------------------------------------

	public FaceButton(Image image) {

		super(image);
		_Count = 1;
	}

	//-------------------------------------------
	/**
	** reset()
	**   This is called when the containing FaceBoard is reset
	*/
	//-------------------------------------------
	public void reset() {
		if (_MailList != null) {
			_MailList.dispose();
			_MailList = null;
		}
	}

	//------------------------------------------------------------------    
	/**
	** setCount
	**	 Set the Count of mailmsgs represented  (used by paint()
	*/
	//------------------------------------------------------------------

	public void setCount(int count) {

		_Count = count;
		_GrayOut = false;
		Graphics g = getGraphics();
		if (g != null) // this seems to get a null value sometimes
			paint(g);
	}

	//------------------------------------------------------------------    
	/**
	** setText
	**	 Set the Text that we will overlay onto the Button
	*/
	//------------------------------------------------------------------

	public void setText(String[] overlay) {

		_OverlayText = overlay;
		//Graphics g = getGraphics();
		//paint(g);
	}

	//------------------------------------------------------------------    
	/**
	** setFace
	**	 Set a reference to the Face Object represented by this ImageButton
	**   The Face object is needed by the action() method
	*/
	//------------------------------------------------------------------

	public void setFace(Face face) {
		_Face = face;
	}

	//-----------------------------------------------------------------
	/**
	** paint:
	**   Draws the number of messages into the left bottom corner
	**   of the imageButton. It first places a white square in this
	**   area to make the number easy to read
	**   _Paint1 will handle when we paint 2 and then need to repaint with 1.
	*/
	//-----------------------------------------------------------------   
	public void paint(Graphics g) {
		//Debug.out("FaceButton: paint()");
		if (g == null)
			return;
		super.paint(g); // Forces main image to get loaded
		if (_GrayOut) {
			if (explicitSize)
				g.drawImage(
					grayImage,
					margin,
					margin,
					width - 2 * margin,
					height - 2 * margin,
					this);
			else
				g.drawImage(grayImage, margin, margin, this);
			drawBorder(g, false);
		}

		if (_Count > 1 || _Paint1) {
			g.setColor(Color.white);
			g.fillRect(2, 32, 14, 14);
			Font f = new Font("Monospaced", Font.BOLD, 14);
			g.setFont(f);
			g.setColor(Color.black);
			g.drawString(Integer.toString(_Count), 4, 44);
			_Paint1 = true;
		}
		if (_OverlayText != null) {
			Font f = new Font("Monospaced", Font.PLAIN, 12);
			g.setFont(f);
			g.setColor(Color.black);
			g.drawString(_OverlayText[0], 4, 12);
			g.drawString(_OverlayText[1], 4, 24);
			g.drawString(_OverlayText[2], 4, 36);

		}
	}

	/**
	 * Method drawBorder.
	 * @param g     Graphis object
	 * @param isUp  True if button is "up"
	 */
	//-----------------------------------------------------------------

	private void drawBorder(Graphics g, boolean isUp) {

		g.setColor(marginColor);
		int left = 0;
		int top = 0;
		int width = this.width;
		int height = this.height;
		for (int i = 0; i < margin; i++) {
			g.draw3DRect(left, top, width, height, isUp);
			left++;
			top++;
			width = width - 2;
			height = height - 2;
		}
	}

	//-----------------------------------------------------------------
	/**
	** action:
	**   It will take control when the user clicks on a Face button.
	**   It creates a MailList object and passes the Face Object to it
	**   MailList will display the Mailmsgs that are assocaited with this Face object
	**
	**   If this is the "NoMail" button there will not be a reference to the Face object.
	*/
	//-----------------------------------------------------------------   
	public boolean action(Event event, Object arg) {
		if (Debug.ON)
			Debug.out("FaceButton.action: " + event.toString());
		if (event.target instanceof FaceButton) {
			if (_Face != null) {
				if (_MailList != null)
					_MailList.dispose();
				_MailList = new MailList(_Face);
				_GrayOut = true;
				Graphics g = getGraphics();
				if (g != null)
					paint(g);

			} else if (Debug.ON)
				Debug.out(
					"FaceButton.action: " + "Ignore select of NoMail Button");
			return true;
		} else
			return false;

	}

	//----------------------------------
	/**
	**Method: mouseEnter
	**
	** Mouse entered this FaceButon - Repaint it
	*/
	//----------------------------------      
	public boolean mouseEnter(Event evt, int x, int y) {

		//Debug.out("FaceButton - Mouse Enter");
		Graphics g = getGraphics();
		if (g != null) {
			paint(g);
		} else {
			Debug.out("Mouse Enter but no Graphic object");
		}
		return true;
	}

} // End of FaceButton Class