package com.jthomas.mailfaces;
/*
 * FaceCache.java
 * Copyright (c) 1996 John Thomas  jthomas@cruzio.com
 *      All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for non-commercial purposes
 * is hereby granted provided that this copyright notice
 * appears in all copies.
 *
 */

import java.io.*;
import java.util.*;
import java.net.*;

import com.jthomas.util.*;
/**
 * Class to Manage the Cache of Face Images used by MailFaces. 
 *   
 *
 * <p>
 * The cache saved on disk consists of a single index file and some
 * number of icon files (gif images).  The index file consists of a list
 * of FaceCachEntry entries in String format.  
 * 
 * At startup, the index file is read and a number of hashtables are built 
 * with the data portion being the FaceCacheEntry
 *    _ByUser  - key is Userid@host
 *    _ByHost  - key is host  
 *    _ByURL   - key is URL
 * There is also a Vecter _Entries that contains all of the FaceCachEntry objects.    
 *
 * When MailFaces is stopped normally, the current entries are written
 * back to the index file.  
 * 
 * @see FaceCacheEntry
 * @author	John Thomas
 * <a href="mailto:jthomas@cruzio.com>jthomas@cruzio.com</a>
 * @version	2.0 ( May 1998 )
 */

public class FaceCache implements FaceEventListener {
	/**
	** Maximum number of images in Cache.  This is not a
	** Hard limit since Hashtables and Vectors can grow
	*/
	final static int MAX_ENTRIES = 200;

	// Public state variables

	/**
	** true if Cache is being used.
	*/
	public boolean _Active = false; // True if Cache being used

	// Private state variables

	/**
	** Index - key is user@host
	*/
	private Hashtable _ByUser;
	/**
	** Index - key is host
	*/
	private Hashtable _ByHost;
	/**
	** Index - key is url (original url)
	*/
	private Hashtable _ByURL;
	/**
	** Collection of all entries
	*/
	private Vector _Entries;
	/**
	** Cache directory
	*/
	private File _DirFile;
	/**
	** Cache Index file
	*/
	private File _IndexFile;
	/**
	** Path for Cache directory
	*/
	private String _Path;
	/**
	** URL Path for Cach directory 
	*/
	private String _PathURL;
	/**
	** Number used to generate filename
	*/
	private int _FileNumber;

	/**
	**True if Cache gets updated
	*/
	private boolean _Updated = false;

	/**
	**Count of entries in cache*/
	private int _Count;

	/**
	** List of PiconDB servers, need to get images from PiconDB
	*/
	private PiconDBServer[] _PiconDBList;

	/**
	** Reference to the FaceOptions object
	*/
	private FaceOptions _Options;

  /**
  ** Reference to the central object
  */
  private MailFaces _MailFaces;
	//-----------------------------------------
	/**
	** Constructor
	**
	** Uses the specified directory to setup the Cache.
	**
	**@param directory where the cache info is stored.
	**@throws IOException if problems writing cache entries
	*/
	//-----------------------------------------

	public FaceCache(String directory) throws IOException {
		// directory is the directory where the Cache will be maintained
		// Verify that the directory exist and can be written
		try {
      _MailFaces = MailFaces.getRef();
			
			// Setup Cache Directory and if already exists, read the Index file. 
			_DirFile = new File(directory);
			if (!_DirFile.exists()) {
				Debug.out("FaceCache: "+directory+" Directory does not exist");
				_DirFile.mkdir();
			}
			if (!_DirFile.canWrite()) {
				Debug.out(0,"FaceCache: "+directory + " Directory is not writable");
				throw new IOException();
			}
			_Path = _DirFile.getAbsolutePath();
			if (!_Path.endsWith(new Character(File.separatorChar).toString()))
				_Path += File.separatorChar;
			_PathURL = "file:" + _Path.replace(File.separatorChar, '/');
			Debug.out(_Path + " ?= " + _PathURL);
			// Setup memory indexes and prime with existing Index file
			this.readIndex();
			// ask to be informed when user enters Cache commands
			if (_MailFaces != null) {
        _MailFaces.addFaceEventListener(this);
        // get a reference to the current Options object 
        _Options = FaceOptions.getRef();
      }

		} catch (Exception e) {
			Debug.out(e);
			Debug.out("Failed to create Cache Object.");
			MsgDialog.create("Failed creating Cache.");
			_Active = false;
		}
	}

	//--------------------------------------------------------------------
	/** 
	** actionPerformed is the required implementation of 
	** FaceEventListener.  
	** It is given control when some other components has sent
	** an event that the Cache object needs to know about
	** 
	*/
	//--------------------------------------------------------------------
	public void actionPerformed(FaceEvent e) {
		switch (e.getID()) {
			case FaceEvent.ID_CACHE_EXPIRE :
				expire();
				break;
			case FaceEvent.ID_CACHE_CLEAR :
				clearIndex();
				break;
			case FaceEvent.ID_CACHE_VERIFY :
				verify(0);
				break;
			case FaceEvent.ID_CACHE_LIST :
				listIndex();
				break;

			default :
				break;
		}

	}

	//--------------------------------------------------------------------
	/** 
	** readIndex will read the Index file and rebuild the
	** Hashtables that keep track of cached images.
	*/
	//--------------------------------------------------------------------
	private synchronized void readIndex() {
		int count = 0;
		// Create the in-memory Indexes 
		_ByUser = new Hashtable(MAX_ENTRIES);
		_ByHost = new Hashtable(MAX_ENTRIES);
		_ByURL = new Hashtable(MAX_ENTRIES);
		_Entries = new Vector(MAX_ENTRIES);
		try {
			_IndexFile = new File(_Path, "Index");
			_FileNumber = 0;
			if (_IndexFile.exists()) {
				BufferedReader datafile;
				String entrystr;
				FaceCacheEntry entry = null;
				datafile = new BufferedReader(new FileReader(_IndexFile));
				while ((entrystr = datafile.readLine()) != null) {
					//Debug.out("->" + entrystr);
					entry = new FaceCacheEntry(entrystr);
					if (entry._CacheURL != null)
						_FileNumber =
							Math.max(_FileNumber, getFileNumber(entry));
					addIndex(entry);
					count++;
				}
				Debug.out("ReadIndex: "+ count+ " entries from Cache index "
						+ _IndexFile+ " Next:"+ _FileNumber);
				_Active = true;
				datafile.close();
				_Count = count;
			} else {
				Debug.out("Cache Index file "	+ _IndexFile
						+ " not found.  Start from scratch");
				_Active = true;
				_Count = 0;
			}
			_Updated = false;
		} catch (IOException e) {
			Debug.out(0,"FaceCache.ReadIndex: "+
				"Failed reading the Index file - " + _IndexFile + " " + e);
			_Active = false;
		}

	}

	//--------------------------------------------------------------------
	/**
	** addIndex(FaceCacheEntry entry)
	**    This adds an entry to the Cache
	**    It updates the in-memory indexes and adds it to the list of entries
	*/
	//--------------------------------------------------------------------
	public synchronized void addIndex(FaceCacheEntry entry) {
		String key;

		_Entries.addElement(entry);

		if (entry._Host != null) {
			if (entry._User != null) {
				key = entry._User + "@" + entry._Host;
				_ByUser.put(key, entry);
			} else
				_ByHost.put(entry._Host, entry);
		}
		if (entry._URL != null) {
			_ByURL.put(entry._URL, entry);
		}
		_Count++;
		_Updated = true;
	}

	//--------------------------------------------------------------------
	/**
	** deleteIndex(FaceCacheEntry entry)
	**    This deletes an entry to the Cache by
	**    deleting it from the in-memory indexes and
	**    and setting a flag so it will not be rewritten
	**    when the index is rewritten
	*/
	//--------------------------------------------------------------------
	public synchronized void deleteIndex(FaceCacheEntry entry) {
		String key;

		entry.setExpired(true);
		if (entry._Host != null) {
			if (entry._User != null) {
				key = entry._User + "@" + entry._Host;
				_ByUser.remove(key);
			} else
				_ByHost.remove(entry._Host);
		}
		if (entry._URL != null) {
			_ByURL.remove(entry._URL);
		}
		_Count--;
		_Updated = true;
	}

	//--------------------------------------------------------------------
	/**
	** writeIndex()
	**    Writes the Cache Index file 
	*/
	//--------------------------------------------------------------------
	public synchronized void writeIndex() {

		if (!_Updated) // Don't bother if we never updated it.
			return;

		DataOutputStream datafile;
		int count = 0;

		try {
			datafile = new DataOutputStream(new FileOutputStream(_IndexFile));

			Enumeration e = _Entries.elements();
			while (e.hasMoreElements()) {
				FaceCacheEntry entry = (FaceCacheEntry) e.nextElement();
				if (entry.getExpired()) {
					continue;
				}
				//Debug.out("<-" + entry.toString());
				datafile.writeBytes(entry.dump());
				count++;
			}
			datafile.close();
			Debug.out(
				"Wrote " + count + " entries to Cache index " + _IndexFile);
		} catch (IOException e) {
			Debug.out("Failed writing the Index file - " + _IndexFile, e);
		}
	}

	//--------------------------------------------------------------------
	/**
	** listIndex()
	**    Lists the Cache Index file 
	*/
	//--------------------------------------------------------------------
	public void listIndex() {

		int count = 0;

		Enumeration e = _Entries.elements();
		while (e.hasMoreElements()) {
			FaceCacheEntry entry = (FaceCacheEntry) e.nextElement();
			//Debug.out("<-" + entry.toString());
			Debug.out(Debug.ALWAYS, "** Cache: " + entry.list());
			count++;
		}
		Debug.out(Debug.ALWAYS, "** Cache: Index has " + count + " entries");

	}
	//--------------------------------------------------------------------
	/**
	** clearIndex()
	**    Clears the Cache Index file 
	*/
	//--------------------------------------------------------------------
	public void clearIndex() {

		int count = 0;

		Enumeration e = _Entries.elements();
		while (e.hasMoreElements()) {
			FaceCacheEntry entry = (FaceCacheEntry) e.nextElement();
			Debug.out("<- Deleting " + entry.toString());
			deleteIndex(entry);
			deleteCacheImage(entry);
			count++;
		}

		Debug.out("Deleted " + count + " entries");

		if (_IndexFile.delete())
			Debug.out("Deleted " + _IndexFile);

		else
			Debug.out("Deletion of " + _IndexFile + " failed.");

		readIndex();

	}

	//--------------------------------------------------------------------
	/**
	** getHost(String host)  returns -> String CacheURL
	**    Return the URL of the cached image if present.
	**    Will return null if not found.
	*/
	//--------------------------------------------------------------------
	public synchronized String getHost(String host) {

		FaceCacheEntry entry = (FaceCacheEntry) _ByHost.get(host);
		if (entry == null)
			return null;
		else {
			entry.reference();
			if (entry._CacheURL == null)
				return ""; // see comments for getUser
			else
				return entry._CacheURL;
		}
	}

	//--------------------------------------------------------------------
	/**
	** getUser(String user,String host)  returns -> String CacheURL
	**    Return the URL of the cached image if present.
	**    Will return null if not found.
	** <p>
	**  There is a special case here.
	**  We can have a cache entry for user@host that has a null url that
	**  represents that we searched the PIconDB and did not find an entry.
	**  So if the url in the entry is null than we return "" instead of null.
	**  to indicate that there is no reason to search the PiconDB again.
	*/
	//--------------------------------------------------------------------
	public synchronized String getUser(String user, String host) {
		String key = user + "@" + host;
		FaceCacheEntry entry = (FaceCacheEntry) _ByUser.get(key);
		if (entry == null)
			return null;
		else {
			entry.reference();
			if (entry._CacheURL == null)
				return ""; // see comments above
			else
				return entry._CacheURL;
		}
	}

	//--------------------------------------------------------------------
	/**
	** getURL(String url)  returns -> String CacheURL
	**    Return the URL of the cached image if present.
	**    Will return null if not found.
	*/
	//--------------------------------------------------------------------
	public synchronized String getURL(String url) {

		FaceCacheEntry entry = (FaceCacheEntry) _ByURL.get(url);
		if (entry == null)
			return null;
		else {
			entry.reference();
			if (entry._CacheURL == null)
				return "";
			else
				return entry._CacheURL;
		}
	}

	//-------------------------------------------------------------------
	/**
	** getCount
	**    Get the current count of index entries
	*/
	//-------------------------------------------------------------------
	public int getCount() {
		return _Count;
	}

	//--------------------------------------------------------------------
	/**
	** CacheImage
	**   Gives a URL address an image and the associated User/host (if not null)
	**   This will create a local copy of the image
	**   It will then create an index entry for this cached image
	**   and return a URL pointing to the cached image.
	*/
	//----------------------------------------------------------------------
	synchronized String cacheImage(
		String urlstr,
		String user,
		String host) {
		
    
    String pathurl;

		if (urlstr == null
			|| urlstr.equals("")) { // cache of the fact there is no image
			FaceCacheEntry entry = new FaceCacheEntry(null, host, user, null);
			addIndex(entry);
			return "";
		} else {
			String type = ".gif";
			if (urlstr.endsWith(".xbm"))
				type = ".xbm";
			else if (!urlstr.endsWith(".gif")) {
				Debug.out("Image type is not .gif or .xbm. ");
				return urlstr;
			}
			String filename = getTempFilename(type);

			String path = _Path + filename;
			if (!copyURL2File(urlstr, path)) {
				Debug.out("Failed to cache the image");
				return urlstr;
			} else {
				pathurl = _PathURL + filename;
				FaceCacheEntry entry =
					new FaceCacheEntry(urlstr, host, user, pathurl);
				addIndex(entry);
				return pathurl;
			}
		}
	}

	//--------------------------------------------------------------------
	/**
	** indexImage
	**   Gives a URL address of an image and the associated User/host (if not null)
	**   This method is used when the image is already cached but for a different
	**   user  (Because it is for a domain)
	**   It just creates a new entry for this user and node
	**                                                  
	*/
	//----------------------------------------------------------------------
	void indexImage(
		String urlstr,
		String user,
		String host,
		String pathurl) {

		FaceCacheEntry entry = new FaceCacheEntry(urlstr, host, user, pathurl);
		addIndex(entry);
	}

	//--------------------------------------------------------------------
	/**
	** expire
	**   Runs thru the index and expires out of date entries
	**                                                  
	*/
	//----------------------------------------------------------------------
	synchronized void expire() {

		int expire = 0;

		Date unusedDate;

		long day = 1000 * 60 * 60 * 24; // a day in milliseconds 

		int maxUnusedAge;
		try {
			maxUnusedAge = Integer.parseInt(_Options.get("CacheExpire", "0"));
		} catch (Exception e) {
			MsgDialog.create("CacheExpire not specified or was invalid.");
			maxUnusedAge = 0;
		}
		if (maxUnusedAge > 0) {
			unusedDate = new Date(new Date().getTime() - maxUnusedAge * day);
			Debug.out("Expire based on Last Used date > " + unusedDate);
			Enumeration e = _Entries.elements();
			while (e.hasMoreElements()) {
				FaceCacheEntry entry = (FaceCacheEntry) e.nextElement();
				if (entry.checkExpire(unusedDate)) {
					// Check if should be expired.
					Debug.out("***Expire this entry***");
					Debug.out(entry.list());
					deleteIndex(entry);
					deleteCacheImage(entry);
					expire++;
				}
			}

			Debug.out("Expired " + expire + " entries.");
			if (expire > 0) {
				_Updated = true;
				writeIndex();
				readIndex();
			}
		}
	}

	//--------------------------------------------------------------------
	/**
	** verify
	**   Runs thru the index and verifies that the entries are correct
	**   This is currently unused (no menuitem to call it)                                               
	*/
	//----------------------------------------------------------------------
	public synchronized void verify(int verifyAge) {

		int reset = 0;
		Date verifyDate;

		long day = 1000 * 60 * 60 * 24; // a day in milliseconds 

		verifyDate = new Date(new Date().getTime() - verifyAge * day);
		Debug.out(verifyDate.toString());

		Enumeration e = _Entries.elements();
		int i = 0;
		while (e.hasMoreElements()) {
			FaceCacheEntry entry = (FaceCacheEntry) e.nextElement();
			i++;
			// checkVerify will return true if verifyDate is > entry.Verified
//			if (entry.checkVerify(verifyDate)) {
//				Debug.out(i + " - " + entry.toString());
//
//				if (entry._CacheURL == null) {
//					// This entry represents a "no hit entry"
//					Debug.out(i + " - " + entry);
//					//entry.setExpired(true);        // For now we will just expire it.
//				} else {
//					String pathurl = entry._CacheURL;
//					int index1 = pathurl.lastIndexOf('/') + 1;
//					String filename = pathurl.substring(index1);
//					filename = _Path + filename;
//					// We should really rename the old file first 
//					if (!copyURL2File(entry._URL, filename)) {
//						Debug.out(
//							i
//								+ " - Caching failed for "
//								+ entry._URL
//								+ " to "
//								+ filename);
//						//entry.setExpired(true);	// if failure then expire entry
//					}
//				}
//				reset++;
//			}
		}

		Debug.out(0,"Verified " + i + " entries");
		if (reset > 0) {
			_Updated = true; // force rewrite
			writeIndex();
			readIndex();
		}
	}

	//---------------------------------------------------------------
	/**
	** getTempFileName
	**    create a filename for the cached image
	**    Uses a counter to generate a filename like 1324.gif
	*/
	//---------------------------------------------------------------
	private synchronized String getTempFilename(String type) {
		_FileNumber++;

		return Integer.toString(_FileNumber) + type;
	}

	//---------------------------------------------------------------
	/**
	** getFileNumber
	**    called when reading the index to get the filenumber of the
	**    most recent entry so we can start from there.
	*/
	//-------------------------------------------------------------
	private int getFileNumber(FaceCacheEntry entry) {
		String pathurl = entry._CacheURL;
		try {

			int index1 = pathurl.lastIndexOf('/') + 1;
			int index2 = pathurl.lastIndexOf('.');
			String filename = pathurl.substring(index1, index2);
			Debug.out("entry=" + entry + " FileNumber=" + filename);
			return Integer.parseInt(filename);
		} catch (Exception e) {
			Debug.out("Could not parse " + pathurl);
			return 0;
		}
	}

	//---------------------------------------------------------------
	/*
	** deleteCacheImage
	**    Delete the cache image if it is no longer used
	*/
	//-------------------------------------------------------------
	private boolean deleteCacheImage(FaceCacheEntry entry) {
		String pathurl = entry._CacheURL;
		if (pathurl == null)
			return false;

		boolean found = false;
		Enumeration e = _Entries.elements();
		while (e.hasMoreElements()) {
			FaceCacheEntry thisentry = (FaceCacheEntry) e.nextElement();
			//Debug.out("<-" + thisentry.toString());
			if (thisentry._CacheURL != null
				&& !thisentry.getExpired()
				&& thisentry._CacheURL.equals(pathurl)) {
				found = true;
				break;
			}
		}
		if (!found) {
			try {
				int index1 = pathurl.lastIndexOf('/') + 1;
				String filename = pathurl.substring(index1);

				File file = new File(_Path, filename);
				if (file.delete())
					Debug.out("Deleted file " + filename);
				else
					Debug.out("Delete failed for file " + filename);
				return true;
			} catch (Exception ex) {
				Debug.out("Could not delete " + pathurl);
				return false;
			}
		}
		return false;
	}

	//----------------------------------------------------------
	/**
	**Method: copyURL2File(urlstr,filename)
	**
	**  Take a GIF image pointed to by a URL and copy it into
	**  the Cache directory.
	**@param url for the image
	**@param filename for output
	**@return true if Copy is sucessfull
	*/
	//----------------------------------------------------------

	private synchronized boolean copyURL2File(String urlstr, String filename) {
		
    String data;
    int size = 0;
		FileOutputStream file;
		InputStream istream = getInputStream(urlstr);
		Debug.out("copyURL2File: " + urlstr + " -> " + filename);
		if (istream == null) {
			Debug.out(0,"copyURL2File: Unable to get data for " + urlstr);
			return false;
		}

		byte[] buffer = new byte[1024];
		try {
			boolean header = true;
			int numbytes = 0;
			file = new FileOutputStream(filename);
			while (true) {
				numbytes = istream.read(buffer);
				if (numbytes == -1)
					break;
        size += numbytes;            
				if (header) {
					data = new String(buffer, 0, 4);
					//Debug.out(data);
					if (data.equals("HTTP")) {
						handleHTTPStream(buffer, numbytes, istream, file);
						break;
					} else
						header = false;
				}
				// Now we have some data to write
				//Debug.out("Imagedata=" + data.substring(0,10)); 	
				file.write(buffer, 0, numbytes);
			}
			file.close();
      
			istream.close();
		} catch (IOException e) {
			Debug.out(0,"copyURL2File: IO Failure " + urlstr + " to " + filename);
			return false;
		}
    if (size == 0) {
      Debug.out(0,"copyURL2File: Zero length file. "+ urlstr + " to " + filename);
      return false;
    }  
		return true;
	}

	//----------------------------------------------------------
	/**
	**Method: getInputStream(String urlstr)
	**
	**  Get the Input Stream needed to read the image from the URL string
	*/
	//----------------------------------------------------------

	private InputStream getInputStream(String urlstr) {
		String[] hosts = null;
		boolean found = false;
		int dbcount = 0;
		URL url = null;

		InputStream istream = null;

		// If this is an image from PiconDB, we will need to check if
		// a proxy server is required to get the image.
		// First we generate and save a list of PIcon servers
		if (_PiconDBList == null && _MailFaces != null) // if not yet setup
			_PiconDBList = _MailFaces.getPiconDBRef().getList();
		if (_PiconDBList != null) {
			hosts = new String[_PiconDBList.length];
			for (int i = 0; i < _PiconDBList.length; i++) {
				hosts[i] = _PiconDBList[i].getHost();
				Debug.out("hosts=" + hosts[i]);
			}
		} else
			hosts = new String[0];

		try {
			url = new URL(urlstr);
			if (urlstr.startsWith("http"))
				for (dbcount = 0; dbcount < hosts.length; dbcount++) {
					if (found = url.getHost().equalsIgnoreCase(hosts[dbcount]))
						break;
				}
			if (found && _PiconDBList[dbcount].getProxy() != null) {
				Debug.out("Get input stream from PiconDB");
				istream = _PiconDBList[dbcount].getInputStream(urlstr);
			} else
				istream = url.openStream();
		} catch (MalformedURLException e) {
			Debug.out(0,"FaceCache: InvalidURL - " + urlstr);
      if (Debug.ON) Debug.out(e);
			return null;
		} catch (Exception e) {
			Debug.out(0,"FaceCache: Exception: " + e.getMessage());
			Debug.out(0,"FaceCache: Cannot connect to server for URL: " + urlstr);
			if (Debug.ON) Debug.out(e);
			return null;
		}
		return istream;
	}

	//----------------------------------------------------------------------
	/**
	** handleHTTPStream
	**
	** This routine is invoked when the input stream that is obtained from
	** the URL is an Http Header stream that may or may not include
	** an image.  It is not clear why this only sometimes happens but
	** we need to handle it.
	*/
	//-------------------------------------------------------------------
	private void handleHTTPStream(
		byte[] buffer,
		int len,
		InputStream istream,
		FileOutputStream file)
		throws IOException {
		Debug.out("handleHTTPStream()");
		ByteArrayOutputStream bytes = new ByteArrayOutputStream();
		// Build an inmemory buffer of the whole thing.
		bytes.write(buffer, 0, len);
		while (true) {
			int numbytes = istream.read(buffer);
			//Debug.out("Read " + numbytes);
			if (numbytes == -1)
				break;
			bytes.write(buffer, 0, numbytes);
		}
		byte[] filebuffer = bytes.toByteArray();
		String filestring = bytes.toString();
		Debug.out("Total=" + filebuffer.length);
		BufferedReader br = new BufferedReader(new StringReader(filestring));
		int headerlen = 0;
		// read thru it looking for a null line.  That's the end of the header
		try {
			while (true) {
				String line = br.readLine();
				Debug.out(line.length() + ":" + line);
				headerlen += line.length();
				if (line.length() == 0)
					break;
			}
			int offset = filestring.length() - br.read();
			Debug.out("offset=" + offset);
			file.write(filebuffer, offset, filebuffer.length - offset);
			file.close();
		} catch (Exception e) {
			Debug.out(e);
		}

	}

	//-----------------------------------------------------
	/**
	** Adding this makes this class self-testing
	*/
	//-----------------------------------------------------
	public static void main(String args[]) {

		String dir = null;
		String user = "jthomas";
		String host1 = "cruzio.com";
		String host2 = "webcom.com";
		String url = "file:/c:/jthomas/my icons/duke.gif";
		boolean list = false;
		boolean expire = false;
		boolean create = false;

		Debug.setDebugOn(true);
		if (args.length < 1) {
			Debug.out("syntax: java FaceCache directory");
			System.exit(4);
		} else {
			for (int i = 0; i < args.length; i++) {
				if (args[i].equals("-l"))
					list = true;
				else if (args[i].equals("-e"))
					expire = true;
				else if (args[i].equals("-c"))
					create = true;
				else
					dir = args[i];
			}
		}

		Debug.out(dir);
		FaceCache cache = null;
		try {
			cache = new FaceCache(dir);
		} catch (IOException e) {
		}
		if (cache.getCount() == 0 && create) {

			String newurl = cache.cacheImage(url, null, host1);
			Debug.out(newurl);
			newurl = cache.cacheImage(url, user, host2);
			Debug.out(newurl);

			newurl = cache.getHost(host1);
			Debug.out(newurl);

			cache.writeIndex();
			Debug.out("Wrote the Index file");
			cache.readIndex();

			newurl = cache.getHost(host2);
			Debug.out(newurl);
		}

		if (list) {

			cache.listIndex();
			System.exit(0);
		}

		if (expire) {
			cache.expire();
			cache.verify(5);
			//cache.writeIndex();
			//cache.readIndex();
			cache.listIndex();
		}
		System.exit(0);
	}

} // End of FaceCache Class

//------------------------------------------------------------------------------
//------------------------------------------------------------------------------

//---------------------------------------------------------------
// FaceCacheEntry class
//---------------------------------------------------------------
/**
 *
 * Private Class to implement the Enties in FaceCache. 
 *   
 *@see FaceCache
 */

class FaceCacheEntry {

	String _User;
	String _Host;
	String _URL;
	String _CacheURL;
	private Date _Created;
	private Date _LastUsed;
	private Date _Verified;
	private boolean _Expired;
	//----------------------------------------------------------
	/**
	** This constructor builds an entry from parameters passed
	** by the FaceCache.cacheImage() method.
	**@param url - URL for the icon
	**@param host - host that it represents
	**@param user - userid that cache entry represents
	**@param pathurl - file path to the icon
	*/
	//----------------------------------------------------------
	FaceCacheEntry(String url, String host, String user, String pathurl) {

		_User = user;
		_Host = host;
		_URL = url;
		_CacheURL = pathurl;
		_Created = new Date();
		_LastUsed = new Date();
		_Verified = new Date();

	}

	//------------------------------------------------------------
	/**
	** This constructor builds an entry using the values written
	** by the dump() method
	** This constructor is used when rebuilding the cache entries from the
	** file. 
	**
	**@param entry - String that has to be parsed to provide the parameters
	**   that are passed to the other constructor
	*/

	//------------------------------------------------------------- 
	FaceCacheEntry(String entry) {
		if (Debug.ON) Debug.out(0, "FaceCacheEntry: " + entry);
		String[] tokens = new String[10];
		StringTokenizer st = new StringTokenizer(entry," ");
		for (int i = 0; st.hasMoreTokens(); i++) {
			String token = st.nextToken(" ");

			if (token.startsWith("\"") && (!token.endsWith("\""))) {
				// handle quoted string with embedded blank  
				String rest = st.nextToken("\"");
				tokens[i] =	token.substring(1)+rest;
				// get rid of the "
        st.nextToken(" ");		
						
			} else if (token.equals("null")) {
				tokens[i] = null;
			} else {
				tokens[i] = token;
			}
      //if (Debug.ON) Debug.out("i="+i+" "+tokens[i]+";");
		}

		_URL = tokens[0];
		_User = tokens[1];
		_Host = tokens[2];
		_CacheURL = tokens[3];
		try {
			_Created = new Date(Long.parseLong(tokens[4]));

			_LastUsed = new Date(Long.parseLong(tokens[5]));

			if (tokens[6] != null)
				_Verified = new Date(Long.parseLong(tokens[6]));
			else
				_Verified = _Created;
		} catch (Exception e) { // probably a date format exception
			Debug.out(0,"FaceCacheEntry: " + entry + "Exception: " + e.getMessage());
			Debug.out(0, "Cache Entry had bad date. Replace with today's date");
			_Created = new Date();
			_LastUsed = new Date();
			_Verified = new Date();
		}

	}

	//----------------------------------------------------------------
	/**
	** Create a string for debugging
	**   
	*/
	//---------------------------------------------------------------
	public String toString() {
		return (_URL + " " + _User + " " + _Host + " " + _CacheURL);

	}

	//----------------------------------------------------------------
	/**
	** Create a string with the persistant elements of the entry
	**   This is used to store the entry in the Index file
	*/
	//---------------------------------------------------------------
	String dump() {
		StringBuffer sb = new StringBuffer(200);
		sb.append(quote(_URL));
		sb.append(" ");
		sb.append(_User);
		sb.append(" ");
		sb.append(_Host);
		sb.append(" ");
		sb.append(quote(_CacheURL));
		sb.append(" ");
		sb.append(Long.toString(_Created.getTime()));
		sb.append(" ");
		sb.append(Long.toString(_LastUsed.getTime()));
		sb.append(" ");
		sb.append(Long.toString(_Verified.getTime()));
		sb.append("\n");

		return sb.toString();
	}

	//----------------------------------------------------------------
	/**
	 * Quote the string if it contains a blank
	 * 
	 * @version str  the String to be checked
	 * 
	 * @returns String theat may be quoted.
	 */

	private String quote(String str) {
		if (str.indexOf(' ') >= 0) {
			return '"' + str + '"';
		} else {
			return str;
		}
	}

	//----------------------------------------------------------------
	/**
	** List the Entry
	**   
	*/
	//---------------------------------------------------------------
	public String list() {
		StringBuffer sb = new StringBuffer(100);

    sb.append( _User+" @ "+ _Host+ " URL="+ _URL
          + " Created: "+ _Created.toString()
          + " Verified: "+ _Verified.toString()
          + " Used: "+ _LastUsed.toString());
		sb.append("\n");

		try {
			if (_CacheURL != null) {
				sb.append("  CachedImage: '" + _CacheURL + "'");
				File image = new File((new URL(_CacheURL)).getFile());
				if (!image.exists())
					sb.append(" <<-- Cached Image file does not exist");
			}
		} catch (Exception e) {
			Debug.out("FaceCacheEntry:list() Failed listing " + _CacheURL);
		}

		return sb.toString();

	}

	//----------------------------------------------------------------
	/**
	** checkExpire
	**   This method is called to check if the entry should be expired.
	**   If so it returns true.
	*/
	//---------------------------------------------------------------
	public boolean checkExpire(Date unusedDate) {

		return unusedDate.after(_LastUsed);
	}

	//----------------------------------------------------------------
	/**
	** checkVerify
	**   This method is called to check if the entry should be verified.
	**   If so, it returns true and resets the date verified to today.
	*/
	//---------------------------------------------------------------
	public boolean checkVerify(Date verifyDate) {

		if (verifyDate.after(_Verified)) {
			_Verified = new Date();
			return true;
		} else
			return false;

	}

	//----------------------------------------------------
	/** 
	**   Update the lastUsed date in the entry
	*/
	//----------------------------------------------------

	public void reference() {
		_LastUsed = new Date();
	}

	/**
	** return true if entry is expired.
	*/
	public boolean getExpired() {
		return _Expired;
	}

	/**
	** Set expiration state
	*/
	public void setExpired(boolean truefalse) {
		_Expired = truefalse;
	}

} // End of class FaceCacheEntry