package com.jthomas.mailfaces;

/*
 * FaceFrame.java
 * Copyright (c) 1996 John Thomas  jthomas@cruzio.com
 *      All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for non-commercial purposes
 * is hereby granted provided that this copyright notice
 * appears in all copies.
 *
 */

import java.awt.*;
import java.awt.event.*;
import java.net.*;
import java.util.*;
import java.io.*;
import com.jthomas.util.*;

/**
 * Class to handle the display of the main MailFaces window.
 * This class sets up an AWT Frame that will contain:<br>
 *   Menubar
 *   FaceBaord  - Panel that contains the faces
 *   FaceStatusLabel - Label that tells what is happening
 * <p>
 * This class defines the Menubar elements and the methods invoked 
 * when the menu option is selected.
 * <p> 
 * @see MailFaces  
 * @see Face  
 * @see FaceButton  
 * @see FaceBoard  
 * @see FaceStatusLabel 
 * @see FaceScrollablePanel 
 * @author	John Thomas
 * <a href="mailto:jthomas@cruzio.com>jthomas@cruzio.com</a>
 * @version	2.0 ( May 1998 )
 * 
 */

public class FaceFrame extends Frame {

  // Class variables
  //static final int ICONSIZE=52;         // 48x48 plus 2 pixel border	
  //static final int VGAP = 4;			// Vertical gap between components
  //static final int HGAP = 4;			// Horz gap between components

  // state variables

  private FaceScrollablePanel _ScrollPanel;
  private Panel _FacePanel; // Panel that will hold Picon buttons
  //private Panel 		_MainPanel;				// Panel that holds scrollable panel and bar.
  private FaceStatusLabel _Status; // Label that will hold status info
  private MenuBar _MBar;

  private MailFaces _MailFaces;
  private FaceOptions _Options;
  private FaceProperties _Properties;
  private FaceSMS _FaceSMS;

  /**
  ** Reference to the FaceBoard panel that contains the Face icons
  */
  private FaceBoard _FaceBoard;
  private boolean _NoMailIcon; // true if NoMailIcon present
  private int _Width; // Width of Main window
  private int _WidthOld; // Width of Main window before entry to paint()
  private int _HeightOld; // Height of Main window before entry to paint()
  private int _Height; // Height of Main window
  private Point _Location;
  // Menu items
  private CheckboxMenuItem _Audio;
  private CheckboxMenuItem _Statusbox;
  private CheckboxMenuItem _Debugbox;
  private CheckboxMenuItem _SMSbox;
  private MenuItem[] _ServerMenuItems;
  private MenuItem _DelayLongOn;
  private MenuItem _DelayLongOff;

  /**
  ** MailList frame used for ListMail menuitem
  */
  private MailList _MailList = null;

  //-------------------------------
  /**
  ** FaceFrame Constructor
  ** It only initializes a few references to other objects.
  ** The real initialization happens in the setup() method
  */
  //-------------------------------
  FaceFrame(FaceBoard faceboard) {
    super("MailFaces"); // Sets the Window title
    _FaceBoard = faceboard;
    // get a reference to the current Options object and Properties Object
    _Options = FaceOptions.getRef();
    _Properties = FaceProperties.getRef();
    _MailFaces = MailFaces.getRef();
    _Status = new FaceStatusLabel();
    _FaceSMS = _MailFaces.getFaceSMSRef();
    //------------
    // Handle Window closing event
    //------------
    addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e) {
        Debug.out("Window Closing Event");
        MailFaces.getRef().setFlagQuitMail(true); // Flag to signal Exit
        MailFaces.getRef().setFlagCheckMail(true); // Flag to end delay loop
        FaceFrame.this.dispose();
      }
    });
    //------------
    // Handle Window resize and window move event
    //------------
    addComponentListener(new ComponentAdapter() {
      public void componentMoved(ComponentEvent e) {
        Debug.out("Component moved Event:" + e);
        handleMove();
      }
      public void componentResized(ComponentEvent e) {
        Debug.out("Component resized Event:" + e);
        handleResize();
      }

    });

 
    
    // Process the Options and call setup    
    
    refreshOptions();
    // ask to be informed if the Options get changed
    _MailFaces.addFaceEventListener( new FaceEventListener() {
    	public void
    	actionPerformed(FaceEvent e) {
    		if (e.getID() == FaceEvent.ID_OPTIONS_CHANGED) 			
    			FaceFrame.this.refreshOptions();
    	}
    } );
    
  }
  //-----------------------------------------------------
  /**
   *Method: refreshOptions()
   *@return void
   *
   *   Invoked at startup and when user refreshes the Option file
   */
  //-----------------------------------------------------
  
  public void
  refreshOptions()  {
  	Debug.out("FaceFrame:refreshOptions()");
  	setup();
  	
  }
  
  //------------------------------------------------------
  /**
  ** setup
  **    Called by refresgOptions to setup the
  **    Main window frame 
  */
  //------------------------------------------------------
  private void setup() {

    this.setLayout(new BorderLayout(0, 0));
    // The following will setup a menubar and add it to the main window
    // First we set the size wide enough that the menubar will fit 
    // so it doesn't try to reserver space for a multiline menubar.
    this.setSize(250, 60);

    _MBar = new MenuBar();
    InitializeMenus(_MBar);
    setMenuBar(_MBar);

    this.add("Center", _FaceBoard);

    boolean status = _Options.getBoolean("StatusMsg", true);
    FaceStatusLabel.setStatusFlag(true); // this will setup minimum size also

    _Status.setVisible(true);
    this.add("South", _Status);

    FaceStatusLabel.setStatusFlag(status);
    // Now setup the size and position of the main frame    

    // Get positioning info from Properties file
    int x = Integer.parseInt(_Properties.getProperty("FaceFrameX", "10"));
    int y = Integer.parseInt(_Properties.getProperty("FaceFrameY", "10"));
    this.setLocation(x, y);
    _Location = this.getLocation();

    // Now set the size of the window.  
    // We have a problem here.  On Win95 if we do a pack()  
    // then insets get set too large.  So we do a little dance.
    // We calculate the size and set the Frame to that size.
    // That will be too small because the insets were not set so
    // then we do a pack() and recalculate the size again.
    // Now we do a setSize and a Show().   
    //this.pack();
    Dimension d = this.getMinimumSize();
    this.setSize(d);

    //Debug.out(this);
    this.pack();
    d = this.getMinimumSize();
    this.setSize(d);
    this.show();
    _HeightOld = d.height;
    _WidthOld = d.width;

  }

  //-----------------------------------------------------------------
  /**
  ** handleMove:
  **    This method is called when the frame has recieved a 
  **    componentMoved event. If the new position is somewhere on 
  **    the screen, then the Window position information is saved in
  **    the properties file 
  */
  //-----------------------------------------------------------------   
  void handleMove() {

    // check if window was moved.
    Point location = this.getLocation();
    Dimension size = this.getSize();
    Debug.out("FaceFrame.handleMove():  at " + location);
    if (!location.equals(_Location)) {
      // check if on screen
      Dimension screen = Toolkit.getDefaultToolkit().getScreenSize();
      Debug.out(screen);
      if (location.x >= screen.width
        || location.x + size.width < 0
        || location.y >= screen.height
        || location.y + size.height < 0)
        return;

      _Properties.put("FaceFrameX", "" + location.x);
      _Properties.put("FaceFrameY", "" + location.y);
      _Properties.save();
      Debug.out("Properties updated to " + location);
      _Location = location;
    }

  }
  //-----------------------------------------------------------------
  /**
  ** handleResize:
  **    This method is called when the frame has recieved a 
  **    componentResized event. We call the Faceboard paint()
  **    method and then cause the components to be readjusted 
  **    as necessary. 
  ** 
  **  There may be better ways to do this!
  */
  //-----------------------------------------------------------------   
  void handleResize() {
    Dimension size = this.getSize();
    Debug.out("FaceFrame.handleResize(): " + size);

    Graphics g = _FaceBoard.getGraphics();
    _FaceBoard.paint(g);
    this.validate();

  }

  //-----------------------------------------------------------------
  /**
  ** paint:
  **    This method is called when the system needs to repaint the window.
  **    including when it has been resized 
  */
  //-----------------------------------------------------------------   
  public synchronized void paint(Graphics g) {

    // Check if window has been resized
    Dimension d = this.getSize();
    int width = d.width;

    if (width == 0 || width == _WidthOld) {
      // No change in width

      super.paint(g); // Handle the components

      return;
    }

    _FaceBoard.paint(g);
    this.validate();
  }

  //------------------------------------------------------------
  /**
  **Method: setCursorBusy(boolean)
  **
  **		Set a Busy cursor when true
  */
  //------------------------------------------------------------
  public void setCursorBusy(boolean busy) {
    if (busy)
      _FaceBoard.setCursor(new Cursor(WAIT_CURSOR));
    else
      _FaceBoard.setCursor(new Cursor(DEFAULT_CURSOR));
  }

  //------------------------------------------------------------
  /**
  **Method: setSMSActive(boolean)
  **
  **	Set the menubar item to reflect the current status
  * @param  onOff   The new status 
  */
  //------------------------------------------------------------
  public void setSMSActive(boolean onOff) {
    
    _SMSbox.setState(onOff);
    Debug.out("FaceFrame.setSMSActive: SMS = "+onOff);
  }

  //-----------------------------------
  /**
  ** This will build the menubar "mbar"
  */
  //-----------------------------------
  private void InitializeMenus(MenuBar mbar) {
    MenuItem mi = null;
    Menu m = new Menu("File");

    //
    // Servers
    //
    Menu serverMenu = new Menu("Servers");
    int n = FaceServer.getCount();
    if (n > 1) {
      _ServerMenuItems = new MenuItem[n];
      for (int i = 0; i < n; i++) {
        String name = FaceServer.getList()[i].getName();
        _ServerMenuItems[i] = new MenuItem(name);
        serverMenu.add(_ServerMenuItems[i]);
        final int servernumber = i;
        _ServerMenuItems[i].addActionListener(new ActionListener() {
          public void actionPerformed(ActionEvent e) {
            MailFaces.getRef().setServer(servernumber);

          }
        });

      }
    } else {
      serverMenu.setEnabled(false);
    }
    m.add(serverMenu);

    //
    // Check for new Mail
    //
    mi = new MenuItem("Check for new Mail");
    m.add(mi);
    mi.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        MailFaces.getRef().setFlagCheckMail(true); // Flag to check for new Mail          

      }
    });

    //
    // Verify Icon Selection
    //
    mi = new MenuItem("Verify Icon Selection");
    m.add(mi);
    mi.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        _MailFaces.getResolverRef().checkAddress();
      }
    });

    //
    // Reset
    //
    mi = new MenuItem("Reset");
    m.add(mi);
    mi.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        MailFaces.getRef().setFlagResetMail(true); // Refresh the FaceFrame
        MailFaces.getRef().setFlagCheckMail(true); // Flag to check for new Mail

      }
    });

    m.addSeparator();
    //
    // Exit
    //
    mi = new MenuItem("Exit");
    m.add(mi);
    mi.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        MailFaces.getRef().setFlagQuitMail(true); // Flag to signal Exit  		
        MailFaces.getRef().setFlagCheckMail(true); // Flag to check for new Mail
      }
    });

    mbar.add(m);

    m = new Menu("Options");


    //------------
    // EditOptions
    //---------------
    if (_MailFaces.editOptionsSupported() ) {
	    mi = new MenuItem("Edit Options");
	    m.add(mi);
	    mi.addActionListener(new ActionListener() {
	      public void actionPerformed(ActionEvent e) {
	        MailFaces.getRef().editOptions(); // Edit the Options file		
	      }
	    });
    }

    //------------
    // RefreshOptions
    //---------------
    mi = new MenuItem("Refresh Options");
    m.add(mi);
    mi.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        MailFaces.getRef().refreshOptions(); // Reread the Options file		
      }
    });

    //
    // Audio
    //
    _Audio = new CheckboxMenuItem("Audio");
    _Audio.setState(_Options.get("AudioSupport", "false").equals("true"));
    m.add(_Audio);
    _Audio.addItemListener(new ItemListener() {
      public void itemStateChanged(ItemEvent e) {
        MailFaces.getRef().setFlagAudio(_Audio.getState());
      }
    });



    //
    // DelayLong or Delay option
    //    
    int delaylong = _Options.getInt("DelayTimeLong", 0);
    int delay = _Options.getInt("DelayTime", 300);
    if (delaylong != 0) {
      _DelayLongOn = new MenuItem("Delay " + delaylong / 60 + " Minutes");
      m.add(_DelayLongOn);
      _DelayLongOff = new MenuItem("Delay " + delay / 60 + " Minutes");
      m.add(_DelayLongOff);

      _DelayLongOff.setEnabled(false);
      _DelayLongOn.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          int d = _Options.getInt("DelayTimeLong", 600);
          MailFaces.getRef().setDelay(d);
          _DelayLongOff.setEnabled(true);
          _DelayLongOn.setEnabled(false);
          MailFaces.getRef().setFlagCheckMail(true); // Flag to end delay loop
        }
      });
      _DelayLongOff.addActionListener(new ActionListener() {
        public void actionPerformed(ActionEvent e) {
          int d = _Options.getInt("DelayTime", 30);
          MailFaces.getRef().setDelay(d);
          _DelayLongOn.setEnabled(true);
          _DelayLongOff.setEnabled(false);
          MailFaces.getRef().setFlagCheckMail(true); // Flag to end delay loop
        }
      });

    }
    
    //
    // SMS on/off
    //
    _SMSbox = new CheckboxMenuItem("SMS");
    Debug.out("FaceFrame.InitializeMenus: SMS = "+_FaceSMS.getSMSActive());
    _SMSbox.setState(_FaceSMS.getSMSActive());
    m.add(_SMSbox);
    _SMSbox.addItemListener(new ItemListener() {
      public void itemStateChanged(ItemEvent e) {
        boolean state = _SMSbox.getState();
        Debug.out("FaceFrame.InitializeMenus: SMS = "+state);
        _FaceSMS.setSMSActive(state);
        
      }
    });

    //
    // Debug on/off
    //
    _Debugbox = new CheckboxMenuItem("Debug");
    _Debugbox.setState(Debug.isDebugOn());
    m.add(_Debugbox);
    _Debugbox.addItemListener(new ItemListener() {
      public void itemStateChanged(ItemEvent e) {
        boolean state = _Debugbox.getState();
        Debug.out(0,"FaceFrame.InitializeMenus: Set Debug = " + state);
        Debug.setDebugOn(state);
      }
    });

    // add "Options" to menubar
    mbar.add(m);

    //-----------------
    // Cache Menubar
    //-----------------
    m = new Menu("View");
    //
    // ListMail
    //
    MenuItem listMail = new MenuItem("ListMail");
    m.add(listMail);
    listMail.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {

        if (_MailList != null)
          _MailList.dispose();

        _MailList = new MailList(_FaceBoard.getAllMailmsgs());

      }
    });

    //
    // StatusLabel on/off
    //
    _Statusbox = new CheckboxMenuItem("Status Message");
    _Statusbox.setState(_Options.getBoolean("StatusMsg", true));
    m.add(_Statusbox);
    _Statusbox.addItemListener(new ItemListener() {
      public void itemStateChanged(ItemEvent e) {
        _Options.put("StatusMsg", (_Statusbox.getState() ? "true" : "false"));
        FaceStatusLabel.setStatusFlag(_Statusbox.getState());
        // cause window to show or hide 
        FaceFrame.this.validate();
        FaceFrame.this.pack();

      }
    });
    //
    // Debuginfo
    //
    MenuItem debugDisplay = new MenuItem("DebugInfo");
    m.add(debugDisplay);
    debugDisplay.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        MailFaces.displayDebugInfo();
      }
    });
    
    //
    // Groups
    //
    MenuItem groupDisplay = new MenuItem("Groups");
    m.add(groupDisplay);
    groupDisplay.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        FaceGroup.getRef().display();
      }
    });
    

    //-----------------
    // Cache Menubar
    //
    Menu cache = new Menu("Cache");

    mi = new MenuItem("Cache Expire");
    cache.add(mi);
    mi.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        _MailFaces.fireFaceEvent(FaceEvent.ID_CACHE_EXPIRE);
      }
    });
    //
    // Cache Clear 
    //
    mi = new MenuItem("Cache Clear");
    cache.add(mi);
    mi.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        _MailFaces.fireFaceEvent(FaceEvent.ID_CACHE_CLEAR);
      }
    });
    //
    // Cache List 
    //
    mi = new MenuItem("Cache List");
    cache.add(mi);
    mi.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        _MailFaces.fireFaceEvent(FaceEvent.ID_CACHE_LIST);
      }
    });
    //
    // Cache Verify 
    //
    mi = new MenuItem("Cache Verify");
    cache.add(mi);
    mi.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        _MailFaces.fireFaceEvent(FaceEvent.ID_CACHE_VERIFY);
      }
    });
    m.add(cache);
    
    // add "View" to menubar
    mbar.add(m);
    
    //-----------------
    // Help Menubar
    //-----------------
    m = new Menu("Help");

    mi = new MenuItem("Help");
    m.add(mi);
    mi.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        String urlstr = FaceOptions.getRef().get("HelpURL"); // URL for Help info            		
        // Now we need to tell the user to use the indicated URL in
        // his browser to display the Help Info
        //String out = "Use your Web browser to view:\n  " + urlstr;
        //MsgDialog.create(FaceFrame.this, out, false);
        Prompter p = Prompter.create();    
        p.setTextInputPrompt("Use your web browser to view:",
                              urlstr,"MailFaces Help");
        p.showPrompt();
      }
    });

    m.addSeparator();

    // About
    mi = new MenuItem("About");
    m.add(mi);
    mi.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        String v = MailFaces.version(); // Java code Version
        String out =
          "MailFaces:\n"
            + " Version "
            + v
            + "\n"
            + " by John Thomas\n"
            + "    jthomas@cruzio.com\n"
            + "    http://www.cruzio.com/~jthomas";
        MsgDialog.create(FaceFrame.this, out, false);
      }
    });

    mbar.setHelpMenu(m);
  }

  /**
  ** Calculates the FaceFrame dimensions
  */

  public Dimension getMinimumSize() {
    int width;
    int height;
    Insets in = this.getInsets();

    Dimension fb = _FaceBoard.getMinimumSize();
    Dimension st = _Status.getMinimumSize();
    width = in.left + in.right + fb.width;
    height = in.top + in.bottom + fb.height + st.height;
    Debug.out("FaceFrame.getMinimumSize: " + width + "x" + height);
    return new Dimension(width, height);

  }

  /**
  ** Calculates the FaceFrame dimensions
  */

  public Dimension getPreferredSize() {
    //Debug.out("PreferredSize");
    return getMinimumSize();
  }

  /**
  ** This will workaround a bug in JDK 1.1.5 and earlier 
  ** where Insets returned a negative value for the bottom
  ** inset.  This seems to have been fixed in JDK 1.1.6
  */
  public Insets getInsets() {
    Insets in = super.getInsets();

    if (in.bottom < 0) {
      //Debug.out("Adjusting bottom inset from "+in.bottom+" to 4");
      return new Insets(in.top, in.left, 4, in.right);
    } else
      return in;
  }

} //------- End of class FaceFrame