package com.jthomas.mailfaces;

/*
 * FaceGroup.java
 * Copyright (c) 1996 John Thomas  jthomas@cruzio.com
 *      All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for non-commercial purposes
 * is hereby granted provided that this copyright notice
 * appears in all copies.
 *
 */

import java.io.*;
import java.util.*;


import com.jthomas.util.Debug;
import com.jthomas.util.Log;

/**
 * The FaceGroup class will handle Group statements from the .opt file.
 *
 * <pre>
 * Group statements have the form:
 *   Group name  filename
 * 
 * The file will contain a list of email addresses in a 
 * comma separated format. .csv 
 * Most Adress books have the capability of creating an 
 * export file that is in CSV format.
 * 
 * It will also support an Eudora Nickname file that contains:
 *   alias name email
 *   note ...
 *  
 * 
 * We will parse the lines in the file and build a hashtable that 
 * contains each email address as the key.  
 * A hashtable with a key of group name will keep track of the 
 * individual group hashtables. 
 *    
 * The matchGroup(groupname,address) method will be used to determine if
 * the specified address is in the specified group.  
 * 
 *
 * @see FaceResolver 
 * @see Mailmsg 
 * @see FaceOptions
 * 
 * @author	John Thomas
 * <a href="mailto:jthomas@cruzio.com>jthomas@cruzio.com</a>
 * @version	1.0 ( April 2002 )
 * 
 */

public class FaceGroup {

 
  private static FaceGroup  _FaceGroup = null;
  // state variables
  
  /**
   * reference to the Options 
   */
  private FaceOptions _Options;
  
   /**
   * Array of Group statements
   */
  private String[] _GroupList;

 
  
  /**
   * Number of Group statements
   */
  private int _GroupCount;

 
  /**
   * Hashtable to keep trak of groups.
   * key: group name
   * data: Hashtable containing member email addresses
   */
  private Hashtable _GroupTable = null;
 
 
  //------------------------------------------------------
  //Method: getRef()
  /**
  ** Static Class Method to access the one and Only FaceOptions object
  */
  //------------------------------------------------------
  static public FaceGroup getRef() {
	    return _FaceGroup;
  }

  //-----------------------------------------
  // Constructor
  //------------------------------------------

  public FaceGroup() {
    //Debug.out(0,"FaceGroup.init()");
    _Options = FaceOptions.getRef();
    _FaceGroup = this;
    // Process the Options
    refreshOptions();
    // ask to be informed if the Options get changed
    MailFaces.getRef().addFaceEventListener(new FaceEventListener() {
      public void actionPerformed(FaceEvent e) {
        if (e.getID() == FaceEvent.ID_OPTIONS_CHANGED)
          FaceGroup.this.refreshOptions();
      }
    });
    
  }
  
  
 

  //------------------------------------------------------------------
  /**
   * refreshOptions() gets control initially and when the 
   * user refreshes the Option file.
   *
  *   The Group statements have the syntax:<br>
  *   Group group1 filename1 
  *   
  *   
  ** <pre>
  **   - Initialize the FaceGroup object.
  **   - Get array of Group statements 
  *    - For each statenment: 
  **     - parse the statement 
  *      - read the file and build a hashtable
  *        of the email addresses
  **     - store the hashtabel in a hashmap with the groupname
  *        as the key.
  *    
  *  </pre>
  */
  //--------------------------------------------------------------------
  public void refreshOptions() {

    // Setup the Group list
    _GroupList = FaceOptions.getRef().getList("GROUP");
    boolean groupDisplay = FaceOptions.getRef().getBoolean("GroupDisplay",false);
    _GroupCount = 0;
    int n = _GroupList.length+1;  // 1 or more slots
    _GroupTable = new Hashtable(n);
    
    int k = 0;
    
    for (int i = 0; i < _GroupList.length; i++) {
      String statement = _GroupList[i];
      Debug.out("FaceGroup: " + statement);
      String[] tokens = parseStatement(statement);
      if (Debug.ON) {
        StringBuffer sb = new StringBuffer("FaceGroup: ");
        for (int di=0;di<tokens.length;di++) {
          sb.append(di+": "+tokens[di]+" ");
        }
        Debug.out(0,sb.toString());
      }
      //
      // if parseStatements returns token[0] == null then
      // it found and error and reported it.  
      // we just ignore it and continue.  
      if (tokens[0] == null) 
        continue;
      
      String groupname = tokens[0]; 
      String filename = tokens[1];
      Debug.out("FaceGroup: " + groupname+" "+filename);
      // get the email addresses from file and load the Hashtable.
      Hashtable members = getAddressesFromFile(filename);
      _GroupTable.put(groupname,members);
      
 
    } // end for()
    _GroupCount = k; 
    // if user requested display of groups 
    if (groupDisplay) 
      display();
      
  } // setup


  //--------------------------------------------------------------------
  /**
  ** Method to parse the Group statement. 
  **
  ** @param statement The String that was read from the file.
  ** @return  Array of tokens The last token is null.
  */
  //--------------------------------------------------------------------
  private String[] parseStatement(String statement) {

    int MAX = 3;
    int n = 0;
    //Debug.out(0,"statement="+statement);
    String[] tokens = new String[MAX];
    
    StringTokenizer tstr = new StringTokenizer(statement);
    while (tstr.hasMoreTokens()) {
      if (n >= MAX) {
        Debug.out(0, "Too many tokens. " + statement);
        String errorMsg ="Error parsing Statement: "+ statement+ "\n"
                + "Too many parameters";
        MsgDialog.create(errorMsg);
        tokens[0] = null;
        break;
      }
      String token = tstr.nextToken();
      tokens[n] = token;
      n++;
    } // end while
    return tokens;
  }


  //--------------------------------------------------------------------
  /**
  ** 
  **   reads the specified file and parses the comma separated lines 
  *    to get the list of email addresses.
  *   
  * 
  * @param filename    The name of the file.
  * 
  * @return HashTable if able to read the specified file and 
  * discover 1 or more email addresses.
  * 
  */
  //----------------------------------------------------------------------
  private Hashtable getAddressesFromFile(String filename) {
    Hashtable group  = new Hashtable(100);
    try {
      //Debug.out(0,"filename="+filename);
      File file = new File(filename);
      FileReader fileReader = new FileReader(file);
      BufferedReader bufferReader = new BufferedReader(fileReader);
      while (true) {
        String line = bufferReader.readLine();
        //Debug.out(0,"Read: "+line);
        if (line == null) 
          break;
        String email;
        if (line.startsWith("alias ") ||
            line.startsWith("note ") ) {
          email = parseEudora(line);
        } else {
          email = parseCSV(line);
        }
        if (email != null) {
          email = email.toLowerCase();
          Debug.out("Group email="+email);
          group.put(email,"");
        }  
      }
    } catch (Exception e) {
      Debug.out(0,e);
      String errorMsg ="Error reading Group file: "+ filename+ "\n"
                + e.getMessage();
      MsgDialog.create(errorMsg);
      
    }
    return group;
  }


  /**
  ** Method to parse a CSV file for the email address. 
  **
  ** @param csv The String that was read from the file.
  ** @return  email address or null if not found.
  */
  private String parseCSV(String csv) {

    String token = null;
    
    //Debug.out(0,"csv="+csv);    
    
    StringTokenizer tstr = new StringTokenizer(csv,",");
    while (tstr.hasMoreTokens()) {
      
      token = tstr.nextToken();
      if (token.indexOf("@") > 0) {
        if (token.startsWith("\"") )
          token = token.substring(1);
        if (token.endsWith("\""))
          token = token.substring(0,token.length()-1);
        
        if (token.indexOf("<") >= 0)
          token = parseEmailAddress(token);
        //Debug.out(0,"email="+token);  
        break;
      } else {
        token = null;
      }
     
    } // end while
    return token;
  }

  /**
  ** Method to parse a Eudora Nickname file for the mail address. 
  ** 
  *  Look for lines that start with "alias"
  *   aliase name email 
  * 
  ** @param line The String that was read from the file.
  ** @return  email address or null if not found.
  */
  private String parseEudora(String line) {

    String token = null;
    
    
    if (line.startsWith("alias ")) {
      StringTokenizer tstr = new StringTokenizer(line);
      int i = 0;
      while (tstr.hasMoreTokens()) {
        token = tstr.nextToken();
        i++;
        if (token.indexOf("@") > 0) {
          if (token.startsWith("\""))
            token = token.substring(1);
          if (token.endsWith("\""))
            token = token.substring(0, token.length() - 1);

          if (token.indexOf("<") >= 0)
            token = parseEmailAddress(token);
          //Debug.out(0,"email="+token);  
          break;
        } else {
          token = null;
        }

      } // end while
    }
    
    return token;
  }
  //--------------------------------------------------------------------
  /**
  ** 
  **   Determines if the sender of the mailmsg is in the specified group.
  *   
  * 
  * @param group    The group name that should be checked.
  * @param mailmsg  The mail that is being checked.
  * @return true if from or reply-to address is a member of the 
  * specified group.
  */
  //----------------------------------------------------------------------
  public boolean resolve(String group, Mailmsg mailmsg) {

    if (Debug.ON) Debug.out("FaceGroup.resolve: "+group + " "+ mailmsg);
    
    Hashtable table = (Hashtable)_GroupTable.get(group);
    if (table == null) {
      Debug.out(0,"Group "+group+" Not found.");
      return false;
    }
    String email = mailmsg.getHeader("from");
    email = parseEmailAddress(email);
    String test = (String)table.get(email);
    if (test != null) 
      return true;
    email = mailmsg.getHeader("reply-to");
    email = parseEmailAddress(email);
    test = (String)table.get(email);
    if (test != null) 
      return true;
    else 
      return false;
  }


  /**
  ** Method to parse a network email addres to extract the user@system portion. 
  **
  ** @param statement The String that was specified in the email.
  ** @return  email address 
  */
  private String parseEmailAddress(String email) {

    
    //Debug.out(0,"Old email="+email);
    int i = 0;
    if ((i = email.indexOf("<")) >=0 ) {
      email = email.substring(i+1);
      if ((i = email.indexOf(">")) >=0 ) {
        email = email.substring(0,i);
      }
    }
    email= email.toLowerCase();
    //Debug.out(0,"New email="+email);
    return email;
  }
  
  
  /**
  ** Method to display the Groups. 
  **
  */
  public void display() {

    Enumeration keys = _GroupTable.keys();
    while (keys.hasMoreElements()) {
      String groupname = (String)keys.nextElement();
      System.out.println("Group: "+groupname);
      Hashtable group = (Hashtable)_GroupTable.get(groupname);
      Enumeration members = group.keys();
      while (members.hasMoreElements()) {
        String memberEmail = (String)members.nextElement();
        System.out.println("  "+memberEmail);
      }
    }

  }
  
  
  
} // End of FaceGroup Class