package com.jthomas.mailfaces;

/*
 * FaceMailbox.java
 * Copyright (c) 1996 John Thomas  jthomas@cruzio.com
 *      All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for non-commercial purposes
 * is hereby granted provided that this copyright notice
 * appears in all copies.
 *
 */

import java.io.*;
import java.util.*;
import java.net.*;

import com.jthomas.util.*;

/**
 * FaceMailbox is an abstract class to handle the interface to the Mail Server. 
 *
 * <p>
 *
 * Class to handle a General purpose FaceMailbox
 * This is an abstract class that is intended to be
 * extended by classes that implement methods unique to
 * the type of mail server (POP3, IMAP4 etc)
 * <p>
 * The overall purpose of this class is to periodically 
 * scan a user's mailbox and when it finds a new email message
 * then it will construct a MailMsg object and place it on the
 * queue for the FaceResolver class.  FaceResolver will determine 
 * which icon should be shown for this email. 
 *
 *
 * <p><pre>
 * Methods:
 *  Constructor:
 *    None
 * *
 * *
 *  Public Methods
 *   void   	init()
 *   void 		run()
 
 *
 *   Mailmsg    getMailmsg(String key)
 *   int        count() 
 *
 *	 void		addMailmsg(key,Mailmsg)   
 *
 *	 void       resetMailmsgs()
 *
 *   void		setModified()
 *   boolean    getNModified()
 *
 * The following are abstract methods imlemented by subclass
 * that implements the code for handling some type of mailserver
 * such as POP3 or IMAP4
 *   void  		open()
 *
 *	 void		read()
 *
 *   void 		close()
 * 
 *   void		terminate();
 * *
 *---------------------------------------------------------------------
 
 *---------------------------------------------------------------------
 *  Simple Usage Example to do something with each message
 *
 *-  FaceMailbox mailbox = server.createMailbox();
 *-  mailbox.getOptions();
 *-  mailbox.open();
 *-  mailbox.read();    
 *-  mailbox.close();
 *-  for (i=0; i<mailbox.count(); i++) {
 *-      Mailmsg msg = mailbox.getMailmsg(i);
 *-   	 ...                 
 *-  }
 
 * </pre>
 * @see FaceResolver 
 * @see MailMsg
 * @author	John Thomas
 * <a href="mailto:jthomas@cruzio.com">jthomas@cruzio.com</a>
 * @version	1.0 ( Sept 1996 )
 *
 */

public abstract class FaceMailbox implements Runnable {
  
  public static final String COPY_HEADER = "X-MAILFACES-COPY";
  /*
   * State Variables
   *
   */
  /**
  ** Reference to the "main" application object.
  **   Used for many things 
  */
  private MailFaces _MailFaces; // reference to MailFaces global object

  /**
  ** GUI interface object
  **   We need this to access the FaceBoard queue
  */
  private FaceBoard _FaceBoard;

  /**
  ** Used to play sound when something happens to the mailbox
  */
  private String _DefaultAudioURL;

  /**
  * count of mailmsgs on server. 
  * This is the value that the server returns at open time 
  */
  protected int _Count = 0;

  /**
  ** Reference to the FaceOptions object so we can access 
  ** the options that the user wants to use.
  */

  protected final FaceOptions _Options = FaceOptions.getRef();
  /**
  ** FaceServer object that is controlling this Mailbox
  */
  protected FaceServer _Server;

  /**
  ** Cache of Icons
  */
  protected FaceCache _Cache;

  /**
  ** Object that handles the selolving of a mainlmsg to a 
  ** matching icon.
  */
  private FaceResolver _Resolver;

  /**
  ** Protocol type  POP3, APOP, IMAP
  */
  protected String _Type = null;

  /** 
   * List of valid supported types.
   * Filled in by subclasses in static constructor.
   */
  protected static Hashtable _ValidTypes = new Hashtable();
  
  /**
   * Mail server host
   */
  protected String _Host = null;

  /**
   * User name for mail account
   */
  protected String _User = null;

  /**
   * password for users mail account
   */
  protected String _Password = null;

  /**
  ** Set to true if all login info is specified.
  */
  protected boolean _ValidInfo; // true if Host,User,Password specified

  /**
  * Default address (userid@server) where message should be copied
  */
  //protected  String _CopyAddress = null;

  /**
  ** A table of all the mailmsgs 
  **
  ** The key is a unique string derived from the mailmsg
  **   UIDL for POP3 mail
  */
  private Hashtable _Mailmsgs = new Hashtable(100);
  ;

  /**
   * keep track of which pass thre the mailbox we are at.
   * Mostly we just need to now about the first pass 
   * since we do something different for failures on the 
   * first pass.
   */
  private int _Passes = 0;
  
  /**
   * Keep track that the previous pass failed to open the server
   */
  protected boolean _OpenFailure = false;
  
  /**
  ** Set true if one or more new messages were processed
  */
  protected boolean _Modified;

  protected FaceMailbox() {
  }

  //---------------------------------------------------------------
  /**
   * setServer(FaceServer server) will set a reference to the Server object for
   * this Mailbox.
   * 
   *
   *  This method is called by MailFaces after creating the Mailbox object 
   *  It used to specify it in the constructor
   *
   *  @param server  FaceServer Object that describes the server
   *  @return  void
   */
  //---------------------------------------------------------------

  public void setServer(FaceServer server) {
    _Server = server;
  }

  //---------------------------------------------------------------
  /**
   *Method: init()
   * 
   *
   *  This method is called by MailFaces to do setup
   *  It will initialize a number of references to 
   *  various MailFaces objects.
   *  @return  void
    
   */
  //---------------------------------------------------------------

  public void init() {
    Debug.out("FaceMailbox.init():");
    _MailFaces = MailFaces.getRef(); // Get Global reference
    Debug.out("access Cache");
    _Cache = _MailFaces.getCacheRef();
    Debug.out("access FaceBoard");
    _FaceBoard = _MailFaces.getFaceBoardRef();
    Debug.out("access Resolver");
    _Resolver = _MailFaces.getResolverRef();
    Debug.out("create Hashtable");

    _Mailmsgs = new Hashtable(100);
    Debug.out("refresh Options");
    refreshOptions();
    // ask to be informed if the Options get changed
    _MailFaces.addFaceEventListener(new FaceEventListener() {
      public void actionPerformed(FaceEvent e) {
        if (e.getID() == FaceEvent.ID_OPTIONS_CHANGED)
          FaceMailbox.this.refreshOptions();
      }
    });
    Debug.out("Init ended");
  }

  //---------------------------------------------------------------
  /**
   *Method: run() 
   *  This method is invoked when the Thread is started.
   * <p>
   *  It will sit in a loop doing the following:
   *  <pre>
   *     -Check Flags set by action of the UI component and perform the
   *         action.  These flags are:
   *  	- Quit
   *      - Reset Server
   *		- Refresh Options
   *      - Check for new mail
   *		- expire
   *      - etc
   *    - Call checkmail rtn to see if there are new mail msgs
   *    - Sleep for number of seconds specified.
   *    - Repeat the loop
   *
   *  </pre>
   * @return  void
   */
  //---------------------------------------------------------------

  public void run() {
    Debug.out("FaceMailbox.run()");
    boolean done = false;
    boolean checkmail = true;
    Runtime runtime = Runtime.getRuntime();

    try {
      while (!done) {
        _Passes++; // Keep track of how many times we have done this
        long total = runtime.totalMemory() / 1024;
        long free = runtime.freeMemory() / 1024;
        Log.out("MailboxOpen",
          "Pass " + _Passes + " Memory: Total=" + total + "K" + "  Free=" + free + "K" + "  Used=" + (total - free) + "K");

        // Check for Flags that have been set from Menubar options
        //
        if (_MailFaces.getFlagQuitMail()) { // User says she wants to quit
          terminate();
          break;
        }
        if (_MailFaces.getFlagResetServer()) { // Reset server so stop the mailbox
          resetMail();
          terminate();
          break;
        }

        if (_MailFaces.getFlagResetMail()) { // Reset and regenerate Faces
          resetMail();
          _MailFaces.setFlagResetMail(false);
        }

        // Now do the check for new mail.
        if (checkmail || _MailFaces.getFlagCheckMail()) { // If we need to check the mailbox

          checkMail();
          //_MailFaces._FlagCheckMail = false;
          checkmail = false;
        }

        // Finished a single pass thru the mail,
        // Now delay and then do it again
        // Because the user may want to refresh it earlier and Java currently
        // has no way to wake up a sleeping thread, we check every few seconds
        // for a CheckMail request

        if (_MailFaces.getFlagQuitMail()) { // check for exit request
          terminate();
          break;
        }

        if (Debug.ON)
          Debug.out("FaceMailbox.run: Sleeping " + _MailFaces.getDelay() + " seconds");
        int delaytimes = _MailFaces.getDelay() / 2; // convert to 2 second intervals
        _MailFaces.setBusy(false);

        try {
          for (int i = 0; i < delaytimes; i++) {
            Thread.currentThread().sleep(2000); // sleeeping 2 seconds
            if (i == 5) // after idle for 10 seconds 
              runtime.gc(); // good time to do garbage collection
            if (_MailFaces.getFlagCheckMail()) { // if user wants to do something
              break;
            }
          }
          checkmail = true;
          if (_MailFaces.getFlagCheckMail()) // if delaytime expired    			

            _MailFaces.setFlagCheckMail(false);

        } catch (InterruptedException e) {
          Debug.out(0, "FaceMailbox.run: Interrupted");
          done = true;
        }
        _MailFaces.setBusy(true);
        if (Debug.ON)
          Debug.out("FaceMailbox.run: Awake");
      } // end of while(true)
    } catch (Exception e) {
      Debug.out("FaceMailbox.run(): ", e);

    }

    Debug.out("FaceMailbox.run() terminated");
  }

  //-----------------------------------------------------------------------
  /**
  **Method: resetMailmsgs
  ** 		Mark all the current mailmsgs with an "Unknown" status.
  **      This is done at the start of a pass that compares the 
  **      messages at the server with the Mailmsg objects.
  **      As each msg is matched, the status is changed to "StatusOld".
  **    
  **      At the end of the pass, any Mailmsgs that still have a 
  **      StatusUnknown are then deletable. 
  **
  */
  //------------------------------------------------------------------------
  protected synchronized void resetMailmsgs() {
    Debug.out("FaceMailbox.resetMailmsgs()");
    if (_Mailmsgs != null) {
      Mailmsg mailmsg;
      Enumeration enum = _Mailmsgs.elements();
      for (int i = 0; enum.hasMoreElements(); i++) {
        mailmsg = (Mailmsg) enum.nextElement();
        mailmsg.setServerStatus(Mailmsg.StatusUnknown);
      }
    }
    return;
  }

  //-----------------------------------------------------------------------
  /**
  ** checkMail will	Open and Read the mail from the Server.
  ** 
  **    This is the main processing that happens on each pass
  **    thru the loop.  The intention is that this code will 
  **    be common for all the mail servers (but we will see)
  **
  **    It will open the mailbox, see if there is anything new,
  **    process any pending deletes, and then close the mailbox.
  ** <p>
  **    This will also handle the default audio play at the end of a scan.
  **
  */
  //------------------------------------------------------------------------
  private synchronized void checkMail() {
    boolean opened;
    

    // On 1st pass, verify that we have valid Login Info
    if (_Passes == 1 && !_Server.getLoginInfo()) {
      _MailFaces.setFlagQuitMail(true);
      Debug.out("FaceMailbox.checkmail: User pressed Cancel");
      return;
    }

    FaceStatusLabel.setStatus("Mailbox", "Contact server " + _User + "@" + _Host);
    opened = open();

    while (!opened) {

      FaceStatusLabel.setStatus("Mailbox", "Unable to Open Mailbox");
      // The first time it may be because of a bad address or password
      // so we will warn the user that it failed.  After that it may
      // just be a network problem so ignore it
      if (_Passes == 1) { // On the first try, let user change login info
        boolean ok = MsgDialog.create("Failed to open Mail Server "+_Host +"\n" + 
                               "Press OK to continue  or Cancel to change address");
        if (!ok) {
          _MailFaces.setFlagResetServer(true);
          _MailFaces.setFlagQuitMail(true);
          _Server.setRespecify(true);
          return;
        } else {
          _FaceBoard.queuePut("NOMAIL");
          opened = open(); // Retry it
          if (!opened) {
            _OpenFailure = true;
            Debug.out("FaceMailbox.checkmail: Mailbox unable to Open");
            Log.out("MailboxOpen", "Mailbox unable to Open on pass " + _Passes);
            return;
          }

        }
      } else {
        //Open failure can happen any time.
        _OpenFailure = true;
        Log.out("MailboxOpen", "Mailbox unable to Open on pass " + _Passes);
        Debug.out("FaceMailbox.checkmail: Mailbox unable to Open on pass " + _Passes);
        return;
      }
    }
    if (_OpenFailure) {
      // was failing now it is OK
      Log.out("MailboxOpen", "Mailbox Open OK. pass " + _Passes);
      Debug.out("FaceMailbox.checkmail: Mailbox Open OK. pass " + _Passes);
      _OpenFailure = false;
    }
    
    // The mail server is open and we are logged into it
    if (_Count == 0) { // There is no mail at the server.
      close();
      Debug.out("FaceMailbox.checkmail: Mailbox closed - count=0");
      if (count() != 0 || _Passes == 1) { // if there used to be mail
        // Reset the FaceBoard to show NoMail    				
        //_MailboxMsgCount = 0;    				
        _FaceBoard.queuePut("NOMAIL");
        _Mailmsgs.clear();        
      }
      FaceStatusLabel.clearStatus("Mailbox");
    } else {
      // Make sure we handle anything that goes wrong in processing the
      // mailbox and then close it so any mail server lock is released.
      try {

        // set a flag so we can tell in postProcess 
        // if the mail is gone
        resetMailmsgs();
        // Now get the new messages
        read();

        close();
      } catch (Exception e) {
        Log.exception(e);
        Debug.out("Exception while processing Mailbox", e);

      }

      if (_MailFaces.getFlagQuitMail()) { // User wants to quit
        terminate();
        return;
      }

      // Do a final PostProcess scan of mailmsgs
      // Delete messages that have been removed by an external process
      // For example by the users real email reader
      postProcessMail();

      if (count() == 0) { // may have deleted all msgs    			  
        // Reset the FaceBoard to show NoMail    				
        //_MailboxMsgCount = 0;    				
        _FaceBoard.queuePut("NOMAIL");
        _Mailmsgs.clear();        
      } 
      FaceStatusLabel.clearStatus("Mailbox");
      
      // If a new mailmsg has arrived or some were deleted, then 
      // play a tune  (if user wanted that)
      if (getModified() && _MailFaces.getFlagAudio() && _DefaultAudioURL != null) {
        _MailFaces.getFaceAudioRef().queuePut(_DefaultAudioURL);
        Debug.out("Audio Play:" + _DefaultAudioURL);
      }
      setModified(false);
    }

    return;
  }

  //---------------------------------------------------------------
  /**
   *Method: refreshOptions()
   * @return  void
   *  This method is invoked when the MailFaces options are modified
   *
   */
  //---------------------------------------------------------------

  public void refreshOptions() {

    //  Get the Name of the Host and the userid and password for the user
    _Type = _Server.getType().toLowerCase();
    _Host = _Server.getHost();
    _User = _Server.getUser();
    _Password = _Server.getPass();
    _DefaultAudioURL = _Options.get("DefaultAudioURL", null);
    //FaceAudio.getRef().queuePut(_DefaultAudioURL); // test  it
    
  }

  //---------------------------------------------------------------
  //Method: verifyLoginInfo()
  //
  //  Private method to verify that the needed Login info has be gotten from user
  //
  //---------------------------------------------------------------

  protected boolean verifyLoginInfo() {

    if (_Server != null)
      return _Server.verifyLoginInfo();
    else
      return false;

  }

  //------------------------------------------------------------------
  /**
   * Method:  resetMail()
   *
   *   Reset the Mailbox and Faceboard so everything gets regenererated
   */
  //------------------------------------------------------------------
  private synchronized void resetMail() {
    Debug.out("FaceMailbox.resetMail():  ");
    // Reset the FaceBoard to show nothing
    _FaceBoard.queuePut("RESET");
    _Mailmsgs.clear();
    _FaceBoard.queuePut("NOMAIL");
    //_MailboxMsgCount = -1;   

  }
  //--------------------------------------------------------------------------------
  /**
   * postProcessMail will do a final scan of the mail..
   *
   * <p>
   * This final scan will discover any mail in the collectionof mailmsgs 
   * that no longer actually resides on the server.  
   * It will remove these from MailFaces
   *
   
   *
   *@return void
   */
  //--------------------------------------------------------------------------------
  private synchronized void postProcessMail() {
    Mailmsg mailmsg;
    Mailmsg[] mailmsgs;
    Debug.out("FaceMailbox.postProcessMail: " + count());
    //Debug.out("Count old=" + _OldMsgCount + " new=" + _MailboxMsgCount);
    mailmsgs = getMailmsgs();
    //Debug.out("_Count=" + _Count + " length=" + mailmsgs.length);

    for (int i = 0; i < mailmsgs.length; i++) { // rumage thru the mailbox

      if (_MailFaces.getFlagQuitMail())
        break;
      mailmsg = mailmsgs[i];
      if (mailmsg.getServerStatus() == Mailmsg.StatusUnknown) {
        // This message was not processed on this pass so it must
        // have already been read and deleted.  Remove the icon
        removeMailmsg(mailmsg);
        setModified(true);
      }
    }

    return;
  }

  //---------------------------------------------------------------
  /**
   * addMailmsg will add a new Mailmsg and begin the icon resolution.
   * <pre>
   * 	  Add mailmsg to the collection of mailmsgs
   *    Add mailmsg to FaceReslver Queue
   * </pre>
   *@return void
   */
  //---------------------------------------------------------------

  protected synchronized void addMailmsg(String key, Mailmsg mailmsg) {
    if (Debug.ON)
      Debug.out("FaceMailbox.addMailmsg(): " + key + ":" + mailmsg);
    Log.out("NewMsg", mailmsg.toString());
    _Mailmsgs.put(key, mailmsg);
    mailmsg.setAction(Mailmsg.ActionResolve);
    if (_Resolver != null)
      _Resolver.queuePut(mailmsg); // queue it to the Resolver Thread  

  }

  //---------------------------------------------------------------
  /**
   * removeMailmsg will remove a Mailmsg and tell the GUI to remove the ICON
   *
   *@return void
   */
  //---------------------------------------------------------------

  protected synchronized void removeMailmsg(Mailmsg mailmsg) {
    if (Debug.ON)
      Debug.out("FaceMailbox.removeMailmsg(): " + mailmsg.getKey());
    mailmsg.setAction(Mailmsg.ActionRemove);
    if (_FaceBoard != null) {
      _FaceBoard.queuePut(mailmsg);
    }
    _Mailmsgs.remove(mailmsg.getKey());
  }

  //---------------------------------------------------------------
  /**
   *Method: getMailmsg(String key)
   *@return Mailmsg
   *
   *  Get the Mailmsg with the specified key.
   *  It returns "null" if no Mailmsg with that key
   */
  //---------------------------------------------------------------

  protected synchronized Mailmsg getMailmsg(String key) {
    if (Debug.ON)
      Debug.out("FaceMailbox.getMailmsg(): " + key);
    return (Mailmsg) _Mailmsgs.get(key);
  }

  /**
  ** Get an array of existing Mailmsg objects.
  */
  protected synchronized Mailmsg[] getMailmsgs() {
    Debug.out("FaceMailbox.getMailmsgs()");
    Mailmsg[] mailmsgs = new Mailmsg[_Mailmsgs.size()];
    Enumeration enum = _Mailmsgs.elements();
    for (int i = 0; enum.hasMoreElements(); i++) {
      mailmsgs[i] = (Mailmsg) enum.nextElement();
    }
    return mailmsgs;
  }

  //---------------------------------------------------------------
  /**
   *Method: count()
   *@return int	Count of number of messages in the mailbox
   *  This method will return the number of messages in the mailbox
   */
  //---------------------------------------------------------------

  public int count() {
    return _Mailmsgs.size();
  }

  //-------------------------------------------------------------------------
  /**
   *Method: setModified(boolean) 
   *
   *    Set flag to indicate if the mailbox contains one or more 
   *    new messages
   *
   *@return void
   */
  //-------------------------------------------------------------------------
  public void setModified(boolean modified) {
    _Modified = modified;
  }

  //-------------------------------------------------------------------------
  /**
   *Method: getModified
   *
   *
   *    query whether the mailbox contains one or more 
   *    new messages.
   *
   *@return boolean
   */
  //----------------------------------------------------------
  public boolean getModified() {
    return _Modified;
  }

  //-------------------------------------------------------------------------
  /**
   *Method: isDeleteSupported queries ability to delete Mailmsgs.
   *
   *
   *    query whether the mailbox supports Deleting of Mailmsgs
   *    This will return false.  The Mailbox subclass can override
   *    this to return true. 
   *
   *@return boolean
   */
  //----------------------------------------------------------
  public boolean isDeleteSupported() {
    return false;
  }
 
  //-------------------------------------------------------------------------
  /**
   *Method: isCopySupported queiries ability to copy Mailmsgs.
   *
   *
   *    query whether the mailbox supports Copying of Mailmsgs
   *    This will return false.  The Mailbox subclass can override
   *    this to return true. 
   *
   *@return boolean
   */
  //----------------------------------------------------------
  public boolean isCopySupported() {
    return false;
  }
  //---------------------------------------------------------------
  /**
   *Abstract Method: open().
   *
   *
   *  Sets up the connection to the MailServer
   *  and logs in to the server.  It also sets the
   *  _Count variable to the number of messages waiting
   *
   *@return booolean  Returns true if Open suceeds.
   */
  //---------------------------------------------------------------

  public abstract boolean open();

  //---------------------------------------------------------------
  /**
   *Abstract Method: read().
   *  This is where the majority of the work is done.  The end result is
   *  that the array "_Mailmsgs" contains the instances of Mailmsgs that
   *  are or were in this mailbox with the _Action value set to
   *  indicate if it is "new", "old" or "gone"
   *  This method creates a Mailmsg object for each waiting Mail msg
   *  This Mailmsg object contains the Mail Headers and the size of the msg
   *@return void
   *@see Mailmsg
   */
  //---------------------------------------------------------------

  public abstract void read();

  //---------------------------------------------------------------
  /**
   *Method to delete any mail marked for deletion.
   * <p>
   *  This method is passed an array of mailmsgs where one or more may have
   *  been marked for deletion.  The implementation of this should 
   *  use the mailmsg.getDeleteFlag() to determine what msgs should be 
   * deleted and ensure that the mailmsg is marked for deletion and 
   * actually deleted.  
   *@return boolean  true if one or more messages are deleted.
   *@exception FaceMailboxException
   *@see Mailmsg
   */
  //---------------------------------------------------------------

  public boolean delete(Mailmsg[] mailmsgs) throws FaceMailboxException {
    return false;
  }
  
    //---------------------------------------------------------------
  /**
   *Method to copy mail to a new server.
   * <p>
   *  This method is passed a mailmsg.  The implementation of this should 
   *  make a copy of the mailmsg and send it to the specified server.
   * 
   * @param msg  Mailmsg to be copied.
   * @param address userid@server address where it is to be copied.
   * 
   *@return boolean  true if the message is successfully copied.
   *@exception FaceMailboxException
   *@see Mailmsg
   */
  //---------------------------------------------------------------

  public boolean copymsg(Mailmsg msg, String addr) throws FaceMailboxException {
    return false;
  }

  //---------------------------------------------------------------
  /**
   *Abstract Method: close().
   *
   *  This close method will be called on each pass,  
   *  Depending on the type of Server, we may or may not 
   *  close the connection to the Mail server
   *@return void
   */
  //---------------------------------------------------------------

  public abstract void close();

  //---------------------------------------------------------------
  /**
   *Abstract Method: terminate().
   *
   *  This terminate method will be called when the server is asked to shutdown  
   *@return void
   */
  //---------------------------------------------------------------

  public abstract void terminate();

  //---------------------------------------------------------------
  /**
   *toString is called to identify the Mailbox Object.
   *
   *  
   *@return String that identifies the FaceMailBox object
   */
  //---------------------------------------------------------------

  public String toString() {

    return "FaceMailbox: " + _Server;
  }
  //-------------------------------------------------------
} // end of Class FaceMailbox