package com.jthomas.mailfaces;
/*
 * FaceMailboxJMail.java
 * Copyright (c) 1996 John Thomas  jthomas@cruzio.com
 *      All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for non-commercial purposes
 * is hereby granted provided that this copyright notice
 * appears in all copies.
 *
 */

import java.io.IOException;
import java.util.Enumeration;
import java.util.Properties;
import java.util.Date;
import java.util.Vector;
import java.util.Random;

import javax.mail.Address;
import javax.mail.FetchProfile;
import javax.mail.Flags;
import javax.mail.Folder;
import javax.mail.Header;
import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.Multipart;
import javax.mail.Part;
import javax.mail.Session;
import javax.mail.Store;
import javax.mail.UIDFolder;
import javax.mail.event.MessageCountAdapter;
import javax.mail.event.MessageCountEvent;
import javax.mail.*;
import javax.mail.internet.*;

import com.jthomas.util.Debug;
import com.jthomas.util.Log;
import com.sun.mail.pop3.POP3Folder;
import com.sun.mail.imap.IMAPFolder;
//import com.sun.mail.pop3.POP3PMessage;

/**
 * SubClass of FaceMailbox for a  server that is supported by JavaMail.
 * 
 * This uses the Sun JavaMail product to access an Pop3 or IMAP server.
 *
 * <pre>
 * The following are implementations of 
 * the FaceMailbox abstract methods
 *
 *   void  		open()
 *
 *	 void		read()
 *
 *   void 		close()
 *   
 *   boolean    delete(Mailmsg[] mailmsgs)
 *
 * </pre>
 *
 * @see FaceMailbox
 * @author	John Thomas
 * <a href="mailto:jthomas@cruzio.com">jthomas@cruzio.com</a>
 * @version	1.0 ( Sept 1996 )
 *
 *
 */

public class FaceMailboxJMail extends FaceMailbox {

  private static int _MaxRead = 100000;
  /*
   * State Variables
   *
   */

  /**
  ** This represents an Mail session.
  */
  private Session _Session = null;
  /**
  ** This represents a Message Store
  */
  private Store _Store = null;
  /**
  ** This represents a mail Folder (inbox) 
  */
  private Folder _Folder = null;
  private static final String DEFAULT_FOLDER = "inbox";

  /**
  **  An array of all the current messages
  */
  private Message[] _Messages = null;
  /**
  ** A Hashtable of all the messages
  **   The key is the UID for the message
  */
  //private Hashtable 	_MsgTable	= null;
  /**
  ** True if the the folder is open.
  */
  private boolean _Opened = false;

  /**
  ** Count of messages in previous pass
  */
  private int _OldCount = 0;

  /**
   * set true if one or more messages are deleted.
   */
  private boolean _deleted = false;

  /**
   * _MaxMsg - Max size of message to be displayed.
   */

  private static int _MaxMsg = 8000;

  /**
   * 
   * used to make unique message identifiers
   */

  int _CopyCount = 0;

  /**
   * Static initialization 
   */
  static {
    _ValidTypes.put("imap", "imap");
    _ValidTypes.put("pop3", "pop3");
  }

  //---------------------------------------------------------------
  /**
   * Method: FaceMailboxJMail constructor that will create 
   * a POP3 impementation of the 
   * FaceMailbox abstract object.
   *
   *  This will create the mailbox object and prepare to open it
   * 
   * @param server  reference to the FaceServer instance
   * @param type    Type of Mailbox "IMAP", "POP3"
   */
  //---------------------------------------------------------------

  FaceMailboxJMail(FaceServer server, String type) {

    if (_ValidTypes.get(type.toLowerCase()) != null) {
      super._Type = type.toLowerCase();
    } else {
      Debug.out(0, "FaceMailboxJMail: Invalid mailbox type specified. " + type);
      return;
    }
    _Server = server;
    Debug.out(_Server.toString());

    Debug.out(
      _Type
        + " "
        + _Host
        + " "
        + _User
        + " "
        + (_Password == null ? "null" : "suppressed"));
    try {
      Properties props = System.getProperties();

      // Get a Session object now
      _Session = Session.getDefaultInstance(props, null);
      // session.setDebug(true);
      _MaxMsg = _Options.getInt("MaxMsgSize", _MaxMsg);
      _MaxRead = _Options.getInt("MaxReadMsgSize", _MaxRead);

      Random random = new Random(new Date().getTime());
      _CopyCount = random.nextInt() % 1000;

    } catch (Exception e) {
      Debug.out("FaceMailboxJMail: Exception: ", e);

    }

  }

  //--------------------------------------------------------
  // This constructor is used only for testing
  //-----------------------------------------------------------
  private FaceMailboxJMail(String type, String host, String user, String pass) {

    if (_ValidTypes.get(type.toLowerCase()) != null) {
      super._Type = type.toLowerCase();
    } else {
      Debug.out(0, "FaceMailboxJMail: Invalid mailbox type specified. " + type);
      return;
    }
    super._Host = host;
    super._User = user;
    super._Password = pass;
    try {
      Properties props = System.getProperties();

      // Get a Session object now
      _Session = Session.getDefaultInstance(props, null);
      // session.setDebug(true);
      Debug.out("Session=" + _Session);

    } catch (Exception e) {
      Debug.out("JavaMail Session.getDefaultInstance: ", e);

    }
  }

  //---------------------------------------------------------------
  /**
   * open() will setup the connection to the server and login.
   *
   * <pre>
   *  Sets up the connection to the MailServer
   *  and logs in to the server.  It also sets the
   *  _Count variable to the number of messages waiting
   * </pre>
   *
   *@return boolean - true if Open OK and false if it fails
   *
   */
  //---------------------------------------------------------------

  public boolean open() {

    if (Debug.ON)
      Debug.out("MaceMailboxJPop.open()");
    if (_Store == null) {
      if (Debug.ON)
        Debug.out("FaceMailboxJMail: Get Store object");
      //_MsgTable = new Hashtable(100);
      _Messages = null; // New connection so need new info

      try {
        Properties props = System.getProperties();

        // Get a Session object
        _Session = Session.getDefaultInstance(props, null);
        // session.setDebug(true);

        // Get a Store object		    
        _Store = _Session.getStore(_Type);
      } catch (MessagingException e) {
        Debug.out("FaceMailboxJMail: getStore: ", e);
        return false;
      }
    }
    if (!_Store.isConnected()) {
      if (Debug.ON)
        Debug.out("FaceMailboxJMail: Connect to Store");
      try {

        // Connect
        _Store.connect(_Host, _User, _Password);
        // Open a Folder
        if (_Folder == null) {
          _Folder = _Store.getFolder(DEFAULT_FOLDER);
        }
        if (_Folder == null || !_Folder.exists()) {
          Debug.out(0, "FaceMailboxJMail: Invalid folder");
          return false;
        }
      } catch (MessagingException e) {
        Log.out("MailboxConnectFailure", e.getMessage());
        Debug.out(0, "FaceMailboxJMail: connect failure: " + e.getMessage());
        if (Debug.ON)
          Debug.out(e);
        return false;
      }
    }

    // We are connected - just open the Folder
    try {
      if (!_Folder.isOpen()) {
        _Folder.open(Folder.READ_WRITE);
        _Count = _Folder.getMessageCount();
      }

    } catch (MessagingException e) {
      Debug.out("POP3 open: ", e);
      return false;
    }

    if (Debug.ON)
      Debug.out("FaceMailboxJMail has " + _Count + " messages");

    return _Folder.isOpen();
  }

  //---------------------------------------------------------------
  /**
   *  The read() method creates a Mailmsg object for each new waiting Mail msg.
   *
   * <p>
   * This will get the POP messages from the server and pass them
   * to the checkForNewmsg method which will determine if it is a new
   * message and if so return a Mailmsg object.
   *  This Mailmsg object contains the Mail Headers and the size of the msg 
   *
   * @return void
   */
  //---------------------------------------------------------------

  public void read() {
    Message jmailMsg;
    //long	uid;
    if (Debug.ON)
      Debug.out("FaceMailboxJMail.read()");
    try {
      if (_Messages == null) {
        super.resetMailmsgs();
      }
      _Messages = _Folder.getMessages();
      FetchProfile fp = new FetchProfile();

      fp.add(UIDFolder.FetchProfileItem.UID);
      _Folder.fetch(_Messages, fp);
      //Debug.out("FaceMailboxJMail.read:" + _Messages.length+ " msgs");
      for (int i = 0; i < _Messages.length; i++) {
        jmailMsg = _Messages[i];
        //Debug.out("jmailMsg="+jmailMsg);

        // Check for new msg and convert to Mailmsg
        checkForNewMsg(jmailMsg);
      }
      // Listen for new messages arriving
      this.listen(); // setup to listen for events 

    } catch (Exception e) {
      Debug.out(e);
    }
  } // end read()

  //---------------------------------------------------------------
  /**
   *Method: close()
   *  This is called on each pass
   *
   */
  //-----------------------------------------------------------

  public void close() {
    if (Debug.ON)
      Debug.out("FaceMailboxJMail.close()");
    try {
      if (_Folder.isOpen()) {
        //Debug.out("FaceMailboxJMail: _deleted="+_deleted);
        _Folder.close(_deleted);
      }
      //_Store.close();
    } catch (MessagingException me) {
      Debug.out(me);
    }
    return;

  }
  //---------------------------------------------------------------
  /**
   *Method: terminate()
   *  We are shutting down so Close the connection to the server
   *
   */
  //-----------------------------------------------------------

  public void terminate() {
    Debug.out("FaceMailboxJMail.terminate()");
    try {
      if (_Folder != null && _Folder.isOpen()) {
        _Folder.close(false);
      }
      if (_Store != null)
        _Store.close();
    } catch (MessagingException e) {
      Debug.out("FaceMailboxJMail.terminate(): ", e);
    }
    return;

  }

  //-------------------------------------------------------------------------
  /**
   *Method: isDeleteSupported queries ability to delete Mailmsgs.
   *
   *  Return true to show that we support the delete method.
   *
   *@return boolean
   */
  //----------------------------------------------------------
  public boolean isDeleteSupported() {
    return true;
  }

  //-------------------------------------------------------------------------
  /**
   *Method: isCopySupported queries ability to copy Mailmsgs.
   *
   *
   *    Return true to show that we support the copy method.
   *
   *@return boolean
   */
  //----------------------------------------------------------
  public boolean isCopySupported() {
    return true;
  }

  //---------------------------------------------------------------
  /**
   *Method to copy mail to a new server.
   * <p>
   *  This method is passed a mailmsg.  The implementation of this should 
   *  make a copy of the mailmsg and send it to the specified server.
   * 
   * @param msg  Mailmsg to be copied.
   * @param address userid@server address where it is to be copied.
   * 
   *@return boolean  true if the message is successfully copied.
   *@exception FaceMailboxException
   *@see Mailmsg
   */
  //---------------------------------------------------------------

  private boolean copyMsg(Mailmsg msg, String addr) throws FaceMailboxException {
    boolean returns = false;

    Debug.out("FaceMailboxJMail.copyMsg: " + msg + " to " + addr);
    String host = _Options.get("SendMailHost", "localhost");

    Properties props = new Properties();
    // fill props with any information
    // Setup mail server
    props.put("mail.smtp.host", host);
    //Debug.out(0,"FaceMailboxJMail.copyMsg: host="+host);
    try {
      Session session = Session.getDefaultInstance(props, null);

      // avoid doing a copy of a copied mail message
      Message origMessage = msg.getMessage();
      if (origMessage.getHeader(FaceMailbox.COPY_HEADER) != null) {
        Debug.out( "FaceMailboxJMail.copyMsg: already copied");
        return true;
      }

      //Date sentDate = origMessage.getSentDate();

      // create the new copy of the message
      Message newMessage = new MimeMessage(session);
      // Read the Mail headers and apply to message copy
      //  	
      Enumeration enum = origMessage.getAllHeaders();
      _CopyCount++;
      while (enum.hasMoreElements()) {
        Header header = (Header) enum.nextElement();
        String value = header.getValue();
        // we want to make unique messages.
        if (value.equals(msg.getKey())) {
          value = value + _CopyCount;
        }
        newMessage.setHeader(header.getName(), value);

        Debug.out( header.getName() + " : " + value);

      }

      // set unique header for us
      // Used to prevent recopying the message
      newMessage.setHeader(FaceMailbox.COPY_HEADER, "MailFaces");
      
      // Set message contents
      //
      Object content = null;
      try {
        String ctype = origMessage.getContentType();
        //if (Debug.ON) Debug.out("ContentType="+ctype);
        
        if (ctype != null) {
          content = origMessage.getContent();
          //if (Debug.ON) Debug.out("Content="+content.getClass());
          if (content instanceof Multipart) {
            newMessage.setContent((Multipart) content);
          } else if (content instanceof String) {
            newMessage.setText((String) content);
          } else {

            newMessage.setContent(content, ctype);
          }
        } else {
          content = null;
          Debug.out(0, "FaceMailboxJMail.copyMsg: " + "Unknown content. ");
        }
      } catch (MessagingException me) {
        if (Debug.ON)
          Debug.out(0, me);
        String emsg = me.getMessage();
        Debug.out(0,
          "FaceMailboxJMail.copyMsg: Failure while getting message content. MessagingException - "
            + emsg);
      } catch (IOException ioe) {
        if (Debug.ON)
          Debug.out(0, ioe);
        
        Debug.out(0,
          "FaceMailboxJMail.copyMsg: IO Failure while getting message content. ");
      }
      
      if (content != null) {
        try {
          Address[] addresses = new Address[1];
          addresses[0] = new InternetAddress(addr);
          //newMessage.addRecipient(Message.RecipientType.TO, addresses[0]);
          Transport transport = session.getTransport("smtp");
          transport.connect(host, null, null);

          transport.sendMessage(newMessage, addresses);
          transport.close();
          if (Debug.ON)
            Debug.out("FaceMailboxJMail.copyMsg:  Sent " + msg + " to " + addr);
          if (Log.LogOn)
            Log.out("CopyMsg", "Copy " + msg + " to " + addr);
          returns = true;
        } catch (MessagingException me) {
          if (Debug.ON)
            Debug.out(0, me);
          
          Debug.out(0, "FaceMailboxJMail.copyMsg: Failure while sending Message. " +
                       me.getClass() + " " +me.getMessage());
        }
      }
    } catch (MessagingException me) {
      if (Debug.ON) Debug.out(0,me);
      String emsg = me.getMessage();
      Debug.out(0,"FaceMailboxJMail.copyMsg: MessagingException - "+ emsg);

    } catch (Exception e) {
      Debug.out(0, e);

    }
    if (!returns) { // failed to send
      Debug.out(0,"FaceMailboxJMail.copyMsg:  Failed Sending " + msg + " to " + addr);
      if (Log.LogOn)
        Log.out("CopyMsg", "Failed Sending  " + msg + " to " + addr);
      msg.setDeleteFlag(false); // in case this was a move
    }
    return returns;
  } // copymsg

  //---------------------------------------------------------------
  /**
   *Method: checkForNewMsg()
   * Create new MailMsg Objects
   * <br>
   * Each POP/IMAP message found will be passed to this meathod.
   * The UID will be optained and used as a key in a call to 
   * the getMailmsg method.  If it returns null then this is a 
   * new email.  Otherwise we have already processed the email 
   * on a previous pass.  
   * <br>
   * If it was already process then it will be checked for 
   * having been deleleted and if so, the Delete flag will be 
   * set for the Pop msg so it will be deleted when the folder 
   * is closed.
   * <br>
   * If this is a new email then a new Mailmsg object will be 
   * built mostly from reading the header info for each new message 
   * 
   *
   *@param jmailMsg   - JavaMail Message object	
   *@return new Mailmsg object or null if it is not new.
   *
   */
  //---------------------------------------------------------------

  private Mailmsg checkForNewMsg(Message jmailMsg) {

    String key = null;
    Mailmsg msg = null;
    String uid;

    try {
      if (_Type.equalsIgnoreCase("pop3")) {
        uid = ((POP3Folder) _Folder).getUID(jmailMsg);
      } else {
        //Debug.out("Folder="+ _Folder.getClass()+" "+_Folder.toString());
        UIDFolder uidf = (UIDFolder) _Folder;
        long uidLong = uidf.getUID(jmailMsg);
        uid = String.valueOf(uidLong);
      }

      //if (Debug.ON)
      Debug.out( "FaceMailboxJMail.checkForNewMsg: uid=" + uid);
      int seqnum = jmailMsg.getMessageNumber();
      if (uid != null) {
        key = uid;
        msg = getMailmsg(key);

        if (msg != null) {
          // The message has already been processed on previous pass				
          msg.setServerStatus(Mailmsg.StatusOld);
          if (Debug.ON)
            Debug.out(
              "FaceMailboxJMail.checkForNewMsg: Msg "
                + (seqnum + 1)
                + " key="
                + key
                + " already processed");
          // check if we should copy the message 
          String copyAddress = msg.getCopyAddress();
          if (copyAddress != null) {
            Debug.out(
              "FaceMailboxJMail.checkForNewMsg: Copy " + msg + " to " + copyAddress);
            copyMsg(msg, copyAddress);
            msg.setCopyAddress(null);
          }
          // check if deleted msg.
          if (msg.getDeleteFlag()) {
            FaceStatusLabel.setStatus("Mailbox", "Delete message ");
            if (Debug.ON)
              Debug.out("FaceMailboxJMail.delete: Delete Msg " + key);
            Log.out(
              "DeleteMsg",
              "From: " + msg.getHeader("from") + ";  Subject: " + msg.getHeader("subject"));
            jmailMsg.setFlag(Flags.Flag.DELETED, true);

            _deleted = true;
            // Indicate that the FaceMailboxPop has less messages	
            super.setModified(true);
            super.removeMailmsg(msg);
          }
          msg = null;

        } else {
          // Process a new mail msg 
          msg = buildNewMsg(key, jmailMsg);
        }

      } else { // the uid was null
        Debug.out(
          0,
          "FaceMailboxJMail.checkForNewMessage: Uid is null.  Should not happen");
      }
    } catch (Exception e) {
      Debug.out(e);
      msg = null;
    }
    return msg;

  }

  //---------------------------------------------------------------
  /**
   *Method: buildNewMsg()
   * Create new MailMsg Objects.
   * <br>
   * If this is a new email then a new Mailmsg object will be 
   * built mostly from reading the header info for each new message 
   * 
   *@param key      - Uid that will be used to locate mail object
   *@param jmailMsg   - JavaMail Message object	
   *@return new Mailmsg object or null if it is not new.
   *
   */
  //---------------------------------------------------------------

  private Mailmsg buildNewMsg(String key, Message jmailMsg) {

    Mailmsg msg = null;
    String text = null;
    try {

      int seqnum = jmailMsg.getMessageNumber();
      msg = new Mailmsg(seqnum);
      msg.setSize(jmailMsg.getSize());
      msg.setKey(key);
      msg.setServerStatus(Mailmsg.StatusNew);
      msg.setMessage(jmailMsg);
      // Read the Mail headers
      //  	
      Enumeration enum = jmailMsg.getAllHeaders();

      while (enum.hasMoreElements()) {
        Header header = (Header) enum.nextElement();
        if (Debug.ON)
          Debug.out(header.getName() + " : " + header.getValue());
        msg.putHeader(header.getName(), header.getValue());
      }

      try {
        text = getTextContent(jmailMsg);
      } catch (Exception e) {
        Debug.out(
          0,
          "FaceMailboxJMail.buildNewMessage: Exception: "
            + e.getMessage()
            + " "
            + msg.toString());
        text = "Exception: " + e.getMessage() + " while reading Mail message";
        if (Debug.ON)
          Debug.out(e);
      }
      msg.setText(text);

      // save the pop message
      //_MsgTable.put(key,jmailMsg);
      // Indicate that the FaceMailboxPop has at least 1 new message   		  
      if (Debug.ON)
        Debug.out("FaceMailboxJMail.checkForNewMsg: New msg=" + key);
      super.setModified(true);
      // store the msg object     		
      super.addMailmsg(key, msg);
    } catch (Exception e) {
      Debug.out(e);
      msg = null;
    }
    return msg;

  } // buildNewMsg

  //---------------------------------------------------------------
  /**
   *Method: getTextContents: get the email body text
   * <br>
   *
   * 
   *@param p      - Part object (really the Message object
   
   *@return Text string to display to user.
   *
   */
  //---------------------------------------------------------------
  private String getTextContent(Part p) throws MessagingException, IOException {

    /*
    * Using isMimeType to determine the content type avoids
    * fetching the actual content data until we need it.
    */
    String result = "?";

    if (Debug.ON)
      Debug.out("FaceMailboxJMail.getTextContext: contentType=" + p.getContentType());
    if (p.getSize() > _MaxRead) {
      return "Message is too large to read from Server.";
    }
    if (p.isMimeType("text/plain")) {
      result = (String) p.getContent();

    } else if (p.isMimeType("multipart/*")) {
      //Debug.out(0,"MimeType(multipart/*)");
      Multipart mp = (Multipart) p.getContent();

      if (Debug.ON)
        Debug.out(0, "Multipart mail: " + p.getContentType());

      result = getMultipart(mp);

    } else if (p.isMimeType("text/html")) {
      result = "HTML Formated mail. It can not be displayed";
    } else {
      /*
      * If we actually want to see the data, and it's not a
      * MIME type we know, fetch it and check its Java type.
      */
      Object o = p.getContent();
      if (o instanceof String) {
        result = (String) o;
      } else {
        result = "Unknown mail format";
      }
    }

    if (result.length() > _MaxMsg) {
      result = result.substring(0, _MaxMsg);
    }
    return result;
  }

  //---------------------------------------------------------------
  /**
   *Method: getMultipart: get the text from a multipart content
   * <br>
   *
   * 
   *@param p      - Multipart object 
   
   *@return Text string to display to user.
   *
   */
  //---------------------------------------------------------------
  private String getMultipart(Multipart mp)
    throws MessagingException, IOException {
    String result = "";

    int count = mp.getCount();

    for (int i = 0; i < count; i++) {
      Part p2 = mp.getBodyPart(i);
      if (Debug.ON)
        Debug.out("--> " + p2.getContentType() + "\n");

      if (p2.isMimeType("text/plain")) {
        String filename = p2.getFileName();
        // if not an attachment 
        if (filename == null) {
          result = result + (String) p2.getContent() + "\n";
        } else {
          result = result + "--> " + p2.getContentType() + "\n";
        }
      } else if (p2.isMimeType("multipart/*")) {
        //Debug.out(0,"MimeType(multipart/*)");
        Multipart mp2 = (Multipart) p2.getContent();

        //result = "Multipart mail: " + p2.getContentType() + "\n";
        result = result + getMultipart(mp2);
      } else {
        if (p2.getFileName() != null) {
          result = result + "--> " + p2.getContentType() + "\n";
        }
      }
    }

    return result;
  }

  //------------------------------------------
  /**
  ** setup to listen for events (like new messages arriving)
  * 
  * This only applies for IMap mailboxes.  I am not sure if this really works.
  */
  //-----------------------------------------
  private void listen() {

    _Folder.addMessageCountListener(new MessageCountAdapter() {
      // one or more Messages have been added to Folder
      public void messagesAdded(MessageCountEvent event) {
        Message imsg = null;
        Message[] msgs = event.getMessages();
        Debug.out(0, "FaceMailboxJMail.listen: messagesAdded = " + msgs.length);
        for (int i = 0; i < msgs.length; i++) {
          imsg = msgs[i];
          //Debug.out(imsg);	    
          // Check for new msg and convert to FaceMsg
          checkForNewMsg(imsg);
        }
      }
      // 1 or more Messages expunged 
      public void messagesRemoved(MessageCountEvent event) {
        //Debug.out(event);
        Message imsg = null;
        Message[] msgs = event.getMessages();
        Debug.out(0, "FaceMailboxJMail.listen: messagesRemoved = " + msgs.length);
        for (int i = 0; i < msgs.length; i++) {
          imsg = msgs[i];

          // Delete this msg
          removeMsg(imsg);
        }
      }

    });
  }
  //---------------------------------------------------------------
  /**
  ** This will get invoked when the specified Message
  ** has been expunged from the folder.
  **
  ** Right now I don't know what to do.  I really want to 
  ** get the UID but can't do it yet
  **
  */
  //---------------------------------------------------------------

  private void removeMsg(Message imsg) {
    Debug.out(0, "FaceMailboxJMail.removeMsg: ");

  }

  public void dumpMsg(Message jmailMsg) {

    try {
      /****/
      Address[] addr;
      String[] strings;

      Debug.out("Subject: " + jmailMsg.getSubject());
      strings = jmailMsg.getHeader("subject");
      if (strings != null)
        for (int i = 0; i < strings.length; i++)
          Debug.out("Subject: " + strings[i]);
      strings = jmailMsg.getHeader("Message-ID");
      if (strings != null)
        for (int i = 0; i < strings.length; i++)
          Debug.out("Message-ID: " + strings[i]);

      if ((addr = jmailMsg.getFrom()) != null) {
        for (int i = 0; i < addr.length; i++)
          Debug.out("From:" + addr[i].toString());
      }

      if (jmailMsg.isSet(Flags.Flag.DELETED))
        Debug.out("Message Deleted Flag");
      /*****/
      /**
      if (jmailMsg instanceof POP3Message) {
      	Enumeration enum = ((POP3Message)jmailMsg).getAllHeaderLines();
      	while (enum.hasMoreElements()) 
      		Debug.out((String)enum.nextElement());
      }
      **/

      //		Enumeration enum = jmailMsg.getAllHeaders();
      //		while (enum.hasMoreElements()) {			
      //			Header header = (Header)enum.nextElement();
      //			Debug.out(header.getName()+" : "+header.getValue());
      //		}

    } catch (Exception e) {
      Debug.out(e);
    }

  }

  //------------------------------------------------------------------
  /**
   * Method: main
   *  code to allow independent testing of this class
   * @param arg[]  Command line arguments
   * @return void
   *
   * <pre>
   *  Args:
   *     host
   *     user
   *     password
   *  </pre>
   */
  //------------------------------------------------------------------
  public static void main(String args[]) {
    String type = "pop3";
    String host = "mail.baymoon.com";
    String user = "jthomas";
    String pass = "*";
    Debug.setDebugOn(true); // Get Debug output
    if (args.length < 4) {
      Debug.out("java FaceMailboxJMail type host user password");

    } else {
      type = args[0];
      host = args[1];
      user = args[2];
      pass = args[3];
    }
    try {
      FaceMailboxJMail mailbox = new FaceMailboxJMail(type, host, user, pass);

      mailbox.open();
      mailbox.read();
      //mailbox.close();
      Mailmsg[] msgs = mailbox.getMailmsgs();
      msgs[0].setDeleteFlag(true);
      Debug.out("Delete: " + msgs[0]);
      //mailbox.delete(msgs);
      mailbox.close();
      mailbox.open();
      mailbox.read();
      mailbox.close();
      mailbox.terminate();
    } catch (Throwable t) {
      Debug.out(t);
    }
    System.exit(0);

  }

  //-------------------------------------------------------

}