package com.jthomas.mailfaces;

/*
 * FaceOptions.java
 * Copyright (c) 1996 John Thomas  jthomas@cruzio.com
 *      All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for non-commercial purposes
 * is hereby granted provided that this copyright notice
 * appears in all copies.
 *
 */
import com.jthomas.util.Debug;

import java.io.*;
import java.util.*;

/**
 *
 *
 * Class to handle the various options used by MailFaces.
 * <p>
 *   This current version obtains the options by reading them from a
 *   a specified file.  There is no dialog yet for modifying the options
 *   other than by editing the file with your favorite editor.
 *   However this is on the TO-DO list.
 * <p>
 *   The OptionFile is either specified on the command line;
 *   obtained from the MailFaces Properties or obtained by 
 *   prompting the user.  
 *  
 *   The Options file is read and each line is parsed into a 
 *   key and a value.  The value is then stored in a Hashtable 
 *   key and the String value.  When a subsequent statement is 
 *   found with the same key, then the String  value in the 
 *   Hashtable is replaced with a Vector that consists of previous 
 *   and current values with that key. .
 *
 * <p>
 * 	
 *
 * @author	John Thomas
 * <a href="mailto:jthomas@cruzio.com>jthomas@cruzio.com</a>
 * @version	1.0 ( Sept 1996 )
 *
 *
 * <p><pre>
 * Methods:
 *  Constructor:
 *  
 *      FaceOptions(String filename)
 *			filename is the name of the file that contains the options
 *
 *	The one and only FaceOptions object is created during the
 *  the initialization of the MailFaces object  (Main executable)
 *  Any other objects that need to get an option can use the
 *  Class getRef method.
 *   For Example:
 *  	  FaceOptions opt = FaceOptions.getRef();
 *     String value = opt.get("AnyOpt","mydefault");
 * *
 * *
 *  Public Methods
 *   parseOptionsFile
 *          Reads the Options file and stores options into a Hashtable
 *          Option files can be nested by using the "Include filename" option.
 *
 *   verifyOptions()
 *          Will verify that required parameters are present and tell user
 *          if Option file is bad.
 *
 *   FaceOptions	getRef()  
 *			Returns a reference to the one and only FaceOptions object
 *
 *   put(String optionName, String value);
 *
 *   String	get(String optionName) -> get(optionName,"")
 *   String	get(String optionName,String default);
 *			Returns the String value for the specified Option
 *          If default specified, that will be returned if option not specified.
 *
 *   String[] getList(String optionName);
 *			Returns the array of Strings for the specified Option
 *  
 *   int       getInt(String optionName,int default);
 *
 *   boolean   getBoolean(String optionName,boolean default);        
 * 
 * 
 * </pre>
 */

public class FaceOptions {

  // Constants

  // state variables
  /**
   * reference to the FaceOptions singleton.
   * It is returned by the getRef() static method.
   */
  private static FaceOptions _TheOptions;

  /**
   * HashTable used to store Options
   */
  private Hashtable _Options;
  
  /**
   * Hashtable used to store Vector instances of 
   * Options that are repeated (Select, Group ...)
   */
  private Hashtable _OptionsList;

  /**
   *  Settings for Default options
   * */
  private final String DefaultImageURL = "http://www.cruzio.com/~jthomas/images/q_mark.gif";
  private final String NoMailURL = "http://www.cruzio.com/~jthomas/images/nomail.gif";

  /**
   * File name for Options.
   */
  private String _OptionsFile;

  /** 
   * reference to Object that handles error messages to the user.
   */
  private Prompter _Prompter;


  
  //------------------------------------------------------
  //Method: getRef()
  /**
  ** Static Class Method to access the one and Only FaceOptions object.
  */
  //------------------------------------------------------
  static public FaceOptions getRef() {
    return MailFaces.getRef().getFaceOptionsRef();
  }

  //------------------------------------------------
  /**
  ** FaceOptions Constructor.
  ** 
  ** The specified file name is checked to verify that it exists.
  ** If null or does not exist, it will prompt the user for the file.
  **
  *  If this is the 1st ime, then if faces.opt does not exist 
  *   and distrib.opt does, then rename distribute.opt to faces.opt
  * 
  ** @param FileName for the MailFaces Options file. 
  */
  //------------------------------------------------

  FaceOptions(String optionsfile) throws FaceOptionsException {
    File file = null;
    boolean verified = false;
    String deflt = "faces.opt";
    String distrib = "distribute.opt";
    String title = "Enter MailFaces Options FileName";

    if (_TheOptions != null)
      Debug.out("Creating a new FaceOptions object");

    Debug.out("FaceOptions(" + optionsfile + ")");

    //_OptionsList = new Hashtable(10);

    if (optionsfile == null) {
      _OptionsFile = FaceProperties.getRef().getProperty("OptionsFile");
      Debug.out("OptionsFile=" + optionsfile);
    } else
      _OptionsFile = optionsfile;

    // If the name of the options file is not specified, we would like
    // to ask the user for the name a File 
    while ((_OptionsFile == null || _OptionsFile.length() == 0) || !verified) {
      if (_OptionsFile == null || _OptionsFile.length() == 0) {
        // if required, frename distrib.opt to faces.opt
        File facesOpt = new File(deflt);
        File distribOpt = new File(distrib);
        if ( !facesOpt.exists() && distribOpt.exists()) {
          distribOpt.renameTo(facesOpt);
        }
               
        
        _OptionsFile = prompt("Enter name of Options File:", deflt, title);

        // If user presses cancel or does not enter it, then we quit
        if (_OptionsFile == null || _OptionsFile.length() == 0)
          throw (new FaceOptionsException("Option file not specified"));
      }
      // verify that the file exists.
      try {
        file = new File(_OptionsFile);
        Debug.out("file=" + file);
      } catch (Exception e) {
        Debug.out("Bad Filename", e);
      }

      if (file == null || !file.exists()) {
        deflt = _OptionsFile;
        title = "Options Filename invalid or not found\nPlease Reenter";
        _OptionsFile = null; // try again
      } else
        verified = true;
    }
    // Now save the name of the Options file in the Properties file
    FaceProperties.getRef().put("OptionsFile", _OptionsFile);

    //_TheOptions = this;  comment out the publishing of this

  }

  //------------------------------------------------------
  /**
  ** This method will read and parse the specified Options file.
  ** The values will be saved in a Hashtable.              
  ** If the same option is found more than once, than the 
  ** values will be saved in a vector and a reference to the
  ** vector will be saved in the Hashtable.
  **
  ** @param A null value is passed on the initial call 
  **   and value decided on by the constructor is used. 
  **   On recursive calls due to Include statements, 
  **   the filename for the included options file is specified.  
  */
  //------------------------------------------------------

  public void parseOptions(String filename) throws FaceOptionsException {

    // The following will be reparsed later so we need to 
    // keep the Quoted string quoted.
    String[] reparseKeys = { "SELECT", "SELECTAFTER", "KILLSELECT", "NAMEDSELECT" };
    boolean reparse = false;

    if (Debug.ON) Debug.out("parseOptions(" + filename + ")");
    String parseFile = filename;
    if (parseFile == null) { // For initial entry.
      parseFile = _OptionsFile;
      _Options = new Hashtable(40);
    }
    FileReader filereader = null;
    try {
      filereader = new FileReader(parseFile);
    } catch (FileNotFoundException e) {
      Debug.out("File not found. '" + parseFile + "'");
      throw new FaceOptionsException("File not found");
    }

    // Use the Java Stream Tokenizer to parse the file.
    StreamTokenizer tokenstream = new StreamTokenizer(filereader);
    // Setup the parse table
    tokenstream.slashSlashComments(true); // Honor C++ "//" coments
    tokenstream.slashStarComments(true); // Honor C "/*" comments
    tokenstream.eolIsSignificant(true); // call at End of Line

    // following get around bug that numbers are always parsed
    tokenstream.ordinaryChars('0', '9');
    tokenstream.wordChars('0', '9');

    try {
      String key = null;
      String value = "";
      while (tokenstream.nextToken() != tokenstream.TT_EOF) {

        //Debug.out("nextToken=" + tokenstream.ttype + " : " + tokenstream.sval);
        switch (tokenstream.ttype) {

          // End of Line
          //    If we found a keyword and value on this line then
          //    save them in the hashtable. (Or set entry  into select table)
          case StreamTokenizer.TT_EOL :
            value = value.trim();
            if (key != null) {
              //
              // Uncomment this to display each statement found
              //
              Debug.out("Statement: " + key + " : " + value);
              if (key.equals("INCLUDE"))
                try {
                  Debug.out("Read Include file: " + value);
                  parseOptions(value);

                } catch (Exception e) {
                  Debug.out("Failed parsing Include file " + value);
                  throw new FaceOptionsException("Could not parse Include File\n" + "'" + value + "'");
                } else {
                Object old = _Options.get(key);
                if (old == null) // new value
                  _Options.put(key, value.trim()); // add to hashtable
                else // already found one or more 
                  if (old instanceof Vector) // allready a vector
                     ((Vector) old).addElement(value);
                  else { //Create new vector and add old and new values
                    Vector newVector = new Vector(5);
                    newVector.addElement(old);
                    newVector.addElement(value);
                    _Options.put(key, newVector);
                  }

              }
              key = null;
              value = "";
              reparse = false;
            }
            break;

            // Quoted String or single word.
            // if key not yet found then this must be the key.
            // otherwise it is part of the value.  	

          case StreamTokenizer.TT_WORD :
            //Debug.out("TT_WORD:" + tokenstream.sval);
            if (key == null) {
              key = tokenstream.sval.trim().toUpperCase();
              for (int i = 0; i < reparseKeys.length; i++) {
                if (key.equals(reparseKeys[i])) {
                  reparse = true;
                  break;
                }
              }
            } else // token is part of the value 		
              value = value + " " + tokenstream.sval;
            break;

            // Quoted String
            //   It is all or  part of the value.  	
          case '"' :
            if (reparse)
              value = value + " \"" + tokenstream.sval + "\"";
            else
              value = value + " " + tokenstream.sval;
            break;

            //   Anything else
            //   Just add to the value string that we are building   	
          default :
            {
              //Debug.out("ttype=" + (char)tokenstream.ttype );
              value = value + String.valueOf((char) tokenstream.ttype);
            }
        }
      }

      filereader.close();
    } catch (IOException e) {
      Debug.out("Error while reading Options File. " + parseFile);
      throw new FaceOptionsException("Error reading Options File");
    }
    if (filename == null)
      MailFaces.getRef().fireFaceEvent(FaceEvent.ID_OPTIONS_CHANGED);

  }

  //------------------------------------------------------
  //Method: VerifyOptions()
  /**
  ** Method to verify that all of the required Options are specified
  ** and warn users if their Option file is out of date or invalid.
  ** 
  ** @returns true if no errors are found.
  ** If errors are found, it will display the error messages to the user
  ** and then return false.
  */
  //------------------------------------------------------

  public boolean verifyOptions() {
    String pgm = MailFaces.version();
    String opt = get("Version", "Unknown");
    String msg;
    boolean errs = false;
    // Check that first 2 characters of program version 
    // matches 1st 2 of Option file version
    if (!pgm.regionMatches(true, 0, opt, 0, 2)) {
      msg = "Your Option file version is Version: " + opt + "\n" + "MailFaces version is Version: " + pgm + "' \n \n";
      errs = true;
    } else
      msg = "Your Option file has errors: \n \n";

    if (get("NoMailURL", null) == null) {
      msg = msg + "Add NoMailURL \n";
      put("NoMailURL", NoMailURL);
      errs = true;
    }
    if (get("DefaultImageURL", null) == null) {
      msg = msg + "Add DefaultImageURL \n";
      put("DefaultImageURL", DefaultImageURL);
      errs = true;
    }

    //--------------------------
    // Make sure that we have at least one Mail Server
    //--------------------------
    String[] serverlist = this.getList("MailServer");
    if (serverlist == null) {
      msg = msg + "No Mail Server specified! \n";
      errs = true;
    } else
      for (int i = 0; i < serverlist.length; i++)
        Debug.out(serverlist[i]);

    //--------------------------
    // Make sure that we have at least one PIconDB
    //--------------------------
    String[] picondblist = this.getList("PIconDB");
    if (picondblist == null) {
      msg = msg + "No PIconDB specified! \n";
      errs = true;
    } else
      for (int i = 0; i < picondblist.length; i++)
        Debug.out(picondblist[i]);

    if (errs) {
      Debug.out(msg);
      MsgDialog.create(msg);
      return false;
    } else
      return true;
  }

  //------------------------------------------------------
  //Method: get(String option)  Returns String value
  //		  get(String option, String default)  Returns String value
  /**
  ** Method to access the value of a named Option.
  ** It returns a string that is the value of the object
  ** or an empty string if the option is not found.  
  **
  ** If the Option is coded as "*PROMPT*" then the user
  ** is promptd for the value.  It will return the value
  ** that the user enters.  
  */
  //------------------------------------------------------
  public String get(String option) {

    return this.get(option, ""); // Use the get(option,default) method with an empty default

  }

  //------------------------------------------------------
  /**
  ** Method to access the value of a named Option.
  ** It returns a string that is the value of the object
  ** or the default if the option is not found.  
  **
  ** If the Option is coded as "*PROMPT*" then the user
  ** is prompted for the value.  It will return the value
  ** that the user enters.  
  */

  //------------------------------------------------------
  public String get(String option, String dflt) {
    String value = null;
    Object obj = _Options.get(option.toUpperCase()); // get value from Hashtable
    if (obj == null)
      value = dflt;
    else if (obj instanceof String)
      value = (String) obj;
    else {
      Vector v = (Vector) obj;
      Debug.out("Requested Single value for key=" + option);
      Debug.out("but there are multiple values. Using last value!");
      value = (String) v.lastElement();
    }

    if (value != null && value.toUpperCase().equals("*PROMPT*")) {
      value = prompt("Enter " + option + ":", dflt);
      if (value == null) // if user pressed Cancel button
        value = dflt;
    }

    return value;
  }
  //------------------------------------------------------
  /**
  ** Method to access the value of a named boolean Option.
  ** if the optin is found, it returns true if the value is "true" 
  ** and false otherwise.
  ** If it is not found then it returns the default.  
  */
  //------------------------------------------------------

  public boolean getBoolean(String option, boolean dflt) {

    String value = this.get(option, null);
    if (value == null)
      return dflt; // Return Default
    return (value.toUpperCase().equals("TRUE"));
  }
  //------------------------------------------------------
  /**
  ** Method to access the integer value of a named Option.
  ** It returns a integer value  if found.
  ** Otherwise if it is not found or is invalid, 
  ** it returns the default value 
  */
  //------------------------------------------------------
  public int getInt(String option, int dflt) {

    String value = this.get(option, null);
    if (value == null)
      return dflt; // Return Default
    try {
      return (new Integer(value)).intValue();
    } catch (NumberFormatException nfe) {
      Debug.out("Error Option: " + option + "=" + value + " " + nfe);
      return dflt;
    }

  }

  //------------------------------------------------------
  /**
  ** Method to set the value of a named Option.
  */
  //------------------------------------------------------
  public void put(String option, String value) {
    Debug.out("FaceOptions.put: " + option + " Value:" + value);
    if (value != null)
      _Options.put(option.toUpperCase(), value); // put value into Hashtable
    else
      _Options.remove(option.toUpperCase());
  }

  //------------------------------------------------------
  /**
  ** Method to access the value of a named Option that is a List.
  ** It returns an array of Strings.
  ** It returns an empty array if the option is not found.
  */
  //------------------------------------------------------

  public String[] getList(String option) {
    String[] value = null;
    Object obj = _Options.get(option.toUpperCase()); // get value from Hashtable
    if (obj == null)
      return new String[0];
    else if (obj instanceof String) {
      value = new String[1];
      value[0] = (String) obj;
    } else {
      Vector v = (Vector) obj;
      int n = v.size();
      value = new String[n];
      for (int i = 0; i < n; i++)
        value[i] = (String) v.elementAt(i);
    }

    //Debug.out("FaceOptions.getList: " + option + "=" + (String[])_OptionsList.get(option) );
    return value;
  }

  //------------------------------------------------------
  //Method: prompt()
  /**
  ** Method to prompt the user for the value of the option
  */
  //------------------------------------------------------
  private String prompt(String prompt, String deflt) {
    return prompt(prompt, deflt, "MailFaces Options");
  }

  private String prompt(String prompt, String deflt, String title) {
    int status;
    Prompter p;

    p = Prompter.create();

    p.setTextInputPrompt(prompt, deflt, title);

    p.showPrompt();
    status = p.getStatus();
    if (status == Prompter.STATUS_OK)
      return p.getTextInput();
    else
      return null;

  }


  //-------------------------------------------------------------
  /**
   * Get the name of the Options file
   * 
   * @return File name of the Options file.
   */
  public String  getFileName() {
    
    return _OptionsFile;
  }
  
  
} // end of class FaceOptions

// Define Exception that FaceOptions can throw.
//   FaceOptionsException(String msg)
//
class FaceOptionsException extends RuntimeException {

  FaceOptionsException(String msg) {
    super(msg);
  }

}