package com.jthomas.mailfaces;

/*
 * FaceResolver.java
 * Copyright (c) 1996 John Thomas  jthomas@cruzio.com
 *      All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for non-commercial purposes
 * is hereby granted provided that this copyright notice
 * appears in all copies.
 *
 */

import java.io.*;
import java.util.*;
import com.jthomas.util.*;

/**
 *
 * Class to resolve which image should be displayed for a given mailmsg.
 * 
 * FaceResolver implements Runnable and in the run method, it will do
 * the following:
 * <br>
 *   Get a mailmsg from the Queue  (blocks if queue empty)
 * <br>
 *   Determine the icon that goes with this mailmsg and
 *   fill it into the mailmsg
 * <br>
 *   Queue the mailmsg to the FaceFrame queue
 * <br>
 *   Repeat
 * 
 *   
 *
 * <p>
 
 *
 * @author	John Thomas
 * <a href="mailto:jthomas@cruzio.com>jthomas@cruzio.com</a>
 * @version	1.0 ( Dec 1996 )
 
 */

public class FaceResolver implements Runnable {

  //-------------------------
  // Class variables
  //-------------------------	

  //---------------------------
  // Instance variables
  //---------------------------
  private MailFaces _MailFaces;
  private FaceOptions _Options;

  private FaceCache _Cache = null;

  // FaceSelect Objects that process the MailMsg against the users Select statements
  private FaceSelect _Select;
  private FaceSelect _SelectAfter;
  private FaceSelect _KillSelect;
  private FaceSelect _NamedSelect;

  // PiconDB is the interface to the PIcon Database
  private PiconDB _PiconDB;
  private boolean _PIconDBDomainOnly = false;
  /**
  ** Default icon used when nothing else found
  */
  private String _DefaultImageURL = null;
  /**
  ** Default icon used when KillSelect is matched
  */
  private String _KillImageURL = null;
  /**
  ** Icon used to build an Icon when none is found
  */
  private String _BuildImageURL = null;

  // Define a Queue that FaceMailbox will place Mailmsgs in.
  // The FaceResolver object will get Mailmsgs from the Queue
  // and look up the appropriate Icon
  private Queue _Queue;

  /**
  ** Reference to the FaceBaord so we can access the its queue
  */
  private FaceBoard _FaceBoard = null;

  /**
   * Reference to class that handles SMS messages
   */
  private FaceSMS  _FaceSMS  = null;
  
  /**
  ** Default constructor for FaceResolver.
  **
  ** It will processs the Options, ask to be informed if the 
  ** options change, setup the Queue and remember its own address
  ** in case anybody asks.
  */

  public FaceResolver() {
    Debug.out("FaceResolver()");
    _MailFaces = MailFaces.getRef();
    _Options = _MailFaces.getFaceOptionsRef();
    _FaceSMS = _MailFaces.getFaceSMSRef();
    
    // Process the Options
    refreshOptions();
    // ask to be informed if the Options get changed
    _MailFaces.addFaceEventListener(new FaceEventListener() {
      public void actionPerformed(FaceEvent e) {
        if (e.getID() == FaceEvent.ID_OPTIONS_CHANGED)
          FaceResolver.this.refreshOptions();
      }
    });

    // Create a Queue that we will get work from.
    _Queue = new Queue();

    Debug.out("FaceResolver() done");

  }

  /**
   * queuePut will submit mailmsgs to the FaceResolver Queue.
   *
   * <p>
   * This method is used by other objects to add mailmsgs to the Resolver
   * queue so they can have the icon resolved.
   * @param obj - A Mailmsg object
   */
  public void queuePut(Object obj) {
    _Queue.put(obj);
  }

  /**
   * This method will get a Mailmsg Object from the queue
   * and attempt to determine what Icon to use to represent it.
   * 
   * The mailmsg object may be scheduled for:
   * <br>
   * actionResolve
   *     This will use the Select and Kill lists to try to 
   *     resolve the mailmsg into an icon.  If not resolved then
   *     it is queued to the PiconDB queue. If it is resolved, 
   *     then the mailmsg is queued to the FaceBoard queue.
   * <br>
   * actionResolveAfter
   *     this will use the SelectAfter list for resolution. 
   *     if not resolved it will check if an image was defered 
   *     because of the PiconDefer list. 
   *     If it is resolved, 
   *     then the mailmsg is queued to the FaceBoard queue.
   *@return void  
   *   
   */

  public void run() {

    try {
      Debug.out("FaceResolver.run()");

      // Setup the access to the ICON database servers
      _PiconDB = _MailFaces.getPiconDBRef();
      // access to the FaceBoard Queue
      _FaceBoard = _MailFaces.getFaceBoardRef();
      _Cache = _MailFaces.getCacheRef();

      while (true) {
        Mailmsg mailmsg = (Mailmsg) _Queue.get();
        boolean resolved = false;
        //---------------------------
        // Figure out which image to display for this mail message
        //---------------------------
        if (mailmsg == null) {
        	// treat a null returned by get() as reason to quit
          Debug.out(0,"Null returned from Queue.  Something is broken");	
          break;
        }
        if (mailmsg.getAction() == Mailmsg.ActionResolve) {
          Debug.out("FaceResolver.run() dequeue mailmsg ActionResolve " + mailmsg.getKey());
          resolved = resolve(mailmsg, true, _PIconDBDomainOnly);
          if (resolved) {
            processSelection(mailmsg, true, _PIconDBDomainOnly);
            // Now queue the image to be displayed
            _FaceBoard.queuePut(mailmsg);
          } else {
            // Search PIcon Database by queueing a request for the PiconDB servers
            Debug.out("FaceResolver.run() queue mailmsg " + mailmsg.getKey() + " for PiconDB");
            mailmsg.setAction(Mailmsg.ActionPiconDB);
            _PiconDB.queuePut(mailmsg);
          }
        } else {
          Debug.out("FaceResolver.run() dequeue mailmsg ActionResolveAfter " + mailmsg.getKey());

          resolved = resolveAfter(mailmsg, true, _PIconDBDomainOnly);
          if (resolved) {
            processSelection(mailmsg, true, _PIconDBDomainOnly);
            // Now queue the image to be displayed
            _FaceBoard.queuePut(mailmsg);
          } else {
            Debug.out(0,"FaceResolver: Mailmsg: "+ mailmsg + " UnResolved mail ???????????????? " );
            // If it reaches here, then no image will be 
            // shown for this Mail, but it should never 
            // get here.  
          }
        }
        if (_MailFaces.getFlagQuitMail())
          break;
      } // end while()	
    } catch (Exception e) {
      Debug.out("FaceResolver: Exception in run(); ", e);
    } catch (Error e) {
      Debug.out("FaceResolver: Error in run(); ", e);
    }

    Debug.out(0, "FaceResolver terminating Thread");
    //System.exit(1);    // So we are forced to know about this!! remove at release ????

  }

  //----------------------------------------------
  /**
  ** This will process a mailmsg that has been resolved into an 
  ** icon.  
  **
  ** It will cache the image if it is a URL pointing to 
  ** another system.
  **
  ** If an audio URL was specified, it will queue it to be played.  
  **
  */
  private void processSelection(Mailmsg mailmsg, boolean usecache, boolean domainonly) {
    String imageurl = mailmsg.getImageURL();
    String audiourl = mailmsg.getAudioURL();
    //Debug.out(0,mailmsg);
    Debug.out("Selected: imageurl='" + imageurl + "'  audiourl='" + audiourl + "'");

    // If this is a remote image, then cache it  
    if (_Cache != null && !imageurl.startsWith("file")) {
      String newurl = null;
      if (usecache)
        newurl = _Cache.getURL(imageurl);
      if (newurl == null) // Not already a cached version    		
        newurl = _Cache.cacheImage(imageurl, null, null); // Cache the url
      if (newurl != null) { // if we now have a cached version then use it.
        Debug.out("Use " + newurl + " for " + imageurl);
        imageurl = newurl;
      }
    }

    // If an Audio URL was specified for this mailmsg, then queue it to be played
    if (audiourl != null && audiourl.length() > 0 && _MailFaces.getFlagAudio()) {
      _MailFaces.getFaceAudioRef().queuePut(audiourl);
    }
    
 

  }

  //---------------------------------------
  /**
   *Method: resolve(Mailmsg,useCacheFlag,domainOnlyFlag)
   * will determine what icon should be displayed for the
   * the specified mailmsg.
   * <p>
   * It will search the Select statements and the KillSelect 
   * statements for a match and if found place a pointer to the 
   * image into the Mailmsg object.
   * If it finds a selected icon that also has a audio file specified, it will
   * place a pointer to the audio file into the Mailmsg object also.to be played.
  
   *@param Mailmsg mailmsg - Mailmsg object to to select an icon for
   *@param boolean usecacheflag - true if cache should be checked to resolve object
   *@param boolean domainOnlyFlag - True if only the domain icon is to be selected
   *@return boolean  - true if icon is selected for this mailmsg
   */
  //----------------------------------------
  public boolean resolve(Mailmsg mailmsg, boolean usecache, boolean domainonly) {
    String imageurl = null;
    String audiourl = null;
    FaceSelection select = null;

    
    // Now we determine what image to display for this mail message
    // First we try to select an image based on User Select statements
    select = _Select.selectImage(mailmsg);
    if (select != null) {

      // FaceSelect.select() may have found an SMS indicator 
      // if so we now send an SMS message
      if (_FaceSMS != null && mailmsg.getSMS()) {
        _FaceSMS.sendSMS(mailmsg);      
      }  
    
      imageurl = mailmsg.getImageURL();
      audiourl = mailmsg.getAudioURL();
      

    } else {
      // Check if Mail that we don't like to get
      select = _KillSelect.selectImage(mailmsg);
      if (select != null) {
        mailmsg.setKillCandidate(true);
        imageurl = mailmsg.getImageURL();
        audiourl = null; // No audio for Kill files
        if (imageurl == null)
          imageurl = _KillImageURL;
// we now handle AutoDelete as a normal operand of Select.          
//        else if (imageurl.toUpperCase().equals("AUTODELETE")) {
//          mailmsg.setDeleteFlag(true);
//          Debug.out("Mailmsg will be deleted on next pass");
//          // We will still assign an image (perhaps we should use a different one??????)
//
//          imageurl = _KillImageURL;
//          mailmsg.setImageURL(imageurl);
//          return false;
//        }
      }
    }
    if (imageurl != null) {
      if (imageurl.equalsIgnoreCase("PICONDB")) {
        // This lets us specify that the Select statement wants 
        // us to just get the image from the PICONDB database 
        // and not do any more selection.
        mailmsg.setImageURL(null);
        return false;
      }
      mailmsg.setImageURL(imageurl);
      mailmsg.setAudioURL(audiourl);
      return true;
    } else
      return false;
  }

  //---------------------------------------
  /**
   *Method: resolveAfter(Mailmsg,useCacheFlag,domainOnlyFlag)
   * will determine what icon should be displayed for the
   * the specified mailmsg when it was not found in the 
   * Select, KillSelect or by searching the PiconDB.
   * <p>
   * It will search the SelectAfter 
   * statements for a match and if found place a pointer to the 
   * image into the Mailmsg object.
   * If it finds a selected icon that also has a audio file specified, it will
   * place a pointer to the audio file into the Mailmsg object also.to be played.
  
   *@param Mailmsg mailmsg - Mailmsg object to to select an icon for
   *@param boolean usecacheflag - true if cache should be checked to resolve object
   *@param boolean domainOnlyFlag - True if only the domain icon is to be selected
   *@return boolean  - true if icon is selected for this mailmsg
   */
  //----------------------------------------

  private boolean resolveAfter(Mailmsg mailmsg, boolean usecache, boolean domainonly) {
    String imageurl = null;
    String audiourl = null;

    // If still not found after even trying the PIcon Database
    // then we try to select an image based on User SelectAfter statements

    FaceSelection select = _SelectAfter.selectImage(mailmsg);
    if (select != null) {
      imageurl = mailmsg.getImageURL();
      audiourl = mailmsg.getAudioURL();
      if (_FaceSMS != null && mailmsg.getSMS()) {
        _FaceSMS.sendSMS(mailmsg);
      }      

    }

    if (imageurl == null) {
      // See if PIconDB had a deferred image
      imageurl = mailmsg.getPiconImageURL();
      // If still no luck, check if we want to build an icon
      if (imageurl == null) {
        if (_BuildImageURL != null) {
          String from = mailmsg.getHeader("reply-to");
          if ( from.length() == 0) {
            from = mailmsg.getHeader("from");
          }
          if ( from.length() > 0) {
            imageurl = _BuildImageURL;
            mailmsg.setBuildIcon(true);
          }
        }
      }
      // If still no luck, get the Default URL
      if (imageurl == null)
        imageurl = _DefaultImageURL;
      
    }
    if (audiourl == null || audiourl.length() == 0) // if not specified by Select 
      audiourl = null;
    if (imageurl != null) {
      mailmsg.setImageURL(imageurl);
      mailmsg.setAudioURL(audiourl);
      return true;
    } else
      return false;
  }

  //-----------------------------------------------------
  /**
   *Method: refreshOptions() get control initially and when the 
   * user refreshes the Option file.
   *
   * It will invoke selectSetup to rebuild the Select lists and 
   * also refresh the default URLs used in the rest of the code
   *
   *@return void
   */
  //-----------------------------------------------------

  public void refreshOptions() {
    Debug.out("FaceResolver:refreshOptions()");
    selectSetup();
    _PIconDBDomainOnly = _Options.get("PIconDBDomainOnly", "true").equalsIgnoreCase("TRUE");

    _DefaultImageURL = _Options.get("DefaultImageURL", null);
    _KillImageURL = _Options.get("KillImageURL", null);
    _BuildImageURL = _Options.get("BuildImageURL", null);

    //Debug.out("FaceResolver:refreshOptions() done");
  }

  //--------------------------------------------------------------------
  /**
  ** selectSetup()
  **   Gets the SelectList from the Options file and sets up
  **   to do selection.
  **   We will build 3 lists of selection:
  **   Select and KillSelectfor before the PIconDB search
  **   and SelectAfter for after it. 
  */
  //----------------------------------------------------------------------
  private void selectSetup() {
    
    // we have to do named selects first
    Hashtable namedSelects = FaceSelection.getNamedSelects();
    namedSelects.clear();
    _NamedSelect = new FaceSelect("NamedSelect");
    _NamedSelect.setup();
    
    _Select = new FaceSelect("Select");
    _SelectAfter = new FaceSelect("SelectAfter");
    _KillSelect = new FaceSelect("KillSelect");
    
    _Select.setup();
    _SelectAfter.setup();
    _KillSelect.setup();
  }

  //------------------------------------------------------------------
  /**
   * Method:  checkAddress()
   * @return void
   *   Bring up a dialog box and ask the user to enter
   *   a mail header line.  Then build a mailmsg with
   *   that header line and simulate the processing of
   *   a real mailmsg.
   */
  //------------------------------------------------------------------
  public void checkAddress() {
    //String imageurl;
    //String audiourl;
    //Debug.out("checkAddress():");
    PromptCheck cd = (PromptCheck) PromptCheck.create();
    cd.setCheckDefaults("", false, false);
    cd.showPrompt();

    if (cd.getStatus() == PromptCheck.STATUS_OK) {
      boolean usecache = cd.getUseCache();
      boolean domainonly = cd.getDomainOnly();
      Mailmsg mailmsg = cd.getMailmsg(); // get fake mailmsg
      //if (usecache) {
      //	mailmsg.setAction(Mailmsg.ActionResolve);
      //	queuePut(mailmsg);
      //	return;
      //}
      mailmsg.setAction(Mailmsg.ActionDummy);
      // will associate a mailmsg with an image

      boolean resolved = resolve(mailmsg, usecache, domainonly);
      if (resolved) {
        processSelection(mailmsg, usecache, domainonly);
        _FaceBoard.queuePut(mailmsg);

      } else {
        //---------------------------
        // Figure out which image to display for this mail message
        // by using PiconDB
        //---------------------------	
        resolved = _MailFaces.getPiconDBRef().resolvePiconDB(mailmsg, usecache, domainonly);
        if (resolved) {
          if (mailmsg.getImageURL() != null) {
            processSelection(mailmsg, usecache, domainonly);
            _FaceBoard.queuePut(mailmsg);
          }
        } else {
          resolved = resolveAfter(mailmsg, usecache, domainonly);
          if (resolved) {
            processSelection(mailmsg, usecache, domainonly);
            _FaceBoard.queuePut(mailmsg);
          } else {
            Debug.out("FaceResolver.run(): Unresloved mailmsg " + mailmsg.getKey());
          }
        }
      }

    }
  }

} // End of FaceResolver class