package com.jthomas.mailfaces;
/*
 * FaceSMS.java
 * Copyright (c) 2002 John Thomas  jthomas@cruzio.com
 *      All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for non-commercial purposes
 * is hereby granted provided that this copyright notice
 * appears in all copies.
 *
 */
import java.util.Properties;
import java.util.StringTokenizer;

import javax.mail.Address;
import javax.mail.Message;
import javax.mail.MessagingException;
import javax.mail.Session;
import javax.mail.Transport;
import javax.mail.internet.InternetAddress;
import javax.mail.internet.MimeMessage;

import com.jthomas.util.Debug;
import com.jthomas.util.Log;

/**
 *
 * FaceSMS will handle the sending of an SMS message 
 * (Short Message Service) to the specified cellphone.
 * 
 * The subject and body of the email will be extracted
 * and made into a 160 byte message and emailed to a 
 * server using the original From: header.  
 * The SMS server (from the SMSEmail parameter in the 
 * faces.opt file) is supplied by the cell phone carrier and 
 * will convert the email and send the SMS message.  
 *
 * @author	John Thomas
 * <a href="mailto:jthomas@cruzio.com>jthomas@cruzio.com</a>
 * @version	1.0 ( April 2002) 
 */

public class FaceSMS {


  //-------------------------
  // Class variables
  //-------------------------	

  //---------------------------
  // Instance variables
  //---------------------------
  /**
   * reference to Mailfaces instance
   */
  private MailFaces _MailFaces;
  
  /**
   * reference to the Options 
   */
  private FaceOptions _Options;
  
  /**
   * EMail address that we send SMS messages to.
   * Supplied from the .opt file
   */
  private String _SmsEmail = null; 
  
  /** 
   * Max length of SMS Message.  
   * It is usuall 160 but we can change it 
   * with the SMSMaxLength parameter.
   */
  private int  _SmsMaxLength = 160;
  
  /** 
   * Switch to turn SMS service on and off
   */
  
  private boolean _SmsActive = false;
  
  private Session  _Session = null;
  
  /**
  ** Default constructor for FaceSMS.
  **
  ** It will processs the Options, ask to be informed if the 
  ** options change, .
  */

  public FaceSMS() {
    //Debug.out(0,"FaceSMS()");
    _MailFaces = MailFaces.getRef();
    _Options = _MailFaces.getFaceOptionsRef();
    // Process the Options
    refreshOptions();
    // ask to be informed if the Options get changed
    _MailFaces.addFaceEventListener(new FaceEventListener() {
      public void actionPerformed(FaceEvent e) {
        if (e.getID() == FaceEvent.ID_OPTIONS_CHANGED)
          FaceSMS.this.refreshOptions();
      }
    });

 

  }
  
  
  //-----------------------------------------------------
  /**
   *Method: refreshOptions() get control initially and when the 
   * user refreshes the Option file.
   *
   * It will setup the information needed to send an SMS message
   * The following entries are used from the Options file:
   * <pre>
   *  SMS
   *  SMSEmail
   *  SendMailHost
   *  SMSMaxLength
   * </pre>
   *@return void
   */
  //-----------------------------------------------------

  public void refreshOptions() {
    //Debug.out("FaceSMS:refreshOptions()");
    _SmsActive = _Options.getBoolean("SMS", _SmsActive);
    _SmsEmail = _Options.get("SmsEmail", null);
    String host = _Options.get("SendMailHost", "localhost");
    _SmsMaxLength = _Options.getInt("SmsMaxLength",_SmsMaxLength);
    Debug.out("FaceSMS: email="+_SmsEmail);
    Properties props = new Properties();
    // fill props with any information
    // Setup mail server
    props.put("mail.smtp.host", host);
    _Session = Session.getDefaultInstance(props, null);
  }

  
  //-----------------------------------------------------
  /**
   * Send an SMS message based on the specified Mailmsg. 
   * 
   * We will extract a 160 character or less message 
   * based on the send,subject and text of the specified
   * Mailmsg.
   * <p>
   * @param mailmsg The instance of Mailmsg that we 
   *    will use to compose the SMS message
   *@return void
   */
  //-----------------------------------------------------

  public void sendSMS(Mailmsg  mailmsg) {
    
    if (! getSMSActive() ) {
      // do nothing if SMS option is off
      Debug.out("FaceSMS:sendSMS: inactive. "+mailmsg);
      if (mailmsg.getSMSCopy()) {
        // turn off the copy option also.
        mailmsg.setCopyAddress(null);
      }
      return;
    }
    if ( mailmsg.getHeader(FaceMailbox.COPY_HEADER).length() > 0) {
      Debug.out("FaceSMS: Current message is a copy of original message");
      return;
    }
    Debug.out("FaceSMS:sendSMS: "+mailmsg);
    try {
      int cum = 0;
      MimeMessage message = new MimeMessage(_Session);
      String subject = mailmsg.getHeader("subject");
      
      cum+=subject.length();
      message.setSubject(subject);
      
      String from = mailmsg.getHeader("from");
      Address address = new InternetAddress(from);
      
      message.setFrom(address);
      cum+=address.toString().length();
      int left = _SmsMaxLength - cum;
      
      
      String text = getText(mailmsg,left); 

      message.setText(text);
      Address toAddress = new InternetAddress(_SmsEmail);
      message.addRecipient(Message.RecipientType.TO, toAddress);
       
      Transport.send(message);
      if (Debug.ON) Debug.out("FaceSMS: To: "+_SmsEmail+" Sent "+mailmsg +"\n"+ " body: "+" "+text );  
      if (Log.LogOn)
         Log.out("SMS","To: "+_SmsEmail+ " Sent " + mailmsg);  

    } catch (Exception e) {
      Debug.out(0,"FaceSMS: Exception: "+e.getMessage() +
                   "\n while sending SMS message for "+mailmsg );
      if (Debug.ON) Debug.out(0,e);
    }
  }

  //-----------------------------------------------------
  /**
   * getText will get the text from the original 
   * mailmessage and prepare it tfor the SMS Message.
   * 
   * It will try to be smart and replace all quoted 
   * text by a single "> ..."
   * 
   * @param mailmsg The original error message
   * @param max  The mas length that we can send in SMS
   *
   *@return resulting text message
   */
  //-----------------------------------------------------

  protected String getText(Mailmsg mailmsg, int max) { 
    StringBuffer returns = new StringBuffer(max); 
    String text = mailmsg.getText();
    //Debug.out(0,"Original: "+text);
    boolean quote = false;
    if (text == null) { 
      returns.append("?");
    } else if (text.length() <= max) {
      returns.append(text);
    } else {  
      // try to be smart.
      StringTokenizer st = new StringTokenizer(text,"\n"); 
      while (st.hasMoreTokens()) {
        String line = st.nextToken();
        //Debug.out(0,"Line: "+line);
        if (line.startsWith(">") ) {
          if (quote) { 
            continue;
          } else {
            quote = true;
            line = "> ...";
          }
        } else {
          quote = false;
        }
          
        if (max > line.length()) {
          max -= line.length()+1;
          returns.append(line);
          returns.append("\n");
        } else if (max ==0) {
        	break;
        } else {             
          returns.append(line.substring(0,max-1));
          break;
        }
      }  
    }
    return returns.toString();
  }  // getText() 
  //-----------------------------------------------------
  /**
   * get/set SMS active flag 
   * 
   * The user can activate and deactivate the SMS service
   * 
   *
   *@return void
   */
  //-----------------------------------------------------

  public void setSMSActive(boolean onOff) { 
     
     _SmsActive = onOff;
     _MailFaces.getFaceFrameRef().setSMSActive(onOff);
  }
  
  public boolean getSMSActive() {
    return _SmsActive;
  }
  
//  public static void main(String[] args) {
//  }
}
