package com.jthomas.mailfaces;

/*
 * FaceSelect.java
 * Copyright (c) 1996 John Thomas  jthomas@cruzio.com
 *      All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for non-commercial purposes
 * is hereby granted provided that this copyright notice
 * appears in all copies.
 *
 */

import java.io.*;
import java.util.*;
import starwave.util.regexp.*;

import com.jthomas.util.Debug;
import com.jthomas.util.Log;

/**
 * The FaceSelect class will determine if a specific Mailmsg 
 * matches any of the Select statements from the Options file.
 *
 * There is a FaceSelect object for each of the 3 types. 
 * Select, SelectAfter and KillSelect.
 * During initialization of the FaceSelect object it processes
 * the SelectList built by FaceOptions for builds
 * a FaceSelection object for each valid Select statement.
 *   
 * The FaceResolver class picks the Mailmsg off of the queue and 
 * invokes the selectImage(Mailmsg) method for the Select  and then 
 * KillSelect object.  If these both returned null,
 * it is sent to the PICON queue.  PiconDB may puts it back in the 
 * FaceResolver queue who then calls the selectImage method for the 
 * SelectAfter method.  
 *
 * @see FaceResolve 
 * @see FaceSelection
 * @see Regexp 
 * @see Mailmsg 
 * @see FaceOptions
 * 
 * @author	John Thomas
 * <a href="mailto:jthomas@cruzio.com>jthomas@cruzio.com</a>
 * @version	1.0 ( Oct 1996 )
 * 
 */

public class FaceSelect {

  // state variables

  /**
  * Array of Select statements
  */
  private String[] _SelectList;

  /**
   * array of FaceSelection objects
   */
  private FaceSelection[] _Selections;

  /**
   * Number of Select statements
   */
  private int _SelectCount;

  /**
  ** The Type of Selection being handled by this object.
  **   Select, SelectAfter, KillSelect ...
  */
  private String _Type;

  public static final String KILLSELECT = "KILLSELECT";
  public static final String SELECT = "SELECT";
  public static final String SELECT_AFTER = "SELECTAFTER";
  public static final String NAMED_SELECT = "NAMEDSELECT";
  /**
   * * reference to the object that handle groups of users
   */
  private FaceGroup _FaceGroup = null;
  
  /**
   * reference to last FaceExit instance invoked.
   */
  FaceExit _ExitInstance = null;
  /**
   * Name of last FaceExit invoked.
   */
  String   _ExitName = null;
  /**
  ** Reference to the "main" application object.
  **   Used for many things 
  */
  private MailFaces _MailFaces; // reference to MailFaces global object
  //-----------------------------------------
  // Constructor
  //------------------------------------------

  public FaceSelect(String type) {
    // _Type is either "Select" "KillSelect" or "SelectAfter"
    _Type = type.toUpperCase();
    //Debug.out(0,"FaceSelect: "+_Type);
    _MailFaces = MailFaces.getRef(); // Get Global reference
    _FaceGroup = FaceGroup.getRef();
  }

  //------------------------------------------------------------------
  /**
  ** Initialization routine called after object is constructed and 
  ** when the Options file is reparsed. 
  **
  *   The Select statement has the syntax:<br>
  *   Select headername [RE] match imageURL options 
  *   KillSelect headername [RE] match [autodelete]
  *   SelectAfter headername [RE] match imageURL options
  * 
  *   Options are:
  *      [SMS] 
  *      [AUDIO audioURL]
  *      [EXIT exitvalue]  
  *   
  ** <pre>
  **   - Initialize the FaceSelect object.
  **   - Get array of Select/killselect/selectafter statements 
  *    - For each statenment: 
  **     - call FaceSelection class method to get 
  *        an instance of FaceSelection. 
  **     - store the FaceSelection instances in an array.
  *    
  *  </pre>
  */
  //--------------------------------------------------------------------
  public void setup() {

    // Setup the list of Selection objects
    _SelectList = FaceOptions.getRef().getList(_Type);

    int n = _SelectList.length;
    int k = 0;
    _Selections = new FaceSelection[n];
    for (int i = 0; i < n; i++) {
      String statement = _SelectList[i];
      if (Debug.ON)
        Debug.out("FaceSelect: " + statement);
      FaceSelection selection = FaceSelection.getFaceSelection(_Type, statement);
      //Debug.out(0,selection);
      if (selection.isValid()) {
        _Selections[k] = selection;
        k++;
      }
    }
    _SelectCount = k; // there may be less than size of array       

  } // setup

  //--------------------------------------------------------------------
  /**
  ** selectImage(Mailmsg msg)  returns the matching Selection Object 
  *  or null if it does not match any statement.
  *  
  *  It is possible that more than one FaceSelection objects may match
  *  If one is found that does not specify an image, 
  *  then it continues searching
  * <pre>
  *    Select subject  SMStest  sms
  *    Select from  xxxx  imageurl
  * </pre>
  * @param mailmsg  The mail that is being checked.
  * @returns The FaceSelection object that was matched
  */
  //----------------------------------------------------------------------

  public FaceSelection selectImage(Mailmsg mailmsg) {
    boolean found = false;
    FaceSelection selection = null;
    String addr = null;
    if (Debug.ON)
      Debug.out("FaceSelect.selectImage: " + mailmsg);

    for (int i = 0;(i < _SelectCount) && (!found); i++) {
      selection = _Selections[i];
      if (selection.selectImage(mailmsg)) {
        found = true;
        //
        if (Debug.ON)
          Debug.out("FaceSelect.selectImage: " + selection);
        // check for various processing options
        // that are required by this Selection entry
        if (selection.getAutoDelete()) {
          // This select statement indicate it should be deleted
          mailmsg.setDeleteFlag(true);
          
        }
        //
        // some options to not apply to email that we copied or moved.
        //
        if (mailmsg.getHeader(FaceMailbox.COPY_HEADER).length() == 0) {

          if (selection.getSMS()) {
            // This select statement indicate SMS message needed
            mailmsg.setSMS(true);
          }
          if (selection.getSMSCopy()) {
            mailmsg.setSMSCopy(true);

          }
          if ((addr = selection.getCopyAddress()) != null) {
            mailmsg.setCopyAddress(addr);
            //Debug.out(0,"FaceSelect: copy = "+ _SelectCopy[found]);
            _MailFaces.setFlagCheckMail(true);
          }
          if ((addr = selection.getMoveAddress()) != null) {
            mailmsg.setCopyAddress(addr);
            mailmsg.setDeleteFlag(true);
            _MailFaces.setFlagCheckMail(true);
          }
        }
        mailmsg.setImageURL(selection.getImageURL());
        mailmsg.setAudioURL(selection.getAudioURL());
        //
        // if the selection requests exit processing 
        // then invoke the process method for it.
        //       
        FaceExit exit = selection.getExit();
        //Debug.out("FaceSelect.selectImage: "+exit);
        if (exit != null) {
          try {            
            // invoke the process method                       
            exit.process(mailmsg, selection);            
          } catch (Throwable t) {
            Debug.out(t);            
          }
        }

        if (_Type.equals(SELECT) && mailmsg.getImageURL() == null) {
          // only a successful find if icon indicated.
          found = false;
        }
      }
    }
    if (found) {
      return selection;
    } else {
      return null;
    }

  }
} // End of FaceSelect Class