package com.jthomas.mailfaces;

/*
 * FaceSelection.java
 * Copyright (c) 1996 John Thomas  jthomas@cruzio.com
 *      All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for non-commercial purposes
 * is hereby granted provided that this copyright notice
 * appears in all copies.
 *
 */

import java.io.*;
import java.util.*;
import starwave.util.regexp.*;

import com.jthomas.util.Debug;
import com.jthomas.util.Log;

/**
 * The FaceSelection represents the Select/KillSelect/SelectAfter statements 
 * that are read from the option file.
 * 
 *
 * @see FaceSelect 
 * @see FaceOptions
 * 
 * @author	John Thomas
 * <a href="mailto:jthomas@cruzio.com>jthomas@cruzio.com</a>
 * @version	1.0 ( Oct 1996 )
 * 
 */

public class FaceSelection {

  /**
   *  Special value in select statement to indicate that we want 
   *  to check if the e-mail is not addressed specifically to me.
   *  This is often an indicator of SPAM 
   *  (but also mailing lists or bcc: )
   */
  private static final String NOT_TO_ME = "NotTo";

  // state variables

  /**
   * Original Select statement from options file.  
   * 
   * 
   */
  private String _SelectStatement;

  /**
   * set to false if Selection entry has errors.
   */
  private boolean _SelectValid = true;
  /**
   * Regular Expression from Select statement.  
   * 
   * Contains null if Select was a matching expression.
   */
  private Regexp _SelectRegexp;

  /**
   *  Matching Expression.  
   * Contains null if Select was for a Regular Expression.
   * The String has been converted to lowercase and will be
   * matched against the header after being converted to lowercase.
   */
  private String _SelectMatch;

  /**
   * header identifier that this checks against.
   */
  private String _SelectHeader;
  /**
   * URL for the icon 
   */
  private String _SelectImageURL;
  /**
   * Audio identifiers
   * 
   */
  private String _SelectAudioURL;

  /**
   * AutoDeleteoption     
   * true if Selected email should be deleted  
   */
  private boolean _SelectAutoDelete;
  
  /**
   * SMS option     
   * true if ShortMessageService should be invoked 
   */
  private boolean _SelectSMS;

  /**
   * SMSCopy option     
   * true if ShortMessageService should be invoked and email copied 
   */
  private boolean _SelectSMSCopy;
  /**
  * Copy options. 
  * 
  * Address the message should be copied to 
  */
  private String _SelectCopy;

  /**
  * Move option . 
  * 
  * Address the message should be copied to and then deleted 
  */
  private String _SelectMove;

  /**
   * Exit option  
   * 
   * If Exit(class) is found then this is set to 
   * instance of classname instead of null
   * 
   */
  private FaceExit  _ExitInstance = null;
  
  /** 
   * negative search argument if true
   */
  private boolean _Negative = false;

  /**
  ** The Type of Selection being handled by this object.
  **   Select, SelectAfter, KillSelect ...
  */
  private String _Type;
  
  public static final String KILLSELECT = "KILLSELECT";
  public static final String SELECT = "SELECT";
  public static final String SELECT_AFTER = "SELECTAFTER";
  public static final String NAMED_SELECT = "NAMEDSELECT";
   /**
   * * reference to the object that handle groups of users
   */
  private static FaceGroup _FaceGroup = null;
  
  /**
  ** Reference to the "main" application object.
  **   Used for many things 
  */
  private MailFaces _MailFaces; // reference to MailFaces global object
  
  /**
   * reference to the FaceOptions object
  */
  
  private static  FaceOptions _Options = null;
  
  /**
   * Name asigned to the NamedSelect instance.
   * NamedSelect  A1 subject Test1
   * NamedSelect  A2 subject Test2 
   * Select All A1 A2   
   */
  private String  _Name = null;
  
  /**
   * Hashtable of all the NamedSelect statements
   */
  private static Hashtable _NamedSelects = new Hashtable(32);
  
  /**
   * array of internal FaceSelection objects that 
   * will be used to satisfy the All( ) or Any( ) 
   * selection criteria.
   */
  private FaceSelection[] _Selections = null;
   
  /**
   * set true for Selcet All(name1 name2 ...)
   */ 
  private boolean _All = false;
  /**
   * set true for Selcet Any(name1 name2 ...)
   */ 
  private boolean _Any = false;
  //-----------------------------------------
  // private constructor
  //------------------------------------------
  /**
   * private constructor. 
   * 
   * Why do we use a static class method to build FaceSelection objects?
   * I must have had a reason at one time??????
   */
  private  FaceSelection() { 
  } 
  
  //-----------------------------------------
  // static class object to return a FaceSelection Object
  //------------------------------------------
  /**
  ** static class object to return a FaceSelection Object. 
  **
  *   The Select statement has the syntax:<br>
  * <pre>
  *   Select [not] headername  match imageURL options 
  *   Select [not] headername re(regexp) [imageurl] [options]
  *   KillSelect [not] headername [RE] match [autodelete]
  *   SelectAfter [not] headername [RE] match imageURL options
  *   
  *   The "not" modifier makes it a negative test and
  *   could also be coded as "!"
  * 
  *   Options are:
  *      SMS 
  *      AUDIO(audioURL)
  *      EXIT[(exitvalue)]
  *      copy[(copyaddress)]
  *      move[(moveaddress)]  
  *      autodelete
  * 
  ** <pre>
  **   - Initialize the FaceSelection object.
  *  : 
  **     - parse the statement 
  **     - store the statement options
  *    
  *  </pre>
  */
  //--------------------------------------------------------------------
  public static FaceSelection getFaceSelection(String type,String statement) {
    // _Type is either "NamedSelect", "Select", "KillSelect" or "SelectAfter"
    
    //Debug.out(0,"FaceSelection.getFaceSelection: "+statement);
    // get any options from the .opt file
    if (_Options == null) 
      _Options = FaceOptions.getRef();
    if (_FaceGroup == null)  
      _FaceGroup = FaceGroup.getRef();
    FaceSelection faceSelection =  new FaceSelection();
    faceSelection.init(type,statement);
    if (faceSelection.isValid()) {
      Debug.out("FaceSelection.getFaceSelection: created "+faceSelection);
      if (type.equals(NAMED_SELECT) ) {
        String name = faceSelection.getName();
        if (_NamedSelects.get(name) != null) {
          /// do something
          //Debug.out("FaceSelection.duplicate error");
          String errorMsg = "Select: " + statement + "\n Name: " + name+ " duplicate found";
          MsgDialog.create(errorMsg);
          faceSelection.setValid(false);
        } else {
          _NamedSelects.put(name,faceSelection);
        }
      } 
    }
    return faceSelection;   
    //_MailFaces = MailFaces.getRef(); // Get Global reference
    //_FaceGroup = FaceGroup.getRef();
  
  
  }
  
  //--------------------------------------------------------------------
  /**
  * access the Hashtable of NamedSelect objects.
  ** 
  */
  //--------------------------------------------------------------------
  static Hashtable getNamedSelects() {
    return  _NamedSelects;
  }
  //--------------------------------------------------------------------
  /**
  * init method for the FaceSelection object.
  ** <pre>
  **   - Initialize the FaceSelection object.
  *  
  **     - parse the statement 
  **     - store the statement options
  *    
  *  </pre>
  */
  //--------------------------------------------------------------------
  private void init(String type, String statement) {
    String errorMsg = null;
    String regexp = null;
    String arg = null;
    String token = null;
    int next = -1;

    _Type = type.toUpperCase();
    _SelectStatement = statement;
    if (Debug.ON)
      Debug.out("FaceSelection: " + statement);
    String[] tokens = parseStatement(statement);
    if (Debug.ON) {
      StringBuffer sb = new StringBuffer("FaceSelection: ");
      for (int di = 0; di < tokens.length; di++) {
        sb.append(di + ": " + tokens[di] + " ");
      }
      Debug.out(0, sb.toString());
    }
    //
    // if parseStatements returns token[0] == null then
    // it found and error and reported it.  
    // we just ignore it and continue.  
    if (tokens[0] == null) {
      Debug.out(0,"FaceSelection: Invalid: " + statement);
      _SelectValid = false;
      return;
    }

    // handle a NamedSelect statement.
    if (_Type.equals(NAMED_SELECT)) {
      _Name = tokens[0];
      removeToken(tokens, 0);
    }

    // check for "not" and remove it if found
    if (tokens[0].equals("not") ) {
      _Negative = true;
      removeToken(tokens, 0);
    }

    String[] args = getArg(tokens[0]);
    token = args[0].toLowerCase();
    arg = args[1];
    // check for all/any(name1 name2 ...)
    if (token.equals("any") || token.equals("all")) {
      processMulti(token,arg);      
      next = 1;
    } else {
      _SelectHeader = token;
      // check for deprecated statement
      if (_SelectHeader.equalsIgnoreCase(NOT_TO_ME)) {
        Debug.out(0, "Deprecated statement: use \"not recipient\"");
        Debug.out(0, "-> " + statement);
        _Negative = true;
        _SelectHeader = "recipient";
      }
      // 1st token is either a regular expression or match string
      
      next = 2;
      args = getArg(tokens[1]);
      token = args[0];
      arg = args[1];
      if (token.equalsIgnoreCase("re")) { // check for Regular expression
        if (arg != null) {
          regexp = arg;
        } else {
          regexp = tokens[2];
          next++;
        }
      } else {
        regexp = null;
        _SelectMatch = token.toLowerCase();
      }      
    }

    //
    // process the remaining options.
    for (int j = next; j < tokens.length; j++) {
      token = tokens[j];

      if (token == null) // done processing 
        break;
      // check if this token has a parameter 
      //   token(arg)  
      args = getArg(token);
      token = args[0];
      arg = args[1];
      // check for special tokens

      if (token.equalsIgnoreCase("autodelete")) {
        _SelectAutoDelete = true;
      } else if (token.equalsIgnoreCase("sms")) {
        _SelectSMS = true;  
      } else if (token.equalsIgnoreCase("smscopy")) {
        _SelectSMS = true;
        if (arg != null) {
          _SelectCopy = arg;
        } else {
          _SelectCopy = _Options.get("DefaultCopyAddress", null);
        }
      } else if (token.equalsIgnoreCase("exit")) {
        String exit = null;
        if (arg != null) {
          exit = arg;
        } else {
          exit = _Options.get("DefaultExit", null);
        }
       
        if (exit != null) {
          try {
            // create an instance of the exit
           
            _ExitInstance = (FaceExit) (Class.forName(exit)).newInstance();
            
               

          } catch (Throwable t) {
            Debug.out(t);
            // disable the exit
            
            _ExitInstance = null;
          }
        }
      } else if (token.equalsIgnoreCase("copy")) {
        if (arg != null) {
          _SelectCopy = arg;
        } else {
          _SelectCopy = _Options.get("DefaultCopyAddress", null);
        }

      } else if (token.equalsIgnoreCase("move")) {
        if (arg != null) {
          _SelectMove = arg;
        } else {
          _SelectMove = _Options.get("DefaultMoveAddress", null);
        }

      } else if (token.equalsIgnoreCase("audio")) {

        if (arg != null) {
          _SelectAudioURL = arg;
        } else {
          _SelectAudioURL = tokens[j + 1];
          j++;
        }

        // handle positional tokens  
      } else if (j == 1) {
        _SelectImageURL = tokens[1];
      } else if (j == 2) {
        if (_SelectImageURL == null) {
          _SelectImageURL = tokens[2];
        } else {
          _SelectAudioURL = tokens[2];
        }
      } else if (j == 3) {
        _SelectAudioURL = tokens[3];
      } else {
        errorMsg = "Select: " + statement + "\n" + "Invalid operand: \"" + token + "\" ";

        break;
      }
    } // for j < tokens.length

    if (regexp != null) { // then compile the regular expresion.
      try {
        _SelectRegexp = Regexp.compile(regexp);
      } catch (MalformedRegexpException e) {
        _SelectRegexp = null;
      }

      if (_SelectRegexp == null) {
        errorMsg = "Select: " + statement + "\n" + "Error in Regular Expression. \"" + regexp + "\" ";
      }

    }
    // some special cases
    if (_SelectSMSCopy && (_SelectCopy == null)) {
      // if smscopy specified but no copy addr
      // then use the default copy address.
      _SelectCopy = _Options.get("DefaultCopyAddress", null);
    }
  
    if (errorMsg != null) {
      MsgDialog.create(errorMsg);
      _SelectValid = false;
      // statement gets ignored if error.
    }
    //Debug.out(0,"FaceSelection.getFaceSelection: "+this);
  }
  
  
  
  
  /**
  ** Method to parse a token to extract the argument if any. 
  **
  ** @param token string in format of "operand(argument)"
  ** @return  "argument"
  *   the input token is changed to be "operand"  
  */  
  private String[] getArg(String token) {
    String[] args = new String[2];
    int paren = token.indexOf("(");
    if (paren > 0) {
      //Debug.out(0,"token="+token);
      args[1] = token.substring(paren + 1, token.length() - 1);
      args[0] = token.substring(0, paren);
      //Debug.out(0,"token="+token+" arg="+arg);
    } else {
      args[0] = token;
      args[1] = null;
    }
    return args;
  }
  
   
  
  
  /**
  * Method to handle the any(name1 name2 ..) or
  * or all(  name1 name2 ..)
  * 
  ** @param token Either "any" or "all"
  *  @param arg  String of names
  ** @return  void
  *   
  */  
  private void processMulti(String token, String arg)  {
    //Debug.out(0,"FaceSelections.processMulti: "+token + " : "+arg);
    if (token.equals("any")) {
      _Any = true;
    } else {
      _All = true;
    }
    Hashtable table = FaceSelection.getNamedSelects();
    StringTokenizer st = new StringTokenizer(arg);
    int n = st.countTokens();
    String name;
    _Selections = new FaceSelection[n];
    for (int i = 0; st.hasMoreTokens(); i++) {
      name = st.nextToken();
      _Selections[i] = (FaceSelection)table.get(name);
      if (_Selections[i] == null) {
        String errorMsg = "Select: " + _SelectStatement + "\n Name: " + name+ " not found";
        MsgDialog.create(errorMsg);
        this.setValid(false);
      }
      
    }
//    for (int i = 0; i < n; i++) {
//      Debug.out(0,"FaceSelection: "+i+" "+_Selections[i]);
//      if (_Selections[i] == null) {
//        Debug.out(0,"FaceSelection: "+arg+ " name not found");
//      }
//    }
  }
      
  /**
  ** Method to parse the Select statement. 
  **
  ** @param statement The String that was read from the file.
  ** @return  Array of tokens The last token is null.
  */
  private String[] parseStatement(String statement) {

    int MAX = 10;
    int n = 0;
    //Debug.out(0,"FaceSelection.parseStatement: "+statement);
    String[] tokens = new String[MAX];
    StringReader sr = new StringReader(statement);
    StreamTokenizer tokenstream = new StreamTokenizer(sr);
    tokenstream.quoteChar('"');
    tokenstream.slashSlashComments(false); // allow //
    tokenstream.slashStarComments(false); // allow /* */  

    try {

      //StringBuffer value = new StringBuffer();
      labelWhile : while (tokenstream.nextToken() != tokenstream.TT_EOF) {
        if (n >= MAX) {
          Debug.out(0, "Too many tokens. " + statement);
          String errorMsg = "Error parsing Statement: " + statement + "\n" + "Too many parameters";
          MsgDialog.create(errorMsg);
          tokens[0] = null;
          break;
        }
        switch (tokenstream.ttype) {
          // unquoted word
          case StreamTokenizer.TT_WORD :
            tokens[n] = tokenstream.sval.trim();
            //Debug.out(n + " - " + tokenstream.sval);
            n++;
            break;

            // Quoted String
            //   sval contains the text inside " "  	
          case '"' :
            tokens[n] = tokenstream.sval.trim();
            //Debug.out(n + " - " + tokenstream.sval);
            n++;
            break;

            //   Anything else

          default :

            String errorMsg = "Error parsing Statement: " + statement + "\n" + "Found unquoted char: " + (char) tokenstream.ttype;
            Debug.printStackTrace(errorMsg);
            MsgDialog.create(errorMsg);
            tokens[0] = null;
            break labelWhile;

        } // end case

      } // end while

    } catch (IOException e) {
      Debug.out(0, "Error parsing Statement: " + statement);
      Debug.out(e);
      String errorMsg = "Error parsing Statement: " + statement + "\n" + "IOException " + e.getMessage();
      MsgDialog.create(errorMsg);
      tokens[0] = null;
    }

    return tokens;
  }
  
  //--------------------------------------------------------------------
  /**
   *  remove a token from an array 
   */
  //--------------------------------------------------------------------
  private void removeToken(String[] tokens, int num) {      
      for (int j = 0; j < tokens.length-1; j++) {
        tokens[j] = tokens[j+1];
      }
      tokens[tokens.length-1] = null;
  }
  
  //--------------------------------------------------------------------
  /**
  ** selectImage(Mailmsg msg)  returns an index to matching Select statement 
  *  or -1 if it does not match any statement.
  * <pre>
  * There are many possible types of Select statements.
  *    recipient  that will match is the to: or cc: headers  
  *           match the specified id
  *    sender      that will match based on from: and reply-to headers
  *    
  *    group  that matches if a call to FaceGroup.resolve() returns true
  *    
  *    header  that matches if the specified header line matches either the 
  *            specified string or the specified Regular Expression.
  *    
  *    any(name1 name2 ...) 
  *    all(name1 name2 ...)
  *            These will cause the names select statements to be tested.
  *  
  *    The "not" operand will make reverse the match
  * 
  * 
  * </pre> 
  * @param mailmsg  The mail that is being checked.
  * @returns index to the Select statement that it matched
  */
  //----------------------------------------------------------------------
  
  public boolean selectImage(Mailmsg mailmsg) {
    String buffer = null;
    String header = null;
    String matchMsg = null; // inserted in log message
    boolean found = false;

    //if (Debug.ON) Debug.out("FaceSelection.selectImage: " + mailmsg);
    
    if ( _Any) { 
      for(int i=0; i<_Selections.length;i++) {
        if (_Selections[i].selectImage(mailmsg) ) {
          found = true;
          break;
        }
      }
    } else if ( _All) { 
      found = true;
      for(int i=0; i<_Selections.length;i++) {
        if ( ! _Selections[i].selectImage(mailmsg) ) {
          found = false;
          break;
        }
      }
    } else if ( _SelectHeader.equalsIgnoreCase("recipient") ) {

      boolean recipient = false;

      // special cases to check for mail not addressesed to me
      // if my mail address is not in to: or cc: header then 
      // it is not to me and is SPAM, mailing list or bcc:

      if (matchHeader(mailmsg, "to")) {
        recipient = true;
      } else {

        if (matchHeader(mailmsg, "cc")) {
          recipient = true;
        }
      }
      matchMsg = " Recipient ";

      found = recipient;      
      
    } else if (_SelectHeader.equalsIgnoreCase("sender")) {

      boolean sender = false;
      // special cases to check for mail with from: or reply-to matching    

      if (matchHeader(mailmsg, "from")) {
        sender = true;
      } else if (matchHeader(mailmsg, "reply-to")) {
          sender = true;
      } else if (matchHeader(mailmsg, "sender")) {
          sender = true;
      } 

      found = sender;
      matchMsg = " Sender ";

    } else if (_SelectHeader.equalsIgnoreCase("group")) {
      // process group statement

      String group = _SelectMatch;

      if (_FaceGroup.resolve(group, mailmsg)) {

        found = true;
        matchMsg = " Group Member ";

      }
    } else {
      header = _SelectHeader;
      buffer = mailmsg.getHeader(header);

      //Debug.out("buffer=" + buffer);
      if (buffer.length() > 0) { // If we have this mail header				
        if (match(buffer)) {
          if (Debug.ON)
            Debug.out(0, "FaceSelection: Matched header " + _SelectHeader + ":" + buffer);

          found = true;
          matchMsg = " header " + header + ":" + buffer;

        }
      }
    }
    
    if (_Negative) {
      found = ! found;
      matchMsg = " Not "+matchMsg;
    }
    if (found) {
      //
      // display a good indicator of why this mailmsg matched
      //
      String subject = mailmsg.getHeader("subject");
      if (Debug.ON)
        Debug.out(0, "FaceSelection: Matched " + _SelectStatement + ":" + matchMsg + " subject: " + subject);
      if (Log.LogOn)
        Log.out("Select", " Matched " + _SelectStatement + ":" + matchMsg + " subject: " + subject);


    }

    return found;

  }


  //--------------------------------------------------------------------
  /**
   * Match the buffer against the specified entry
   * It may be a regular expression or 
   * a string comparison.
   * 
   * @param buffer  The string to be matched to 
   * 
   * @return true if expression matches the buffer.
   */
  //--------------------------------------------------------------------
  public boolean match(String buffer) {
    if (_SelectRegexp != null) {
      return matchRE(buffer, _SelectRegexp);
    } else {
      String match = _SelectMatch;
      // An empty string will match any header that actually exists
      if (match.length() == 0 && buffer.length() > 0) {
        //Debug.out(0,"match="+match+";buffer="+buffer); 
        return true;
      }
      int found = buffer.toLowerCase().indexOf(match);
      //Debug.out(0,"match="+match+";buffer="+buffer+";found="+found);    
      return found >= 0;
    }
  }

  //--------------------------------------------------------------------
  /**
   * Match the buffer against the specified header. 
   * It may be a regular expression or 
   * a string comparison.
   * 
   * @param header  The name of the headerto match against 
   * @param entry   The entry to compare
   * @return true if expression matches the header,
   *     false if not matched or header does not exist
   */
  //--------------------------------------------------------------------
  public boolean matchHeader(Mailmsg mailmsg,String header) {
    String buffer = mailmsg.getHeader(header);
    if ( buffer.length() == 0)  // header does not exist
      return false;
    if (_SelectRegexp != null) {
      return matchRE(buffer, _SelectRegexp);
    } else {
      String match = _SelectMatch;
      // An empty string will match any header that actually exists
      if (match.length() == 0 ) {
        //Debug.out(0,"match="+match+";buffer="+buffer); 
        return true;
      }
      int found = buffer.toLowerCase().indexOf(match);
      //Debug.out(0,"match="+match+";buffer="+buffer+";found="+found);    
      return found >= 0;
    }
  }

  //--------------------------------------------------------------------
  /**
   * Do a regular Expression match.
   * 
   * @param buffer  The string to be matched to 
   * @param re      The Regualr expression
   * @return true if regular expression matches the string.
   */
  //--------------------------------------------------------------------
  boolean matchRE(String buffer, Regexp re) {
    return re.searchForward(buffer, 0) != null;
  }

  /**
  ** method used to get the ImageURL for the matching Select statement.
  **
  ** The selectImage method returns an index to the matching statement
  ** if it finds a match.  This method is used to get the imageURL for
  ** from the Select statement at that index.
  * 
  */

  public String getImageURL() {
    return _SelectImageURL;
  }

  /**
  ** method used to get the AudioURL for the matching Select statement.
  **
  ** The selectImage method returns an index to the matching statement
  ** if it finds a match.  This method is used to get the audio URL for
  ** from the Select statement at that index.
  *
  */

  public String getAudioURL() {
    return _SelectAudioURL;
  }

  /**
  ** method used to get the AutoDelete flag.
  **
  * 
  */

  public boolean getAutoDelete() {
    return _SelectAutoDelete;
  }
  
  /**
  ** method used to get the SMS message flag.
  **
  * 
  */

  public boolean getSMS() {
    return _SelectSMS;
  }

  /**
  ** method used to get the SMSCopy flag.
  **
  * 
  */

  public boolean getSMSCopy() {
    return _SelectSMSCopy;
  }

  /**
  ** method used to get the copy address.
  **
  * @return copy email address.
  */

  public String getCopyAddress() {
    return _SelectCopy;
  }

  /**
  ** method used to get the move address.
  **
  * @return move email address.
  */

  public String getMoveAddress() {
    return _SelectMove;
  }

  /**
  ** method used to get the Exit classname.
  **
  * @return classname.
  */

  public FaceExit getExit() {
    return _ExitInstance;
  }
  
  /**
  ** method used to get the Name assigned to the Select statement.
  **
  * @return name.
  */

  public String getName() {
    return _Name;
  }
  
  /**
  ** method used to get the SelectValid flag.
  **
  * 
  */

  public boolean isValid() {
    return _SelectValid;
  }
  /**
  ** method used to set the SelectValid flag.
  **
  * 
  */

  public void setValid(boolean onOff) {
    _SelectValid = onOff;
  }
  /**
   * toString display method 
   */
  public String toString() {
    
    StringBuffer sb = new StringBuffer(100);
    
    sb.append("FaceSelection:");
    if (_Name != null) {
      sb.append(" Name: "+_Name+" ");
    }
    if (_Negative) 
      sb.append(" Not ");
    if (_Any ) { 
      sb.append(" Any() \n");
      for(int i=0; i<_Selections.length;i++) {
        sb.append("  "+_Selections[i] +"\n");          
      }
      
    } else  if (_All ) { 
      sb.append(" All() \n");
      for(int i=0; i<_Selections.length;i++) {
        sb.append("  "+_Selections[i] +"\n");          
      } 
    } else { 
      sb.append(_SelectHeader);
      sb.append(" ");
      sb.append(_SelectMatch!=null?_SelectMatch:_SelectRegexp.toString());
    }
    if (_SelectImageURL != null) {
      sb.append( " Image  "+_SelectImageURL);
    }
    if (_SelectAudioURL != null) {
      sb.append( " Audio  "+_SelectAudioURL);
    }
    if (! isValid()) {
      sb.append( " Invalid ");
    }
    if (_SelectSMS) {
      sb.append( " SMS ");
    }
    if (_SelectSMSCopy ) {
      sb.append( " SMSCopy  ");
    }
    if (_SelectCopy != null) {
      sb.append( " Copy  "+_SelectCopy);
    }
    if (_SelectMove != null) {
      sb.append( " Move  "+_SelectMove);
    }
    if (_SelectAutoDelete) {
      sb.append( " AutoDelete  ");
    }
    if (_ExitInstance != null) {
      sb.append( " Exit  "+_ExitInstance);
    }
    return sb.toString();
  } 
} // End of FaceSelection Class