package com.jthomas.mailfaces;

/*
 * FaceServer.java
 * Copyright (c) 1996 John Thomas  jthomas@cruzio.com
 *      All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for non-commercial purposes
 * is hereby granted provided that this copyright notice
 * appears in all copies.
 *
 */

import java.io.*;
import java.util.*;

import com.jthomas.util.*;

/** 
 * Each FaceServer object contains the information needed to 
 * create a Mailbox object for the specified server. 
 * 
 * The FaceServer class object maintains a list of these FaceServer 
 * objects. 
 * 
 * It contains a createMailbox method that acts as a factory for 
 * FaceMailbox objects. 
 *
 *@see FaceMailbox 
 *
* @author	John Thomas
 * <a href="mailto:jthomas@cruzio.com>jthomas@cruzio.com</a>
 * @version	2.0 ( May 1998 )
 */

public class FaceServer  {

// Class variables
/**
** List of FaceServer objects maintained by the Class 
*/
private static FaceServer[] 	_ServerList;
private static int		_ServerCount;

// state variables

private String		_Type=null;
private String 		_Host=null;
private	String		_User=null;
private	String		_Pass=null;

private boolean     _Respecify; 	// if true, ask user for new info

private FaceProperties _Properties;


//-----------------------------------------
/**
** Constructor for a FaceServer object
**
** It will parse the FaceServer information from the Options file 
** and get the following information for each server (if available)
** <br>Type  ( Pop3,APOP,IMAP)
** <br>Host
** <br>Userid
** <br>Password
** 
**@param String containing the MailServer information from the Options file.
*/
//------------------------------------------

public
FaceServer(String mailServer) {

	String[] tokens = parseServerInfo(mailServer);
	//  Get the Name of the Host and the userid and password for the user
    if ( tokens[0] != null ) 
    	_Type=tokens[0].toUpperCase();       
    _Host = tokens[1];
    _User = tokens[2];
    _Pass = tokens[3];

    _Properties = FaceProperties.getRef();

}

//-----------------------------------------------------------
/**
** routine to parse statement from the options file or properties 
** file.  
** @param Inputstring with format "type  host  userid  password"
** @return array with format {type,host,userid,password}
**     Missing items are null
*/
//----------------------------------------------------------- 
private String[] 
parseServerInfo(String mailServer) {
	StringTokenizer st = new StringTokenizer(mailServer);
	String[] tokens = new String[4];
	for (int i=0; st.hasMoreTokens() && i< 4; i++ ) {
    	tokens[i] = st.nextToken();
    	if ( tokens[i].toUpperCase().equals("*PROMPT*") )
    		tokens[i] = "";
    	//Debug.out(tokens[i]);
	}
	
    return tokens;	
}






//-----------------------------------------
/**
**   createList is a Class method that will 
**   create an array of FaceServer Objects 
**   based on the current FaceOptions information.
*/
//------------------------------------------

public static void
createList() 	{
	String[] list;
	list = FaceOptions.getRef().getList("MailServer");   // get list of Mailservers
    if (list.length == 0) {
    	Debug.out("No Mail Servers found");
        list = new String[1];
        list[0] = "";
    }
    _ServerCount = list.length;
    
	_ServerList = new FaceServer[_ServerCount];
	for (int i=0; i < _ServerCount ; i++) {
		_ServerList[i] = new FaceServer(list[i]);
		Debug.out("Server " + i +": " + _ServerList[i].toString());
	}
		
}



/**
** Class method to return a list of FaceServer objects.
*/
public static FaceServer[]
getList()	{
	return _ServerList;
}

/**
** Class method to return a count of the number of servers.
*/
public static int
getCount()  {
	return _ServerCount;
}

//-----------------------------------------
//------------------------------------------

/**
 * Method: createMailbox()   Create the appropriate type of FaceMailbox object
 * for this FaceServer.
 * 
 * @see Mailbox
 * @return FaceMailbox
 *
 */
//-----------------------------------------------------------------------
public FaceMailbox
createMailbox()  	{
  
  FaceMailbox mailbox=null;
  if (_Type != null) {
    if (_Type.equalsIgnoreCase("IMAP")) {
      try {
        mailbox = new FaceMailboxJMail(this,"IMAP");

      } catch (Throwable e) {
        Debug.out(0,"Unable to create FaceMailboxIMAP object", e);
        MsgDialog.create("Unable to create FaceMailboxIMAP object");
        return null;
      }
    } else if (_Type.equalsIgnoreCase("pop3")) {
      try {
        mailbox = new FaceMailboxJMail(this,"pop3");

      } catch (Throwable e) {
        Debug.out(0,"Unable to create FaceMailboxJMail object", e);
        MsgDialog.create("Unable to create FaceMailboxJMail object");
        return null;
      }  
    
    }
    //  We would add checks here to return other subclasses of FaceMailbox
  } 
  if (mailbox == null) {
    Debug.out(0,"Invalid mailbox Type: "+_Type);
    MsgDialog.create("Invalid mailbox Type: "+_Type);
  }
  return mailbox;
}


public String
getType()  {
	return _Type;
}
public String
getHost()  {
	return _Host;
}
public String
getUser()  {
	return _User;
}
public String
getPass()  {
	return _Pass;
}
public String
getName()  {
	return _User + "@" + _Host;
}

//-------------------------------------------------------------
/**
**    Use the PromptLogin dialog to get any missing Login info
*/
//-------------------------------------------------------------
public boolean
getLoginInfo() {

	boolean ok = verifyLoginInfo();     // see if it already exists
	if (ok)
		return true;
		
	PromptLogin p = (PromptLogin)PromptLogin.create();    
    
	p.setLoginDefaults(_Type,_Host,_User,_Pass);
    p.showPrompt();
    int status = p.getStatus();
    
    if (status == PromptLogin.STATUS_CANCEL) {    	
		Debug.out("User pressed Cancel");
    	return false;
    }
    _Type = p.getType();
    _Host = p.getHost();
    _User = p.getUser();
    _Pass = p.getPass();
	Debug.out("getLoginInfo: result= " + this.toString());
	Debug.out("getLoginInfo: Status= " + p._Status);
	_Respecify = false;
	// save the info in the ini file so it can be reused 
	// the next time.  
	_Properties.put("FaceServer",_Type+" "+_Host+" "+_User);
	_Properties.save();
	return true;
}

/**
** Set a flag so that User is forced to respecify the Server information
** This could be done when the connection could not be made so we 
** want to let the user check if all the info was supplied correctly.
*/
public void
setRespecify(boolean flag) {
	_Respecify = flag;
}
	
//---------------------------------------------------------------
/**
** verifyLoginInfo
**	Check that all of the required Login information is avaiable
**  If there is no information at all, then check the Properties 
**  file. 
*/
//----------------------------------------------------------------
public boolean
verifyLoginInfo() {
	if ( _Respecify )
		return false;
	// If none of the parameters are found then look at the
	// Properties to get the set of parameters used last.	
    if ( _Type == null) {
    	String mailServer =_Properties.getProperty("FaceServer",null);
    	if (mailServer == null) 
    		return false;
    	String[] tokens = parseServerInfo(mailServer);
		//  Get the Name of the Host and the userid and password for the user
    	_Type=tokens[0].toUpperCase();       
    	_Host = tokens[1];
    	_User = tokens[2];
    	_Pass = tokens[3];
		if ( _Type == null) 		
    		return false;
    }
    // Check if the Tpe is correct
    if ( 	! _Type.equals("POP3") &&     	 
    	  ! _Type.equals("HOTMAIL") &&
			  ! _Type.equals("YAHOO") &&
    	  ! _Type.equals("IMAP") ) {
    	Debug.out(0,"Required Mail Server information not specified.\n" +
    					"MailServerType=" + _Type);
    	return false;
	}
	
	// Check if the Host name and User name are supplied
    if ( _Host == null || _User == null ){
    	Debug.out("Required Mail Server information not specified.\n" +
    					"MailHost=" + _Host + "\n" +
    					"MailUserid=" + _User  );
    	return false;
	}
	// check if the password is supplied
	if ( _Pass == null || _Pass.length() == 0 ||
	     _Pass.toUpperCase().equals("*PROMPT*") ){
    	Debug.out("Password not specified");
    	return  false;
	}
    // We have all the needed info.
    return true;
}

public String
toString() {
	return _Type + "," + _Host + "," + _User + "," + (_Pass==null? "null":"suppressed");
}

} // End of FaceServer Class