package com.jthomas.mailfaces;

/*
 * FaceStatusLabel.java
 * Copyright (c) 1996 John Thomas  jthomas@cruzio.com
 *      All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for non-commercial purposes
 * is hereby granted provided that this copyright notice
 * appears in all copies.
 *
 */
import java.awt.*;
import java.util.*;
import com.jthomas.util.*;

/**
 *    This class implements a Status label that is used to
 *    display the status while MailFaces is running.
 *    The reason for making it a subclass of Label is that we plan
 *    to have only one instance of _StatusLabel class and we want to be
 *    able to update the text from anywhere using a Class method.
 * <p>
 *    Because different threads can be writing status methods this 
 *    uses a stack where each thread would use a specific key to identify
 *    messages written by that thread.  The stack will only contain the 
 *    latest message written with that key.  When a tread clears it's message 
 *    then the previous message is popped from the stack and displayed.
 * <p>
 * 
 *  Usage: 
 *   To write a status message<br>
 *      FaceStatusLabel.setStatus(key,"any status message");
 *   To clear the above status message
 *      FaceStatusLabel.clearStatus(key);
 * 
 *   If the clearStatus method is used, then if the specified 
 *   method is still being shown, it  is cleared and 
 *   the previous status message(if any) is displayed.    
 *-------------------------------------------------------------
 */


public 
class FaceStatusLabel extends Label {


private static final int MINWIDTH=100;

// We will call the constructor when we create the Class.
private static	 FaceStatusLabel	_StatusLabel;

private static   boolean			_FaceStatusFlag;
/**
** Keep track of the dimmensions of the FaceStatusLabel
*/
private int		_Width;
private int		_Height;
/**
** Current message
*/
private String  _Message="Status";

/**
** Current message key
*/
private String	_MessageKey="";
/**
** Stack of messages
*/
private static Stack	_MessageStack = new Stack();

private FontMetrics		_FM;



//-----------------------
/**
** Private constructor
*/
//------------------------
public
FaceStatusLabel() {
	super("Starting MailFaces");
	setBackground(new Color(0xD2B48C));
	Font font = new Font("Dialog",Font.BOLD,12);
	this.setFont(font);
	_FM = this.getFontMetrics(font);
	_Height = _FM.getHeight();
	_Width = MINWIDTH;
	_StatusLabel = this;
}









/**
** Get the reference to our one and only Status label
*/
public static FaceStatusLabel
getRef() {
	return _StatusLabel;
}







/**
** Set a new Status label
*/
public static synchronized void
setStatus(String key, String statusinfo) {
	
	Debug.out("FaceStatusLabel.setStatus: " +key+";"+ statusinfo);
	/****
	if (! _FaceStatusFlag ) {
		return;
	} 
	****/
	if (_StatusLabel == null) {
		Debug.out("StatusLabel has not yet been created");
		return;
	}
	String[] status = new String[2];
	status[0] = key;
	status[1] = statusinfo;
	removemsg(key);      // Remove any old message with this key.
	_MessageStack.push(status);   // Save this one on the stack
	// Now display this message.
	_StatusLabel.setText(statusinfo);
	_StatusLabel._Message = statusinfo;
	_StatusLabel._MessageKey = key; 	
	_StatusLabel._Height = _StatusLabel._FM.getHeight();
	int width = _StatusLabel._FM.stringWidth(statusinfo);
	_StatusLabel._Width = Math.max(MINWIDTH,width);
	
	
}



/**
** clear all Status labels 
*/
/***
public static synchronized void
clearStatus() {
	_MessageStack.removeAllElements();
	_StatusLabel.setText(""); 

}
******/



/**
** clear specified Status label 
*/
public static synchronized void
clearStatus(String key) {
	
	Debug.out("FaceStatusLabel.clearStatus: " + key);
	/*****
	if (! _FaceStatusFlag ) {
		return;
	} 
	*****/
	if (_StatusLabel == null) {		
		return;
	}
	String statusinfo = "";
	String oldkey = _StatusLabel._MessageKey;
	if ( _MessageStack.empty() ) 
		return;
		
	if (oldkey.equals(key)) {		
		_MessageStack.pop();  
		if (! _MessageStack.empty()) {
			
			String[] status = (String[])_MessageStack.peek();
			statusinfo = status[1];
			_StatusLabel._MessageKey= status[0];
		}
		_StatusLabel.setText(statusinfo);
		int width = _StatusLabel._FM.stringWidth(statusinfo);
		_StatusLabel._Width = Math.max(MINWIDTH,width);
		_StatusLabel._Message = statusinfo; 
	} else {
		removemsg(key);	
	
	}
	
	
	
	
}


/**
** Remove a Message from the stack of messages
*/
private static void
removemsg(String key) {
	Enumeration enum = _MessageStack.elements();
	while (enum.hasMoreElements()) {
		String[] entry = (String[])enum.nextElement();
		if ( entry[0].equals(key)) 
			_MessageStack.removeElement(entry);
	}
}



/**
** The user can choose to show or hide the Status label.
** This is the routine that does it.
*/
public synchronized static void
setStatusFlag(boolean statusflag) {
	_FaceStatusFlag = statusflag;
	if ( statusflag)  {		// if this is turning on the flag
		_StatusLabel.setText(_StatusLabel._Message);
		_StatusLabel.setVisible(true);     // show the label
	}
	else
		_StatusLabel.setVisible(false);     // hide the label
	
}









/**
** Get the current state of whether we are showing the 
** Status label or not.
*/
public synchronized static boolean
getStatusFlag() {
	return _FaceStatusFlag;
}





/**
** Override this method to specify special dimensions
*/
public Dimension
getMinimumSize() {
	
	Debug.out("StatusLabel.MinimumSize: " + MINWIDTH + "," + _Height);		
	if ( _FaceStatusFlag)
		return new Dimension(MINWIDTH, _Height);
	else 
		return new Dimension(0,0);	
}





/**
** Override this method to specify special dimensions
*/

public Dimension
getPreferredSize() {
  	Debug.out("StatusLabel.PreferredSize: " + _Width + "," + _Height);		
	if ( _FaceStatusFlag)
		return new Dimension(_Width, _Height);
	else 
		return new Dimension(0,0);
}

}		// end of FaceStatusLabel class