package com.jthomas.mailfaces;

/*
 * MailFaces.java
 * Copyright (c) 1996-1998 John Thomas  jthomas@cruzio.com
 *      All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for non-commercial purposes
 * is hereby granted provided that this copyright notice
 * appears in all copies.
 *
 */
import java.util.*;

import com.jthomas.util.Debug;
import com.jthomas.util.Queue;
import com.jthomas.util.Log;
import com.jthomas.util.JEditor;
/**
 *  This is the main class for the MailFaces application.  
 
 *   <p>
 *   This application will examine the mail waiting at a Mail Server and
 *   display an Icon that represents the sender of the Mail.
 *   The user can either select unique images to be displayed based
 *   on information from the Mail headers or by default it will look
 *   up the From: user@host and search the PIcon Database for a
 *   matching image.
 *   <p> 
 *
 *  MailFaces creates 4 main Class Objects that work together along with some
 *  supporting Classes
 *  <ul>
 *  <li>FaceMailbox<br> 
 *       This represents an instance of Mail Server with the 
 *       users associated mailbox.  When a new mailmessage arrives in the 
 *       mailbox, a Mailmsg object is created and placed into the FaceResolver queue.
 *       FaceMailbox is an abstract object and is implemented by 
 *    	 a class that know how to read a designated type of Mailbox.
 *		 For example: FaceMailboxJMail which uses JavaMail for POP3 and IMAP
 * 
 *   <li>FaceResolver<br>
 *		This component gets Mailmsgs from its queue
 *		and determines what images should be displayed 
 *    for the Mailmsg object. It uses the various FaceSelect instances to 
 *    match the Select, KillSelect and SelectAfter statements to 
 *    against the Mailmsg headers. 
 *		If a match is found, then the image from the corresponding Select statement is used
 *		and the Mailmsg is queued to the FaceBoard queue.  
 * 		If there is no match, then the Mailmsg is queued to the PiconDB queue.
 *    If the Slect statement has contains an SMS keyword, then the FaceSMS sendMessage
 *    method is called to send an SMS (Short Message Service) to a cellphone.
 *
 *	 <li>PiconDB<br>
 *		This component takes the From: header from the mailmsg and sends a query to the 
 *		PiconDB server as defined in the Options file.  The PiconDB server returns
 *        a URL for a Picon that represents either the user@host or the Picon that represents
 *        the most specific host address.  If it can find no match it returns an "unknown"
 *    	icon.  If an icon is found, it is checked against a list of PiconDefer nodes.
 *        If no icon is found or if it is in the PiconDefer list, then the Mailmsg is sent
 *		back to FaceResolver to check the SelectAfter list, otherwise it is sent to the 
 *		FaceBoard queue.   
 *
 *   <li>FaceBoard<br>
 *        This is a subclass of java.AWT.panel. 
 *        This handles the display of the images based on instructions from FaceResolver
 *		It gets a Mailmsg from its queue and uses the specified URL to build an image 
 *		button.
 *  </ul><p>
  
 *  The following are Classes that are used by the above.
 *  
 *  <ul>
 *    <li>FaceFrame<br>
 *       This contains all of the GUI elements and contains the 
 *       FaceBoard panel.  The logic for the menubar elements is
 *       contained in this class.
 *    <li>Face<br> 
 *       This represents the Face Icon that will be shown for 
 *       one or more mail messages.
 *	  <li>FaceAudio<br>
 *		This class has a queue that other classes can place URLs
 *  	     for "au" sound files.  It will remove the URL form the 
 *		queue and play the sound.
 *	  <li>FaceServer<br>
 *		This represents an instance of a Mail server that the 
 *        user has specified.  It is a factory for FaceMailbox 
 *        objects.
 *    <li>FaceCache<br>
 *        This class manages a cache of icons that is maintained on the users
 *        local file system.
 *    <li>FaceGroup<br>
 *        This class manages a list of members of a group.  The groups are 
 *        read at initialization from an exported Address book.
 *
 *    <li> FaceSelect<br>
 *         There are instances of this class repesenting the Select, KillSelect
 *         and SelectAfter processing.  These use FaceSelection objects which 
 *         are built from the individual Select statements.
 *    <li> FaceSMS<br>
 *         Extract a SMS (Short MessageService) message from the mailmessage and 
 *         send it to a cell phone
 *    <li> FaceGroup<br> 
 *         Read a CSV (Comma separated Variables) file containing email addresses 
 *         from an address book and support matching the from header to the 
 *         the list of addresses.        
 *  </ul>
 *  There are also a number of GUI objects that are involved in 
 *  displaying the results.  These are documented in the FaceBoard
 *  documentation.  
 *	<p>
 *  There are also some utility Objects that are involved
 *  
 *  <ul>
 *    <li>FaceOptions<br> 
 *       This handles the parsing and storage of the Options file
 *       that specifieds the control options for using the application
 *    <li>Log<br>
 *       Writes major events to a log file
 *     <li>Queue<br>
 *       Provides a simple queuing of requests. 
 *        MailFaces makes heavy use of this class to communicate 
 *        between classes.  
 *	</ul>
 *
 *
 *  <p>
 *   For additional details, refer to the
 *   <a href="http://wwww.cruzio.com/~jthomas/mailfaces">
 *     MailFaces online documentation</a>
 *
 * <p>
 * Credits:<br>
 *   <pre>
 *   ---------------------------
 *   ImageButton object from:
 *      Marty Hall:
 *      <A href="http://www.apl.jhu.edu/~hall/java">
 *        Java Resource Page.</A> 
 *      <A href="mailto:hall@apl.jhu.edu">
 *         hall@apl.jhu.edu</a>
 *   -------------------------
 *   Regexp Object from:
 *      Copyright (c) 1995 Starwave Corporation.  All Rights Reserved.
 *      A perl-like regular expression matcher for java.
 *      Version 1.10, 02 Feb 1996
 *      Author Jonathan Payne
 *  
 *   --------------------------
 *   </pre>
 * <p>
 *Class: MailFaces.java
 *  <p>
 *   Main application object for MailFaces.
 *	<p>
 *  MailFaces.java contains the "main" method that is given control initially.
 *  It will create a MailFaces object and give control to the "init()" method.
 *  
 *  The init method will do the initial setup of the Mailfaces environment and then 
 *  invoke the "run()" method that will start up the various threads that will 
 *  actually do the work and then wait for the other threads to terminate.  
 
 *  The main purpose of the MailFaces class (other than getting things started) 
 *  is to act as the 
 *  global namespace for communication between other components.  
 * 
 ** 
 * @see FaceResolver 
 * @see FaceBoard
 * @see FaceFrame  
 * @see FaceServer 
 * @see FaceOptions 
 * @see FaceSelect
 * @see FaceSelection
 * @see FaceSMS
 * @see FaceGroup 
 * @see Queue  
 * @see FaceMailbox 
 * @see Mailmsg
 * @see PiconDB
 * @see FaceOptions
 * @author	John Thomas <a href="mailto:jthomas@cruzio.com">jthomas@cruzio.com</a>
 * @version	2.1 ( August 1998 )
 */

public class MailFaces implements Runnable {

  /**
  ** Reference to the one and Only MailFaces object
  ** Other objects can access via the MailFaces.getRef() method
  */
  static MailFaces _MailFaces; // object for this application

  /**
  ** MailFaces Version
  ** <p> See http://www.cruzio.com/~jthomas/mailfaces
  ** for History of changes with notes on what is included in each release
  */
  static final String _Version = "2.1.5";
  static final String _VersionDate = "July 25, 2002";

  /**
  ** Public(for package) State variables
  */

  // Private State variables

  /**
  ** This is the number of seconds to delay between MailChecks
  */
  private int _Delay;

  /**
  ** MailFaces Properties object
  */

  /**
  ** The following flags are set in FaceFrame.actions and used in MailFaces
  ** There are get/set methods to set and query these flags
  */

  /**
  ** When set, then the Mailbox will be checked for new mail 
  ** even if the timer has not expired.
  */
  private boolean _FlagCheckMail = false;

  /**
  ** When set, the entire contents of the mailbox will be
  ** reprocessed
  */
  private boolean _FlagResetMail = false;

  /**
  ** When set, a new MailServer will be accessed
  */
  private boolean _FlagResetServer = false;

  /**
  ** When set the Mailbox process will terminate
  */
  private boolean _FlagQuitMail = false;

  /**
  ** When set, Audio clues will be given
  */
  private boolean _FlagAudio = false;

  /**
  ** Properties object that remembers the cross session information
  */
  private FaceProperties _Properties;

  /**
  ** FileName for Options file prepared by user
  */
  private String _OptionsFile;
  /**
  ** Object that stores the parsed Options.
  */
  private FaceOptions _Options;
  
  /**
  ** Object that determines what image should be displayed for 
  ** a selected Mailmsg
  */
  private FaceResolver _Resolver;
  private Thread _ResolverThread;

  /**
  ** A cache of images and information about what image should 
  ** be displayed for a Mailmsg (used by FaceResolver) 
  */
  private FaceCache _Cache;

  /**
  ** FaceFrame is the main window that contains the menubar, FaceBoard and Status
  */
  private FaceFrame _FaceFrame;
  /**
  ** FaceBoard GUI objects
  */
  private FaceBoard _FaceBoard; // Panel that contains Faces    
  private Thread _FaceBoardThread; // Controlling Thread

  /**
  ** Current Mailbox object
  */
  private FaceMailbox _Mailbox;

  private int _MailboxMsgCount;
  private int _OldMsgCount;
  private Thread _MailboxThread;

  /**
  ** Object that interfaces to a PIcon DataBase
  */
  private PiconDB _PiconDB;
  private Thread _PiconDBThread;

  /**
  ** Current Mail Server object
  */
  private FaceServer _Server; // Current Server
  /**
  ** List of Mail Servers
  */
  private FaceServer[] _ServerList; // List of available Servers

  private int _ServerCurrent = 0; // Number of current server	

  /**
  ** Vector that keeps track of who wants know when a FaceEvent occurs 
  ** 
  */
  private Vector _FaceEventListeners = null;

  /**
  ** Object that handles the playing of audio
  */
  private FaceAudio _FaceAudio; // Audio player     
  private Thread _FaceAudioThread; // Controlling Thread

  /**
   * reference to class that handles the SMS messaging
   */
  private FaceSMS  _FaceSMS = null;
  
  /**
   * reference to class that handles the Address book groups
   */
  private FaceGroup  _FaceGroup = null;
  //------------------------------------------------------------------
  // Static class methods
  //------------------------------------------------------------------

  //------------------------------------------------------------------
  /**
   *   Returns the MailFaces application Version and release Date.
   *   
   *   For Example:
   *   <br>V 1.2.1 - February 7, 1997
   *
   *@return String  
   */
  //------------------------------------------------------------------
  public static String version() {
    return _Version + " - " + _VersionDate;
  }

  //------------------------------------------------------
  /**
   *Class method that returns the reference to the only MailFaces Object
   *
   *@return MailFaces    Reference to MailFaces object 
   */
  //------------------------------------------------------
  static public MailFaces getRef() {
    Debug.asserted((_MailFaces != null), "MailFaces instance not available");

    return _MailFaces;
  }

  //------------------------------------------------------------------    
  // MailFaces Constructor
  //------------------------------------------------------------------
  /**
  ** Default Constructor is private so it can only be created by main()
  */
  private MailFaces() {
  }

  //------------------------------------------------------------------
  /**
   *    Called by MailFaces.main() to do initial setup.
   * 
   *
   * <pre>
   *  - Setup FaceProperties object from MailFaces.ini file.
   *  - Call setupOptions to read the MailFaces Options file
   *  - Setup the FaceBoard object that will handle User interface
   
   *  - Setup the Mailbox object
   *  - Setup the Image Cache 
   *  - Setup the Resolver Object 
   * </pre>
   * @return void
   */
  //---------------------------------------------------------------
  public void init() {

    _Properties = FaceProperties.create("MailFaces.ini");

    if (!this.setupOptions()) {
      _FlagQuitMail = true; // failed while reading options.
      return;
    }
    try { 
      boolean log = _Options.get("Log", "true").equals("true");
      Log.setLogOn(log, _Options.get("LogFile", "faces.log"));
      Log.out("Start", "Start MailFaces Version " + version());
      String time = (new Date()).toString().substring(0,19);
      Debug.println(time + " Start MailFaces Version " + version());
      // Setup the Server Object and get any missing login info 	
      // Create a list of MailServers    
      FaceServer.createList();
      _ServerList = FaceServer.getList();
  
      _FaceAudio = new FaceAudio();
  
      _FaceSMS = new FaceSMS();
      
      _FaceGroup = new FaceGroup();
      
      // Setup the Cache Object.
      String dir;
      if ((dir = _Options.get("CacheDirectory", null)) != null) {
        try {
          if (_Cache != null)
            _Cache.writeIndex(); // Save current Cache index    		
          _Cache = new FaceCache(dir);
        } catch (Exception e) {
          Debug.out("Unable to open the Image Cache");
          _Cache = null;
        }
      }
  
      // Setup the object that resolves mailmsgs into appropriate icons
      _Resolver = new FaceResolver();
  
      // Create the PiconDB object 
      _PiconDB = new PiconDB();
  
      // Setup the FaceBoard. It is a panel that will display the Faces that
      // represent the mail messages that are in the users mailbox
  
      _FaceBoard = new FaceBoard();
      _FaceBoard.setup();
  
      // Setup the FaceFrame.  It is the Main GUI window that the user 
      // interfaces with.  It contains the menubar, FaceBoard and FaceStatus 
      _FaceFrame = new FaceFrame(_FaceBoard);
      _FaceFrame.setCursorBusy(true);
          //----------
      // Startup the threads that do all the work
      //----------
      if (_FaceAudioThread == null) {
        Debug.out("Start FaceAudioThread");
        _FaceAudioThread = new Thread(_FaceAudio, "FaceAudioThread");
        _FaceAudioThread.start();
      }
      _MailboxMsgCount = -1;
      _OldMsgCount = 0;
  
      // We have a problem with the following thread startup
  
      if (_ResolverThread == null) {
        Debug.out("Start FaceResolverThread");
        _ResolverThread = new Thread(_Resolver, "ResolverThread");
        _ResolverThread.start();
      }
  
      // Let the above task get going - May avoid a strange deadlock
      try {
        Thread.sleep(200);
      } catch (Exception e) {
      }
      Debug.out(_ResolverThread.toString() + " isAlive=" + _ResolverThread.isAlive());
      //_Resolver.queuePut(null); 
  
      if (_PiconDBThread == null) {
        Debug.out("Start PiconDBThread");
        _PiconDBThread = new Thread(_PiconDB, "PiconDBThread");
        _PiconDBThread.start();
      }
      // Let the above task get going - May avoid a strange deadlock
      try {
        Thread.sleep(200);
      } catch (Exception e) {
      }
  
      Debug.out(_PiconDBThread.toString() + " isAlive=" + _PiconDBThread.isAlive());
      //_PiconDB.queuePut(null); 
  
      if (_FaceBoardThread == null) {
        Debug.out("Start FaceBoardThread");
        _FaceBoardThread = new Thread(_FaceBoard, "FaceBoardThread");
        _FaceBoardThread.start();
      }
  
      // Let the above tasks get going - May avoid a strange deadlock
  
      try {
        Thread.sleep(200);
      } catch (Exception e) {
      }
      Debug.out("Threads should be started!");
    } catch (Throwable te) {
      Debug.out(te);
      _FlagQuitMail = true; // failed during initialization  
    }
  }

    
  //------------------------------------------------------------------
  /**
  ** editOptionsSupported() will verify that JEditor 
  *  can be invoked  (Requires Java 2)
  * @return true if can be JEditor can be created.
  */
  //------------------------------------------------------------------
  public boolean editOptionsSupported() {
    
 
    boolean returns = true;
    try {
      final JEditor editor = new JEditor();
    } catch (NoClassDefFoundError e) {
      Debug.out(0,"MailFaces:  Editing of Options file not supported on Java 1.1");
      returns = false;
    }
    return returns;
  }
  //------------------------------------------------------------------
  /**
  ** editOptions() will invoke the JEditor internal editor on
  *  the .opt file  
  *
  * @return void
  */
  //------------------------------------------------------------------
  public void editOptions() {
    Debug.out("editOptions():");

  
    try {
      final JEditor editor = new JEditor();
//	  editor.addWindowListener(new WindowAdapter() {
//		public void windowClosing(WindowEvent e) {
//			Debug.out(0,"windowClosing");
//		}
//		public void windowClosed(WindowEvent e) {
//			Debug.out(0,"windowClosed");
//			
//		}
//	  });
      editor.setFileName(_Options.getFileName());

      editor.setup();
      

    } catch (Throwable e) {
      Debug.out(e);
    }
  }
  //------------------------------------------------------------------
  /**
  ** setupOptions() will invoke the FaceOptions object to parse and 
  ** store the various options that the user has chosem.
  ** These options can later be queried using the various "get(name)"
  ** methods.
  ** @return true if processed, false if unable to process Option file
  */
  //------------------------------------------------------------------
  private boolean setupOptions() {
    Debug.out("setupOptions():");

    try {
      if (_Options == null)
        _Options = new FaceOptions(_OptionsFile);
      _Options.parseOptions(null);
    } catch (Exception e) {
      String msg = "Options Error: " + e.getMessage();
      Log.out("Options", msg);
      MsgDialog.create(msg);

      return false;
    }
 
    // Verify that certain important options are specified correctly.
    _Options.verifyOptions();

    // set to produce Debug output if user wants it.
    // It may have already been turned on by the command line option. 

    if (_Options.getBoolean("Debug", false)) {
      Debug.setDebugOn(true);

    }
    // allow console output to be turned on by Options file
    // The distributed default is to only use a file
    boolean console = Debug.getDebugOut();
    if (!console) {  
      console = _Options.getBoolean("DebugConsole", false);
    }
    if (console) {
      Debug.setDebugOut(true);
    }
    String file = _Options.get("DebugFile", null);
    if (file != null) {
      Debug.setDebugFile(file);
    }

    if (Debug.isDebugOn()) {
      // display some useful info if we ever have to 
      // look at a log file generated elsewhere.	
      Debug.out(
        "Start MailFaces Version "
          + version()
          + " at "
          + (new Date()).toString().substring(0, 19));
      String version = System.getProperty("java.version");
      Debug.out("java.version=" + version);

      String os = System.getProperty("os.name");
      Debug.out("os.name=" + os);

      String userdir = System.getProperty("user.dir");
      Debug.out("user.dir=" + userdir);

      String compiler = System.getProperty("java.compiler");
      Debug.out("java.compiler=" + compiler);

    }
    setDelay(_Options.getInt("DelayTime", 30));
    _MailFaces._FlagAudio = _Options.getBoolean("AudioSupport", false);

    return true;
  }

  //------------------------------------------------------------------
  /**
   *   Reprocess the Options file when User has made a Refresh request.
   */
  //------------------------------------------------------------------
  public void refreshOptions() {
    Debug.out("refreshOptions():");
    this.setupOptions();

   
  }

  //------------------------------------------------------------------
  /**
   *   Change the Server to be accessed.
   *   <p>
   *   This method is called
   *   from the event handler for the menu bar "Server" item.
   */
  //------------------------------------------------------------------

  public void setServer(int ServerNumber) {
    Debug.out("MailFaces.setServer " + ServerNumber);
    if (_ServerCurrent != ServerNumber) {
      _ServerCurrent = ServerNumber;
      _FlagResetServer = true;
      setFlagCheckMail(true);

    }
  }

  //------------------------------------------------------------------
  /**
   * Main processing loop for MailFaces.
   *
   * <pre>
   *   start processing threads
   *   While (not done)
   *      open Mailbox
   *      Check if any new mail and process it
   *      delay until time to check it again
   * </pre>
   * @return void
   */
  //------------------------------------------------------------------

  public void run() {
    Debug.out("MailFaces.run()");

    //--------------------------
    // Startup a thread for the Mail Server
    //   It will go thru this do-while only once normally
    //   unless more than one mailservers were specified 
    //   and the user has asked to switch servers.  
    //-----------------------------
    do {
      _FlagResetServer = false;
      Debug.out("Access MailServer");
      // Setup the current MailServer
      _Server = _ServerList[_ServerCurrent];

      while (!_Server.verifyLoginInfo()) {
        if (!_Server.getLoginInfo()) { // If cancel while asking for login info
          _FlagQuitMail = true;
          break;
        }
      }
      // Create the Mailbox
      Debug.out("Create Mailbox object for " + _Server.toString());
      _Mailbox = _Server.createMailbox();
      if (_Mailbox == null) { // Unable to create a mailbox.

        _Server.setRespecify(true);
        if (!_Server.getLoginInfo())
          _FlagQuitMail = true;
        else
          _FlagResetServer = true;
        continue; // then just continue
      }
      _Mailbox.init();
      Debug.out("Start Mailbox thread");

      Log.out("OpenMailbox", _Mailbox.toString());

      _FaceFrame.setTitle(_Server.getUser() + "@" + _Server.getHost());
      _MailboxThread = new Thread(_Mailbox, "MailboxThread");
      _MailboxThread.start();
      //Debug.out("priority="+_MailboxThread.getPriority());

      _MailboxThread.setPriority(_MailboxThread.getPriority() + 1);
      // now wait for MailboxThread to end
      try {
        Debug.out("waiting...");
        _MailboxThread.join();
      } catch (Exception e) {
      }
      // we only get here if the Mailbox Thread terminates
      _FlagQuitMail = false;
      // If FlagResetServer was set true then _ServerCurrent has new server #
    }
    while (_FlagResetServer); // Flag would be set by user from Servers menu item

    // Update the Cache Index file.	
    if (_Cache != null)
      _Cache.writeIndex();

    _Properties.save();
    Debug.out("MailFaces is done!");
    Log.out("Stop", "MailFaces is done!");
    System.exit(0);
    return;
  }

  //------------------------------------------------------------------
  /**
  ** displayDebugInfo is called by FaceFrame when the user clicks the DebugInfo 
  ** 
  **  It will display some debug info.  What it displays may change from time to time
  **  since this will often be used to chase different problems.
  **
  */
  //------------------------------------------------------------------

  public static void displayDebugInfo() {
    Runtime runtime = Runtime.getRuntime();
    long total = runtime.totalMemory()/1024;
    long free = runtime.freeMemory()/1024;
    Debug.out(0,"Memory: Total=" + total + "K" + 
                "  Free=" + free + "K" +
                "  Used=" + (total-free) + "K");
    //Thread.dumpStack();

    Thread t = _MailFaces._MailboxThread;
    Debug.out(0, t.toString() + " isAlive=" + t.isAlive());

    t = _MailFaces._ResolverThread;
    Debug.out(0, t.toString() + " isAlive=" + t.isAlive());

    _MailFaces._Resolver.queuePut(null);

    t = _MailFaces._PiconDBThread;
    Debug.out(0, t.toString() + " isAlive=" + t.isAlive());
    _MailFaces._PiconDB.queuePut(null);

    t = _MailFaces._FaceBoardThread;
    Debug.out(0, t.toString() + " isAlive=" + t.isAlive());
    _MailFaces._FaceBoard.queuePut(null);

    t = _MailFaces._FaceAudioThread;
    Debug.out(0, t.toString() + " isAlive=" + t.isAlive());
    _MailFaces._FaceAudio.queuePut(null);


  }

  //------------------------------------------------------------------
  /**
  ** Adds the FaceEventListener to the collection of those
  ** interested in new FaceOptionsEvent
  **
  ** @param rcl the FaceEventListener to add
  */
  //------------------------------------------------------------------

  public synchronized void addFaceEventListener(FaceEventListener rcl) {
    if (_FaceEventListeners == null)
      _FaceEventListeners = new Vector(10);
    if (!_FaceEventListeners.contains(rcl))
      _FaceEventListeners.addElement(rcl);
  }

  //------------------------------------------------------------------
  /**
  ** Removes the FaceEventListener from the collection of
  ** those interested in new FaceOptionsEvents
  **
  ** @param ml the FaceEventListener to remove
  */
  //------------------------------------------------------------------

  public synchronized void removeFaceEventListener(FaceEventListener rcl) {
    if (_FaceEventListeners == null)
      return;
    _FaceEventListeners.removeElement(rcl);
  } // removeFaceEventListener

  //------------------------------------------------------------------
  /**
  ** Send a FaceEvent to all registered listeners
  * 
  * @see FaceEvent
  * 
  * @param id   FaceEvent identifier
  */
  //------------------------------------------------------------------

  public void fireFaceEvent(int id) {
    Vector targets = null;
    synchronized (this) { // be very careful
      if (_FaceEventListeners == null) { // if never used             
        return; // we have no listeners
      }
      targets = (Vector) _FaceEventListeners.clone();
    }
    // define the Event object
    FaceEvent event = new FaceEvent(this, id);
    for (int i = 0; i < targets.size(); i++) {
      FaceEventListener target = (FaceEventListener) targets.elementAt(i);
      target.actionPerformed(event);
    }
  }

  //------------------------------------------------------
  /**
   * This returns the reference to the one and Only FaceCache object
   *
   *@return FaceCache    Reference to FaceCache object 
   */
  //------------------------------------------------------
  public FaceCache getCacheRef() {
    return _Cache;
  }

  //------------------------------------------------------
  /**
   * This returns the reference to the current Mailbox object
   *
   *@return FaceMailbox    Reference to current FaceMailbox object 
   */
  //------------------------------------------------------
  public FaceMailbox getMailboxRef() {
    return _Mailbox;
  }

  //------------------------------------------------------
  /**
   * This returns the reference to the one and Only 
   * FaceResolver object
   *
   *@return _Resolver    Reference to FaceResolver object 
   */
  //------------------------------------------------------
  public FaceResolver getResolverRef() {
    return _Resolver;
  }

  //------------------------------------------------------
  /**
   * Method to access the one and Only PiconDB object
   *
   *@return Reference to PiconDB object 
   */
  //------------------------------------------------------

  public PiconDB getPiconDBRef() {
    return _PiconDB;
  }

  //------------------------------------------------------
  /**
   * Method to access the one and Only FaceBoard object
   *
   *@return FaceBoard    Reference to FaceBoard object 
   */
  //------------------------------------------------------

  public FaceBoard getFaceBoardRef() {
    return _FaceBoard;
  }
  
  //------------------------------------------------------
  /**
   * Method to access the one and Only FaceFrame object
   *
   *@return _FaceFrame    Reference to FaceFrame object 
   */
  //------------------------------------------------------

  public FaceFrame getFaceFrameRef() {
    return _FaceFrame;
  }
  //------------------------------------------------------
  /**
   * Method to access the one and Only FaceAudio object
   *
   *@return _FaceAudio    Reference to FaceAudio object 
   */
  //------------------------------------------------------

  public FaceAudio getFaceAudioRef() {
    return _FaceAudio;
  }

  //------------------------------------------------------
  /**
   * Method to access the one and Only Options object
   *
   *@return _Options    Reference to FaceOptions object 
   */
  //------------------------------------------------------

  public FaceOptions getFaceOptionsRef() {
    return _Options;
  }

    //------------------------------------------------------
  /**
   * Method to access the one and Only Options object
   *
   *@return _Options    Reference to FaceOptions object 
   */
  //------------------------------------------------------

  public FaceSMS getFaceSMSRef() {
    return _FaceSMS;
  }
  //------------------------------------------------------
  /**
   * Method to allow the user to get/set the Delay interval
   *
   *@param int delay -  new delay interval
   */
  //------------------------------------------------------
  public void setDelay(int delay) {
    _Delay = delay;
  }

  public int getDelay() {
    return _Delay;
  }

  //------------------------------------------------------
  /**
   * Method to allow the components to signal that the application is busy
   *
   *@param boolean true if application is busy.
   */
  //------------------------------------------------------
  public void setBusy(boolean busy) {
    if (Debug.ON) Debug.out("Cursor busy = " + busy);
    _FaceFrame.setCursorBusy(busy);
  }

  //------------------------------------------------------
  // Methods to set and get the various processing Flags
  //------------------------------------------------------

  /**
  ** return/set the _FlagCheckMail flag
  **
  */
  public boolean getFlagCheckMail() {
    return _FlagCheckMail;
  }
  public void setFlagCheckMail(boolean flag) {
    _FlagCheckMail = flag;
  }

  /**
  ** return/set the _FlagResetMail flag
  **
  */
  public boolean getFlagResetMail() {
    return _FlagResetMail;
  }
  public void setFlagResetMail(boolean flag) {
    _FlagResetMail = flag;
  }

  /**
  ** return/set the _FlagResetServer flag
  **
  */
  public boolean getFlagResetServer() {
    return _FlagResetServer;
  }
  public void setFlagResetServer(boolean flag) {
    _FlagResetServer = flag;
  }

  /**
  ** return/set the _FlagQuitMail flag
  **
  */
  public boolean getFlagQuitMail() {
    return _FlagQuitMail;
  }
  public void setFlagQuitMail(boolean flag) {
    _FlagQuitMail = flag;
  }

  /**
  ** return/set the _FlagAudio flag
  **
  */
  public boolean getFlagAudio() {
    return _FlagAudio;
  }
  public void setFlagAudio(boolean flag) {
    _FlagAudio = flag;
  }

 
  //------------------------------------------------------------------
  /**
   *  The "main" class method is invoked when MailFaces is invoked as a
   *  Java application.
   *
   *  It will create a MailFaces object.  Then it will process the 
   *  command line operands if present.  
   *
   * <pre>
   *  Args:
   *     -d 			Debug flag
   *     
   *     name.opt 	Option file to be used
   *  </pre>
   *
   * It will then call the MailFaces init method to get everything prepared.
   * It will then startup the a new thread that invokes the MailFaces.run() method.
   * 
   
   
   * @param arg[]  Command line arguments
   * @return void
   */
  //------------------------------------------------------------------
  public static void main(String args[]) {

    try {

      _MailFaces = new MailFaces(); // Create Global Application object   
      Debug.setDebugOut(false);  // do not write to console by default
      // If user specified the name of the options file, then use it
      for (int i = 0; i < args.length; i++) {
        if (args[i].equals("-d")) { // check for Debug flag
          Debug.setDebugOn(true); // Get Debug output
          Debug.setDebugOut(true);  // write to console if -d
        } else if (args[i].equals("?")) { // check for help	      	
          System.out.println("Options:  -d   your.opt");
          System.exit(0);
        } else
          _MailFaces._OptionsFile = args[i];
      }
      // Initial setup
      _MailFaces.init();
      if (_MailFaces._FlagQuitMail) { // We failed our initilization    
        Debug.out(0, "Failed during initialization");
        System.exit(0);
      }

      // Startup the run() method for MailFaces 
      Thread jf = new Thread(_MailFaces, "MailFaces");
      jf.start();
      // wait for thread to end
      jf.join();
    } catch (Exception e) {
      Debug.out(e);
    }
  }

  //------------------------------------------------------------------  
} // End of MailFaces