package com.jthomas.mailfaces;

/*
 * Mailmsg.java
 * Copyright (c) 1996 John Thomas  jthomas@cruzio.com
 *      All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for non-commercial purposes
 * is hereby granted provided that this copyright notice
 * appears in all copies.
 *
 */

import java.io.*;
import java.util.*;

import javax.mail.Message;

import com.jthomas.util.Debug;

/**
 *
 * Mailmsg contains the information and methods to handle 
 * a mailmsg contained in a Mailbox object.
 * 
 * It contains the mail headers that were retrieved from the 
 * Mail server.  
 *<p>
 * The mailmsg object is what is passed from component to component until 
 * it finally arrives in the queue for the display component that then 
 * displays the icon that reprecents this email message.
 *
 * <pre>
 *---------------------------------------------------------------------
 *  Simple Usage Example to display the subject of a message.
 *
 *-  Mailmsg mailmsg = mailbox.getMailmsg(number);
 *-  System.out.println("Subject=" + mailmsg.getHeader("Subject");
 *
 * </pre>
 * @author	John Thomas
 * <a href="mailto:jthomas@cruzio.com>jthomas@cruzio.com</a>
 * @version	1.0 ( Sept 1996 )
 */



public
class Mailmsg {

//-------------------------------------
// State Variables
//-------------------------------------

/**
 * Hashtable of _Headers 
 * Key is the header name.
 */ 
private	Hashtable 	_Headers;

/**
 * Text message 
 */
protected String 		_TextContents;

/**
 * key used to store this Mail msg. 
 * This is usually the UID for the msg
 */
private	String 		_Key;

/**
 * Reference to the JavaMail Message object 
 * that was used to build this Mailmsg object.
 * If this is not a JavaMail related object then 
 * this would be null.
 */
private Message  _Message = null;

/**
 * Size of the mail msg
 */
private	int			_Size=-1;




/**
 * Message number 
 */
private int			_MsgNum;

/**
 * Set true if user asks to delete this message at the server
 */
private boolean		_ToBeDeleted=false;

/**
 * email addres to copy message to if copy or move  
 * is specified in the Select statement
 */
private String _CopyAddress = null;

/**
 * Set true if mail message should be automatically deleted.
 */
private boolean 	_KillCandidate=false;

/**
 * Image to associate with this mail
 */
private String		_ImageURL;
/**
 * Sound to associate with this mail
 */				
private String      _AudioURL;

/**
** imageurl obtained from PiconDB
*/
private	String		_PiconImageURL;

/**
 * indicator that SMS Message should be sent for this message.
 */
private  boolean   _SendSMS = false;

/**
 * indicator that SMS Message should be sent for this message
 * and the email should be copied to an alternative mailbox
 */
private  boolean   _SendSMSCopy = false;

/**
** If _BuildIcon is true, then we will use the From header to 
** construct a special icon.
*/
private boolean		_BuildIcon		= false;
/**
** _Action contains a value that indicates what action is pending
** for this mailmsg.  
*/
private int			_Action;

public static final int ActionDone=0;
public static final int ActionDelete=1;
//public static final int ActionNew=2;
//public static final int ActionOld=3;
public static final int ActionRemove=4;
public static final int ActionResolve=5;
public static final int ActionResolveAfter=6;
public static final int ActionPiconDB=7;
public static final int ActionPiconDBDefer=8;

public static final int ActionDummy=10;


/**
** _Status contains a flag that indicates if the Mailmsg still 
** exists on the server. 
*/
private int			_ServerStatus = 0;
public static final int StatusUnknown	=0;
public static final int StatusOld 		=1;
public static final int StatusNew 		=2;

/** 
** Reference to the Face object that represents this mailmsg
*/
private	Face		_Face;

/**
 *Method: Mailmsg constructor
 *
 *  This will just create the object. No work is done
 *
 */

public
Mailmsg(int msgnum) {
	_Headers = new Hashtable(40);
    _MsgNum = msgnum;
}

//-------------------------------------------------------
/**
 *Method: putHeader(String type, String value)
 *     
 *  This will save the value of the specified header into a Hashtable
 *  
 */
//-------------------------------------------------------
public synchronized  void
putHeader(String type, String value) {
  String key = null;
  
  if (type == null) {
    // special case to have3 us do the parsing
    String[] parsed = parse(value);
    key = parsed[0].trim().toUpperCase();
    value = parsed[1];
  } else { 
    key = type.trim().toUpperCase();
  }
  if (key == null || value == null) {
    Debug.out(0,"Mailmsg.putHeader: Invalid header: "+ type + "=" + value);
    return;
  }
  
	Debug.out("Mailmsg.putHeader " + key + "=" + value);
		
  String old = (String) _Headers.get(key);
  if (old != null) {
    value = old +"\n\t"+ value;
      }
	_Headers.put(key, value);
	if (key.equals("FROM")) {
	  verifyFromHeader();
	}
}

//-------------------------------------------------------
/**
 * This will return the specified Header line
 * 
 * @param The Type of Header (e.g. "from" to get the from: mail header
 * @return String HeaderLine ("" if headerline not present)
 * 
 */
//-------------------------------------------------------
 
public synchronized String getHeader(String type) {

  // 
  // The first time here, we have to build the Hashtable
  // from the array of header lines.
  //	
//  if (!_HeadersParsed && _HeaderLines != null) { // We need to parse the headers first
//    processHeaders();
//  }
    

  // 
  // get the header line from the hashtable
  //
  String value = (String) _Headers.get(type.toUpperCase());
  return (value == null) ? "" : value;
  
}

/**
 *  Build a Hashtable from the array of Header lines.
 * 
 *  The mailmsg was initial created with the Mail headers stored as 
 * an array of strings.  
 * When this method is called, we will parse the array of 
 * strings and create a Hashtable. The header line type (with ":" removed) is
 * used as the key.  For duplicate keys (e.g. multiple "cc:" lines) we 
 * will concatenate the header data  
 *   cc: joe@a.com
 *   cc: jane@b.edu
 *   >>>> key "cc"  data "joe@a.com jane@b.edu" 
 * 
 * Once the hashtable is built, we can just get the header line we want 
 * by using the "type" parameter as a key.  (Case ignored) 
 *
 * Some special handling of "From" occurs.
 *   If the email has no From header than "?@?" is returned
 *   If From has no "@" then "from@?" is returned
 *  This will allow one to Select this mail for special handling (SPAM)
 */

//private void processHeaders() {
//  String line;
//  for (int i = 0; i < _HeaderLines.length; i++) {
//    line = _HeaderLines[i];
//    Debug.out(0,"processHeaders: in="+ line);
//    if (line == null || line.length() == 0)
//      break;
//    for (int j = i + 1; j < _HeaderLines.length && _HeaderLines[j] != null; j++) {
//      if (_HeaderLines[j].length() > 0 && Character.isSpaceChar(_HeaderLines[j].charAt(0))) {
//        line = line + _HeaderLines[j];
//        i++;
//      } else
//        break;
//    }
//    //Debug.out(">>'" + line + "'");
//    String[] headerinfo = parse(line);
//    if (headerinfo[0] != null && headerinfo[0].length() > 0) {
//      headerinfo[0] = headerinfo[0].toUpperCase();
//      String old = (String) _Headers.get(headerinfo[0]);
//      if (old != null) {
//        headerinfo[1] = old + headerinfo[1];
//      }
//      putHeader(headerinfo[0], headerinfo[1]);
//      Debug.out(0,"processHeaders: out="+headerinfo[0] + ": '" + headerinfo[1] + "'");    
//
//    }
//  }
//  verifyFromHeader();
//  _HeadersParsed = true;
//}
  

/**
** Special handling of missing or invalid From: headers
**
*/
private void
verifyFromHeader() {	
	    
    
	String value = (String)_Headers.get("FROM");
  // Special handling of From:
	//Debug.out("From: "+value);
	if (value == null || value.length() == 0) 
		value = "-@-";
	else {
		String[] tokens = parseNetworkAddress(value);
		if (tokens[0].equals("-") || tokens[1].equals("-") ) {			
			String newvalue = tokens[0]+"@"+tokens[1];
			Debug.out("Mailmsg.verifyFromHeader: Replace '"+value+"' with '"+newvalue+"'"); 
			value = newvalue;
		}
	}
	Debug.out("Mailmsg.verifyFromHeader: From: "+value);
 	_Headers.put("FROM",value);
	
}


//-------------------------------------------------------------------------
/**
 * setKey
 *    Set the Key used to store this message          
 *
 */
//-------------------------------------------------------------------------
public synchronized void
setKey(String key) {
	_Key = key;
}
/**
** returns a Key that will be used to access this mail message 
**
** If the mailmsg has a _Msgid value, then it is used.
** otherwise email is checked for a "Message-ID" header
** and if found then it is used
** If neither UIDL or MessageID is present then the 
** message number is used.
**
** @return key to use to store and access this email message
*/
//---------------------------------------------------
public synchronized String
getKey() {
	
	if ( _Key == null ) {
		_Key = this.getHeader("Message-ID");
		if (_Key.length() > 0 )
			return _Key;
		else
		    return _Key = Integer.toString(_MsgNum);
		    
	} else
		return _Key;

}

//-------------------------------------------------------------------------
/**
 * setAction(int ActionCode)
 * getAction()
 *
 *    Set and query the action status of this message
 */    
//-------------------------------------------------------------------------
public void
setAction(int code) {
	_Action = code;
}

public int
getAction() {
	return _Action;
}

//-------------------------------------------------------------------------
/**
 * setServerStatus(int ActionCode)
 * getServerStatus()
 *
 *    Set and query the Server Status of this message
 */    
//-------------------------------------------------------------------------
public synchronized void
setServerStatus(int code) {
	_ServerStatus = code;
}

public synchronized int
getServerStatus() {
	return _ServerStatus;
}

//-------------------------------------------------------------------------
/**
 * setSize
 * getSize
 *    Set and query the size of the message in bytes
 *
 */
//-------------------------------------------------------------------------
public void
setSize(int size) {
	_Size = size;
}

public int
getSize() {
	return _Size;
}


//-------------------------------------------------------------------------
/**
 * setSMS
 * getSMS
 *    Set and query an indicator that an SMS message should 
 *    be sent for this email
 *
 */
//-------------------------------------------------------------------------
public void
setSMS(boolean onOff) {
	_SendSMS = onOff;
}

public boolean
getSMS() {
	return _SendSMS;
}


//-------------------------------------------------------------------------
/**
 * setSMSCopy
 * getSMSCopy
 *    Set and query an indicator that the email should be copied 
 *    also.  
 *
 */
//-------------------------------------------------------------------------
public void
setSMSCopy(boolean onOff) {
	_SendSMS = onOff;
}

public boolean
getSMSCopy() {
	return _SendSMS;
}

//-------------------------------------------------------------------------
/**
 * getMsgNum
 *    Get the message number for this email          
 *
 */
//-------------------------------------------------------------------------
public synchronized int
getMsgNum() {
	return _MsgNum;
}

//-------------------------------------------------------------------------
/**
 * get/setMessage
 *    Set the reference to the JavaMail Message object          
 *    This will be null if JavaMail is not used as the 
 *    FaceMailbox implementation.
 */
//-------------------------------------------------------------------------
public synchronized void
setMessage(Message jMessage) {
	_Message = jMessage;
}

//-------------------------------------------------------------------------
public synchronized Message
getMessage() {
	return _Message;
}

//-------------------------------------------------------------------------
/**
 * get/setText
 *    Get and set the contents of the Mail message         
 *
 */
//-------------------------------------------------------------------------
public synchronized void
setText(String text) {
	_TextContents = text;
}

//-------------------------------------------------------------------------
public synchronized String
getText() {
	return _TextContents;
}



//-------------------------------------------------------------------------
/** 
 * getHeaders
 *    return the complete set of Headers 
 *    from the Hashtable that results from 
 *    parsing the array of header lines          
 */    
//-------------------------------------------------------------------------
public synchronized String[]

getHeaders() {
	String[] headers = new String[_Headers.size()];
	Enumeration keys = _Headers.keys();
	Enumeration elements = _Headers.elements();
	int i = 0;
	while(keys.hasMoreElements()) {
	  String key = (String)keys.nextElement();
	  String element = (String)elements.nextElement();
	  headers[i] = key+": "+element;
	  i++;
	}
	return headers;
}


//-------------------------------------------------------------------------
/*
 * 
 *    Set a flag _ToBeCopied stating that this message should be copied          
 * toBeDeleted()
 *    Report the setting of the _ToBeDeleted flag.
 */
//-------------------------------------------------------------------------
public synchronized void
setCopyAddress(String address) {
	_CopyAddress= address;
}

public synchronized String
getCopyAddress() {
	return _CopyAddress;
}


//-------------------------------------------------------------------------
/*
 * delete
 *    Set a flag _ToBeDeleted stating that this message should be deleted          
 * toBeDeleted()
 *    Report the setting of the _ToBeDeleted flag.
 */
//-------------------------------------------------------------------------
public synchronized void
setDeleteFlag(boolean flag) {
	_ToBeDeleted = flag;
}

public synchronized boolean
getDeleteFlag() {
	return _ToBeDeleted;
}

//-------------------------------------------------------------------------
/**
 * setKillCandidate
 *    Set a flag that this mail message was match by a KillSelect statement
 * getKillCandidate
 *    Report the setting of the KillCandicate flag.
 */
//-------------------------------------------------------------------------
public synchronized void
setKillCandidate(boolean flag) {
	_KillCandidate = flag;
}

public synchronized boolean
getKillCandidate() {
	return _KillCandidate;
}


//-------------------------------------------------------------------------
/**
 * setImageURL setAudioURL
 *    Set the URL for an image to associate with this mailmsg
 * getImageURL getAudioURL
 *    Report the URL for the image or audio file
 */
//-------------------------------------------------------------------------
public void
setImageURL(String imageURL) {
	_ImageURL = imageURL;
}

public String
getImageURL() {
	return _ImageURL;
}
//-------------------------------------------------------------------------
/**
 * setPiconImageURL setAudioURL
 *    Set the URL for an image to associate with this mailmsg
 * getPiconImageURL getAudioURL
 *    Report the URL for the image or audio file
 */
//-------------------------------------------------------------------------
public void
setPiconImageURL(String imageURL) {
	_ImageURL = imageURL;
}

public String
getPiconImageURL() {
	return _ImageURL;
}

public void
setAudioURL(String audioURL) {
	_AudioURL = audioURL;
}

public String
getAudioURL() {
	return _AudioURL;
}

//----------------------------------------------------------------------
/**
 *Method: setFace
 *Method: getFace
 *
 */
//---------------------------------------------------------------------- 
public  void
setFace(Face face) {
	_Face = face;
}

public  Face
getFace() 			{
	return _Face;
}


//----------------------------------------------------------------------
/**
** setBuildIcon
** isBuildIcon
**
** set/get methods for flag that controls whether we build a 
** special icon using the From: header.
*/
//---------------------------------------------------------------------- 
public  void
setBuildIcon(boolean setting) {
	_BuildIcon = setting;
}

public boolean
isBuildIcon() {
	return _BuildIcon;
}

//-----------------------------------------------------------------------
/**
 * Parse(String line) -> returns String[]
 *    
 * Parse a Mail Header line.
 *  
 *  
 * It will return an array of strings. The first string is the name of the header line (i.e. Subject)
 * the 2nd entry is the value.
 *   Subject: A new subject
 *    return[0] = "Subject"
 *    return[1] = "A new subject"
 */
//----------------------------------------------------------------------------
private String[]
parse(String headerline) {
    String[] tokens = null;
    
	tokens = new String[3];
 	tokens[0] = "";
 	tokens[1] = "";

 	if (headerline != null) {
 		
        int colonpos = headerline.indexOf(':');
        if (colonpos > 0) {
          tokens[0] = headerline.substring(0,colonpos); //header type without :
          tokens[1] = headerline.substring(colonpos+1).trim();   //value
        }
 	}
  	//Debug.out("Tokens=" + tokens[0] + ":" + tokens[1]);	
	return tokens;
}

//-----------------------------------------------------------------------
/**
 * Static class method to parse a Network mail address and return
 * an array containing {userid,host,fullname}
 *
 *   0:  Userid
 *   1:  Host
 *   2:  Name (if present)
 *    Examples:
 *     John Q Public <jpublic@netcom.com>  -> [jpublic] [netcom.com] [John Q. Public]
 *     jpublic@netcom.com (John Q Public)  -> [jpublic] [netcom.com] [John Q. Public]
  *@return It will return an array of strings.
 */
//----------------------------------------------------------------------- 
public static String[]
parseNetworkAddress(String address) {
    String[] tokens = { "-", "-", "?"};
    int start;
    int end;
    boolean failed = false;
    String netaddress = "";
	if (address == null) 
		return tokens;
	else { 	
 		end = address.lastIndexOf('>');
 		if ( end != -1 ) {
 			// parse "name <user@host>" style
 			start = address.lastIndexOf('<',end);
 			if (start != -1) {
 				tokens[2] = address.substring(0,start).trim();
 				netaddress = address.substring(start+1,end).trim();
 			} else {
 				failed = true;
 			}
 		} else {
 			end = address.lastIndexOf(')');
 			if ( end != -1 ) {
 				// parse "user@host (name)" style
 				start = address.lastIndexOf('(',end);
 				if (start != -1) {
 					tokens[2] = address.substring(start+1,end).trim();
 					netaddress = address.substring(0,start).trim();
 				}
 			} else {
 				end = address.lastIndexOf('@');
 				if ( end != -1) {
 					// Parse a simple "Userid @ host" style 					
 					netaddress = address.trim();
 				} else {
 					failed = true;
 				}
 			}
 		}

 		if ( ! failed ) {
 			end = netaddress.indexOf('@');
 			if (end == -1)
 			  tokens[0] = netaddress.trim();
 			else {
 				if (end == 0) 
 					tokens[1] = netaddress.substring(1).trim();

 				else {
 					tokens[0] = netaddress.substring(0,end).trim();
 					tokens[1] = netaddress.substring(end+1).trim();
 				}
 			}	
 		}
 	}
 	//Debug.out("0:" + tokens[0] + " 1:" + tokens[1] + " 2:" + tokens[2]);
 	return tokens;	
}

//-----------------------------------------------------------------------
/**
 * toString method to display the idendification of a 
 * MailMsg object.
 * 
 * @return It will return a string representing the Mailmsg.
 */
//----------------------------------------------------------------------- 
public String
toString() {

  StringBuffer sb = new StringBuffer(256);
  sb.append("Mailmsg: ");
  sb.append(this.getKey());
  sb.append(" From: ");
  sb.append(this.getHeader("from"));
  sb.append(" Subject: ");
  sb.append(this.getHeader("subject"));
  return sb.toString();
}

//-------------------------------------------------------
} // end of Class Mailmsg