package com.jthomas.mailfaces;

/*
 * PiconDB.java
 * Copyright (c) 1996 John Thomas  jthomas@cruzio.com
 *      All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for non-commercial purposes
 * is hereby granted provided that this copyright notice
 * appears in all copies.
 *
 */

import java.io.*;
import java.util.*;
import java.net.*;

import com.jthomas.util.Debug;
import com.jthomas.util.Queue;
import com.jthomas.util.Log;
/**
 * Interface to the PICON Database(s).
 * <pre>
 * Operation:<br>
 *   The PiconDB constructor will create a PiconDBServer object for each
 *   PIconDBServer statement in the Opt file. 
 *   It will also allocate a queue for search requests.
 *
 *   The MailFaces main routine will create a thread for the Picondb object
 *   and issue the picondbobj.start().
 *
 *   The run method will be given control. It will take requests from the
 *   queue (in the form of mailmsgs) and issue a search command for the
 *   appropriate image to each of the PiconDBServer Objects.
 *   
 * 
 * Based on Steve Kinzler's Picon Archive<br> 
 *   <a href="http://www.cs.indiana.edu/picons/ftp/index.html">Picons Archive</a>
 * </pre>
 * @author	John Thomas
 * <a href="mailto:jthomas@cruzio.com">jthomas@cruzio.com</a>
 * @version	2.0 ( May 1998 )
 */

public class PiconDB implements Runnable {

  //-------------------
  // Class variables
  //-------------------
  //-------------------
  // instance variables
  //-------------------

  /**
  ** List of PICONDB servers
  */
  private PiconDBServer[] _DBServerList;
  private int _DBServerCount;

  /**
  ** List of icon results that we will defer using 
  ** until after we call SelectAfter
  */
  private String[] _PiconDeferList = null;
  /**
  ** List of icon results that we will ignore 
  */
  private String[] _PiconIgnoreList = null;

  private MailFaces _MailFaces;
  private FaceOptions _Options;
  private FaceResolver _Resolver;
  private FaceBoard _FaceBoard;
  private FaceCache _Cache = null;

  private boolean _PIconDBDomainOnly = false;

  // Define a Queue that we will get work (Mailmsgs) from
  private Queue _Queue;

  //----------------------------------------------
  /**
   * PiconDB()  - PiconDB constructor.
   * <p>
   *  This will create a PiconDBServer object for each PIconDB
   *  statement that it finds in the Options file.
   *  The singleton PiconDB object will be created by MailFaces 
   *  during initialization
   *  
   */
  //----------------------------------------------
  public PiconDB() {
    Debug.out("PiconDB constructor");
    _MailFaces = MailFaces.getRef();
    _Options = FaceOptions.getRef();

    // Process the Options
    refreshOptions();
    // ask to be informed if the Options get changed
    _MailFaces.addFaceEventListener(new FaceEventListener() {
      public void actionPerformed(FaceEvent e) {
        if (e.getID() == FaceEvent.ID_OPTIONS_CHANGED)
          PiconDB.this.refreshOptions();
      }
    });

    // Create a Queue that we will get work from. 
    _Queue = new Queue(100);

    // Create a list of PiconDB server objects
    String[] list;
    list = _Options.getList("PIconDB"); // get list of PIconDBs

    //
    // handle default if no PIconDB statements at all.
    //
    if (list.length == 0) {
      Debug.out("No PiconDB Servers found. Use defaults");
      _DBServerList = new PiconDBServer[1];
      _DBServerList[0] = new PiconDBServer();
      _DBServerCount = 1;
      return;
    }

    // 
    // process 1 or mor PIconDB statements
    //  

    _DBServerCount = list.length;

    _DBServerList = new PiconDBServer[_DBServerCount];
    for (int i = 0; i < _DBServerCount; i++) {
      // Parse the list entry
      String stmt = list[i];
      StringTokenizer st = new StringTokenizer(stmt, " ");
      int n = st.countTokens();
      String[] tokens = new String[n];
      for (int j = 0; st.hasMoreTokens(); j++) {
        tokens[j] = st.nextToken().trim();
        Debug.out(j + ": " + tokens[j]); 
      }     
      //
      // PIconDB "http://www.cs.indiana.edu:800/piconsearch" "users+domains" 
      //
      if (n >= 2) {
        _DBServerList[i] = new PiconDBServer(tokens[0], tokens[1]);
        if (n == 2)
          continue;
      } else {
        MsgDialog.create("PIconDB:  Invalid statement. " + stmt);
        continue;
      }
      //
      //PIconDB "http://www.cs.indiana.edu:800/piconsearch" "users+domains" PROXY localhost 4480
      //
      if (n >= 5) {
        int port = 0;
        if (tokens[2].toUpperCase().equals("PROXY")) {
          try {
            port = Integer.parseInt(tokens[4]);
          } catch (Exception e) {
            MsgDialog.create("PIconDB option - Proxy port was invalid.");
          }
          _DBServerList[i] = new PiconDBServer(tokens[0], tokens[1], tokens[3], port);
          if (n == 5) 
            continue;
        } else {
          MsgDialog.create("PIconDB: - Invalid Proxy statement. " + stmt);
        }
      } else {
        MsgDialog.create("PIconDB:  Invalid statement. " + stmt);
        continue;
      }
      //
      //PIconDB "url" "users+domains" PROXY localhost 8080 username password
      //
      if (n == 7) {
        _DBServerList[i].setProxyUser(tokens[5], tokens[6]);

      } else {
        MsgDialog.create("PIconDB:  Invalid statement. " + stmt);
        continue;
      }
    
    } // for

  }  // end constructor

  /**
   * queuePut will place an object in the PiconDB work queue.
   *
   */
  public void queuePut(Object obj) {
    _Queue.put(obj);
  }

  //-----------------------------------------------------
  /**
   *Method: refreshOptions()
   *@return void
   *
   *   Invoked when user refreshes the Option file
   */
  //-----------------------------------------------------

  public void refreshOptions() {

    //  We should refresh the PIcondb options
    _PIconDBDomainOnly = _Options.getBoolean("PIconDBDomainOnly", true);

    _PiconDeferList = _Options.getList("PiconDefer");
    _PiconIgnoreList = _Options.getList("PiconIgnore");
    // currently we are not refreshing the server address.  
    // Maybe we should.  ???????   
  }

  //-----------------------------------------
  /**
   *   getList will return the list of PiconDBServer objects.
   *    
   */
  //------------------------------------------
  public PiconDBServer[] getList() {
    return _DBServerList;
  }

  //-----------------------------------------
  /**
   *   getCount is a Class method that will return the count of PiconDB objects.
   */
  //------------------------------------------
  public int getCount() {
    return _DBServerCount;
  }

  //-----------------------------------------
  /**
   *   run() will get control when the Thread is started
   */
  //------------------------------------------
  public void run() {
    Debug.out("PiconDB.run()");

    _FaceBoard = _MailFaces.getFaceBoardRef();
    _Cache = _MailFaces.getCacheRef();
    _Resolver = _MailFaces.getResolverRef();
    while (true) {
      try {

        Mailmsg mailmsg = (Mailmsg) _Queue.get();
        if (mailmsg == null) {
          Debug.out(0,"Null returned from Queue.  something is broken"); 	
          return;   // treat a null as a reason to quit.
        }
        Debug.out("PiconDB.run() dequeued mailmsg " + mailmsg.getKey());
        //---------------------------
        // Figure out which image to display for this mail message
        //---------------------------
        boolean resolved = false;

        resolved = resolvePiconDB(mailmsg, true, _PIconDBDomainOnly);
        if (_MailFaces.getFlagQuitMail())
          break;

        if (!resolved) {
          mailmsg.setAction(Mailmsg.ActionResolveAfter);
          Debug.out(
            "PiconDB.run() queued mailmsg " + mailmsg.getKey() + " for ResolveAfter");
          _Resolver.queuePut(mailmsg); // Requeue to Resolver for further searching

        }

      } catch (Exception e) {
        Debug.out("PiconDB: Exception in run(); ", e);
        //Debug.asserted(false);
        // So we are forced to know about this!! remove at release ???		

      } catch (Throwable e) {
        Debug.out("PiconDB Error", e);
        //Debug.asserted(false);
        // So we are forced to know about this!! remove at release ???		
      }
    }

  }

  //--------------------------------------------------------------------
  /**
   * resolvePiconDB() will handle issueing the search request to one or more
   * PiconDBServers.  
   *
   * Note that this can be called directly from FaceResolver.
   *
   *@return true if the mailmsg is resolved into an appropriate Icon
   */
  //----------------------------------------------------------------------
  public boolean resolvePiconDB(
    Mailmsg mailmsg,
    boolean usecache,
    boolean domainonly) {
    boolean exceptionflag = false;
    String imageurl = null;
    StringBuffer logmsg = new StringBuffer(256);

    // get the user and host from the mail
    //Debug.out("  From: " + mailmsg.getHeader("from"));
    String from = mailmsg.getHeader("from");
    logmsg.append("Mail from: "+from+";");
    if (from.length() == 0) {
      Debug.out("PiconDB: Invalid mailmsg: No From: Header. Cannot use PIconDB");
      Log.out("PiconDB",logmsg.toString());
      return false;
    }
    
    String[] address = mailmsg.parseNetworkAddress(from);
    String user = address[0];
    String host = address[1];
    String replyto = mailmsg.getHeader("reply-to");

    if (replyto.length() > 0) {
      address = mailmsg.parseNetworkAddress(replyto);
      if (!user.equalsIgnoreCase(address[0]))
        user = address[0] + "+" + user;
      if (!host.equalsIgnoreCase(address[1]))
        host = address[1] + "+" + host;
    }
    Debug.out("PiconDB: " + user + " @ " + host);

    // Check if cached image for this user@host and use it if found
    if (usecache && _Cache != null) {
      imageurl = _Cache.getUser(user, host);
      if (imageurl == null || imageurl.length() == 0)
        if (domainonly) {
          imageurl = _Cache.getHost(host);
          Debug.out("Cache Found " + imageurl + " for host " + host);
        } else {
          Debug.out("Cache Found " + imageurl + " for " + user + "@" + host);
        }
      // if Cache returned "" then we know PIcondb does not have it.  
      if (imageurl != null && imageurl.length() == 0) {
        logmsg.append(" Cache returns null.");
        Log.out("PiconDB",logmsg.toString());
        return false;
      }
    }

    //
    // if not already discovered her URL in the Cache
    // Then search the PICON DB servers
    if (imageurl == null) {

      String searchfor = domainonly ? host : user + "@" + host;
      String searchUser = user;
      FaceStatusLabel.setStatus("PiconDB", "Search for " + searchfor);
      if (domainonly)
        searchUser = "unknown";
      for (int i = 0; i < _DBServerList.length && imageurl == null; i++) {
        try {
          imageurl = _DBServerList[i].search(host, searchUser);
          exceptionflag = true;
          if (imageurl != null) 
            break;
        } catch (IOException e) {
          exceptionflag = true;
          imageurl = null;
        }
      }
      FaceStatusLabel.clearStatus("PiconDB"); // turn off the status msg

      if (imageurl != null && (mailmsg.getAction() == Mailmsg.ActionPiconDB)) {
        if (checkDeferList(imageurl, host)) { // if host icon in PiconDefer	 
          Debug.out("checkDeferList returned true for " + imageurl);
          mailmsg.setPiconImageURL(imageurl); // save url in case we need it later 
          imageurl = null;
        } else if (checkIgnoreList(imageurl, host)) { // if host icon in PiconIgnoreD	 
          Debug.out("checkIgnoreList returned true for " + imageurl);
          imageurl = null;
        }

      }
    }
    logmsg.append("imageurl: "+imageurl);
    Log.out("PiconDB",logmsg.toString());
    if (imageurl != null) {
      processURL(mailmsg, imageurl, user, host, usecache);
      // this will have spawned another task that will cache the URL and queue  it for  display
      return true;
    } else {
      // imageurl will be null if search for her userid@node failed.
      // Cache this fact so we will not search for her again
      if (_Cache != null && !exceptionflag) {
        _Cache.cacheImage(null, user, host); // Cache the lack of url

      }
      return false;
    }
  }

  /**
  ** This rtn is called by the ResolvePiconDB method when it has located an
  ** icon in the database and returned the url pointing to the image. 
  ** 
  ** This rtn will check the cache and see if the actual icon for that URL 
  ** has already be placed in the cache and if so it will place a pointer to the 
  ** cached icon into the Mailmsg and queue the mailmsg to the FaceBoard for 
  ** display.
  **
  ** If it is not already in the cache, then it will start up a thread (using and
  ** anonymous inner class) that will cache the image and place a pointer to the 
  ** cached image into the mail msg and queue it to FaceBoard.  This is done because  
  ** getting the icon from the database is the slowest part of the whold process and 
  ** we don't want to wait for it.  
  ** Note that this means that we have 1 thread that calls the Picon search and 
  ** a new tread for each icon that is read.  
  */
  private void processURL(
    Mailmsg mailmsg,
    String imageurl,
    String user,
    String host,
    boolean usecache) {
    if (_Cache == null || imageurl.startsWith("file:")) {
      // place the url into the mailmsg	
      mailmsg.setImageURL(imageurl);
      // Queue the mailmsg for display
      Debug.out("Use " + imageurl + " for " + user + "@" + host);
      Debug.out("PiconDB queued mailmsg " + mailmsg.getKey() + " for FaceBoard");
      _FaceBoard.queuePut(mailmsg);
      return;
    }

    if (usecache) {
      String newurl;
      newurl = _Cache.getURL(imageurl);
      if (newurl != null) {
        // if already cached then add entry for this user/host  
        _Cache.indexImage(newurl, user, host, newurl);
        // place the url into the mailmsg	
        mailmsg.setImageURL(newurl);
        // Queue the mailmsg for display
        Debug.out("Use " + imageurl + " for " + user + "@" + host);
        Debug.out("PiconDB queued mailmsg " + mailmsg.getKey() + " for FaceBoard");
        _FaceBoard.queuePut(mailmsg);
        return;
      }
    }

    // create some final variable that can be referenced inside the inner class
    final String f_imageurl = imageurl;
    final String f_user = user;
    final String f_host = host;
    final Mailmsg f_mailmsg = mailmsg;
    // We found a URL so now cache the results using new thread
    Thread inner = new Thread(new Runnable() {
      public void run() {
          // create a new cached image.
        Debug.out("Cache image: " + f_imageurl);
        String cacheURL = _Cache.cacheImage(f_imageurl, f_user, f_host);
        if (cacheURL != null)
          // place the url into the mailmsg	
          f_mailmsg.setImageURL(cacheURL);
        else
          f_mailmsg.setImageURL(f_imageurl);
        // Queue the mailmsg for display
        Debug.out("Use " + cacheURL + " for " + f_user + "@" + f_host);
        Debug.out("PiconDB queued mailmsg " + f_mailmsg.getKey() + " for FaceBoard");
        _FaceBoard.queuePut(f_mailmsg);

      }
    }, "CacheURLThread");
    inner.start(); // start the thread. 

  }

  //--------------------------------------------------------------------
  /**
  ** This will check the URL returned from the PICONDB server 
  ** against the List of "PiconDefer" items specified by the user.
  ** It will return true if the URL that that PiconDB returned from 
  ** the database is in the PiconDefer list.
  **  
  */
  //----------------------------------------------------------------------
  public boolean checkDeferList(String imageurl, String host) {
    if (_PiconDeferList.length > 0) {
      String urlhost = getHostFromPiconURL(imageurl, host);
      if (urlhost == null)
        return false;
      for (int i = 0; i < _PiconDeferList.length; i++)
        if (urlhost.equalsIgnoreCase(_PiconDeferList[i]))
          return true;
    }
    return false;
  }
  //--------------------------------------------------------------------
  /**
  ** This will check the URL returned from the PICONDB server 
  ** against the List of "PiconIgnore" items specified by the user.
  ** It will return true if the URL that that PiconDB returned from 
  ** the database is in the PiconIgnore list.
  **  
  */
  //----------------------------------------------------------------------
  public boolean checkIgnoreList(String imageurl, String host) {
    if (_PiconIgnoreList.length > 0) {
      String urlhost = getHostFromPiconURL(imageurl, host);
      if (urlhost == null)
        return false;
      for (int i = 0; i < _PiconIgnoreList.length; i++)
        if (urlhost.equalsIgnoreCase(_PiconIgnoreList[i]))
          return true;
    }
    return false;
  }
  //--------------------------------------------------------------------
  /**
  ** This will parse the URL returned from the PICONDB server 
  ** and pick out the host name for the icon that is being returned.
  **
  ** For example, if the Picon server returned 
  **  http://www.cs.indiana.edu/picons/db/domains/com/sun/unknown/face.gif 
  ** and we were searching for "system.sun.com" 
  ** then this rtn would return
  **   sun.com
  **
  ** @param the URL returned from Picon database
  ** @param the host name we were searching for
  ** @return the host name for the icon
  */
  //----------------------------------------------------------------------

  private String getHostFromPiconURL(String imageurl, String hostInput) {

    String list;
    Debug.out("getHostFromPiconURL: " + imageurl + " " + hostInput);
    int x = imageurl.indexOf("//");
    int y = imageurl.lastIndexOf("/unknown/");
    // change "cruzio.com+pobox.com" to ".cruzio.com.pobox.com"
    String searcharg = "." + hostInput.replace('+', '.');

    if (x > 0 && y > x) {
      list = imageurl.substring(x + 9, y);

      StringTokenizer t = new StringTokenizer(list, "/");
      int n = t.countTokens();
      String[] tokens = new String[n];
      for (int i = 0; i < n; i++)
        tokens[n - i - 1] = t.nextToken();
      // now we have an array of host qualifiers like {"sun","com","domains","db","picons"}

      String host = "";
      for (int i = 0; i < n; i++) {
        String check = "." + tokens[i];
        if (searcharg.indexOf(check) >= 0) {
          if (host.length() > 0)
            host = host + ".";
          host = host + tokens[i];
        }
      }
      Debug.out(host);
      return host;
    } else
      return null;
  }

  //-----------------------------------------------------
  // Adding this makes this class self-testing
  //-----------------------------------------------------
  public static void main(String args[]) {

    Debug.setDebugOn(true);
    if (args.length < 2) {
      Debug.out("syntax: java PiconDB hostname userid");
      System.exit(0);
    }

    PiconDBServer db = new PiconDBServer();
    try {
      String url = db.search(args[0], args[1]);
      Debug.out("URL=" + url);
    } catch (IOException e) {
      Debug.out(0,e);
    }

    System.exit(0);
  }

  //-------------------------------------------------------
} // end of Class PiconDB