package com.jthomas.mailfaces;

/*
 * PiconDB.java
 * Copyright (c) 1996 John Thomas  jthomas@cruzio.com
 *      All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for non-commercial purposes
 * is hereby granted provided that this copyright notice
 * appears in all copies.
 *
 */

import java.io.*;
import java.util.*;
import java.net.*;

import com.jthomas.util.Debug;
import com.jthomas.util.Queue;
import com.jthomas.util.Log;

//-----------------------------------------------------------------------
// PiconDBServer Class
//-----------------------------------------------------------------------
/**
 * Maintains the Interface to an instance of a PICON Database.
 * <pre>
 *
 * </pre>
 */
//------------------------------------------------------------------------

public class PiconDBServer {
  
  /**
   * timeout value for URLConnection connect and read (ms)
   */
  private static final String  _TIMEOUT = "30000"; 
  //----------------------------------
  // State Variables
  //---------------------------------
  /**
   * URL to access the PBIcon server
   */
  private String _DbUrl;
  
  /**
   * The string that tells the server if both domains and users
   * are to be searched or just the domains.
   */
  private String _DbTypes;
  /**
   * Optional Proxy host
   */
  private String _DbProxy;
  /**
   * Optional Proxy port
   */
  private int _DbPort;
  
  /** 
   * Optional Proxy userid
   */
  private String _DbProxyUser;
  /** 
   * Optional Proxy userid password
   */
  private String _DbProxyPassword;
  /** 
   * Host for the PBIcon server
   */
  private String _DbHost = null;

  /**
   * set true when a failure is detected on the connection.
   */
  private boolean _ConnectionDown = false;
  
  /**
   * time to wait before retrying the connection failure.
   */
  private long  _ConnectionRetryTime =  120*1000;  
  private long  _ConnectionDownTime =  0; 
  //----------------------------------------------
  /**
   * PiconDBServer constructor.
   * <p>
   *  This will just create the object with the default parameters 
   */
  //----------------------------------------------
  public PiconDBServer() {

    _DbUrl = "http://www.cs.indiana.edu:800/piconsearch";
    _DbTypes = "users+domains";
    Debug.out("PicondbServer() - Default to " + _DbUrl + " , " + _DbTypes + " )");
  }
  //----------------------------------------------
  /**
   *Method: PiconDBServer(String url, String types)  - PiconDBServer constructor
   *@param String url		- The URL string that will be used to access the CGI program
   *@param String dbtype  - The names of the PIcon type collections.  (users+domains) 
   * <p>
   *  This will just create the object with the specified values 
   */
  //----------------------------------------------
  public PiconDBServer(String url, String types) {
    _DbUrl = url;
    _DbTypes = types;
    Debug.out("PicondbServer() - Settings: " + _DbUrl + " , " + _DbTypes);
  }

  //----------------------------------------------
  /**
   *Method: PiconDBServer(String url, String types, String proxy, String port)  - PiconDB constructor
   *@param String url		- The URL string that will be used to access the CGI program
   *@param String dbtype  - The names of the PIcon type collections.  (users+domains)
   *@param String proxy   - Host name for the HTTP Proxy system
   *@param String port    - Network port used for Proxy
   * <p>
   *  This will just create a PiconDBServer object that will access the database via a
   *  HTTP Proxy gateway.  
   */
  //----------------------------------------------
  public PiconDBServer(String url, String types, String proxy, int port) {
    _DbUrl = url;
    _DbTypes = types;
    _DbProxy = proxy;
    _DbPort = port;
    Debug.out(
      "PicondbServer() - Settings: "
        + _DbUrl
        + " , "
        + _DbTypes
        + " Proxy: "
        + _DbProxy
        + ":"
        + _DbPort);
        
  }
  
  //----------------------------------------------------------
  /**
   *  setProxyUser will setup the username and password for the 
   *  Proxy if the user specified them.
   *
   *  @param user     The Username required by the Proxy
   *  @param password The password required by the Proxy
   */
  //----------------------------------------------------------
  public void setProxyUser(String user, String password) {
     _DbProxyUser = user;
     _DbProxyPassword = password; 
  }


  //----------------------------------------------------------
  /**
   *  getHost() returns the hostname for this PIconDB.
   *
   *  Needed so we can retrieve the image using 
   *  any proxy that may be required.
   */
  //----------------------------------------------------------
  public String getHost() {
    if (_DbHost == null) {
      try {
        _DbHost = (new URL(_DbUrl)).getHost();
      } catch (Exception e) {
      }
      Debug.out("PicondDB.getHost():" + _DbHost);
    }

    return _DbHost;
  }

  //----------------------------------------------------------
  /**
   *  getProxy() returns the Proxy hostname for this PIconDB.
   *
   */
  //----------------------------------------------------------
  public String getProxy() {
    return _DbProxy;
  }

  //----------------------------------------------------------
  /**
   *  getPort() returns the Proxy hostname for this PIconDB.
   *
   */
  //----------------------------------------------------------
  public int getPort() {
    return _DbPort;
  }

  //----------------------------------------------------------
  /**
   * search(String mailhost, String mailuserid) will invoke a query to the Picon Database.
   *
   * <p>
   *
   *  Input is the mailhost name for the user and the userid
   *  A search of the remote database will be invoked and the results 
   *  will be parsed to determine the URL of the GIF for the best 
   *  match.
   *
   *@param String mailhost  - 1 or most hostnames separated by +
   *@param String mailuser  - 1 or more userids separated by +
   *@return String url  - Returns the URL needed to access the image or null
   *@exception IOException
   */
  //----------------------------------------------------------

  public String search(String mailhost, String mailuserid) throws IOException {

    String urlstr;
    
    // if recent connection failure 
    if (_ConnectionDown) {
      long now = new Date().getTime();
      if (( now - _ConnectionDownTime) < _ConnectionRetryTime) {
        Debug.out(0,"PIconDBServer: No connection for " + mailuserid + " @ " + mailhost);
        
        throw (new IOException("PIconDB connection is down"));
      } else {
        _ConnectionDown = false;   // try again
      }
    }
    urlstr = _DbUrl + "/" + mailhost + "/" + mailuserid + "/" + _DbTypes + "/up/on";
    if (Debug.ON) Debug.out("PIconDB Search URL='" + urlstr + "'");

    // The input stream is the answer from piconsearch.
    // It will be placed in "buffer" as a single long String
    String data;
    StringBuffer buffer;
    try {
      InputStream istream = getInputStream(urlstr);
      if (istream == null) {
        Debug.out(0,"PIcondDBServer: Failure: Failed in getInputStream");
        throw (new IOException("Failed in getInputStream"));
      }

      BufferedReader url_data = new BufferedReader(new InputStreamReader(istream));

      buffer = new StringBuffer();

      while ((data = url_data.readLine()) != null) {
        // Now we have a line of data, add it to the buffer		
        buffer.append(data);
      }
    } catch (IOException e) {
      Debug.out(0, "PIcondDBServer: Failure: " + e.getMessage());
      Debug.out(0, "PIconDB Search for " + mailuserid + " @ " + mailhost);
      Log.out(
        "PiconDBServer",
        "Failure: "
          + "Search for "
          + mailuserid
          + " @ "
          + mailhost
          + "; "
          + e.getMessage());
      _ConnectionDown = true;
      _ConnectionDownTime = new Date().getTime(); 
      throw e;
    }
    //--------------------------------------------
    // The data returned from picon search looks like the following:
    //
    //-> <html><head>
    //-> <title>Picons Search - Results</title>
    //-> </head><body>
    //-> 
    //-> <h1><a href="http://www.cs.indiana.edu/picons/search.html">Picons Search</a> Results</h1>
    //-> 
    //-> <strong>user</strong> = <kbd>adams</kbd><br>
    //-> <strong>host</strong> = <kbd>ibeam.intel.com</kbd><br>
    //-> <strong>db</strong> = <kbd>users usenix domains</kbd><br>
    //-> <strong>type</strong> = <kbd>up</kbd><br>
    //-> <strong>show</strong> = <kbd>off</kbd><br>
    //-> <strong>limit</strong> = <kbd>10</kbd><br>
    //-> <p>
    //-> 
    //-> users/com/intel/adams/face.<a
    //-> href="http://www.cs.indiana.edu/picons/db/users/com/intel/adams/face.xbm">xbm</a><br>
    //-> domains/com/intel/unknown/face.{<a
    //-> href="http://www.cs.indiana.edu/picons/db/domains/com/intel/unknown/face.gif">gif</a>,<a
    //-> href="http://www.cs.indiana.edu/picons/db/domains/com/intel/unknown/face.xbm">xbm</a>,<a
    //-> href="http://www.cs.indiana.edu/picons/db/domains/com/intel/unknown/face.xpm">xpm</a>}<br>
    //-> 
    //-> </body></html>

    //------------------------------------------------------------
    String bufstr = buffer.toString();
    //Debug.out("PiconDB returned:"+bufstr.length()+"\n"+bufstr);
    if (bufstr.indexOf("Picons Search - Results") == -1) {
      Debug.out(0,"Unexpected Message from PiconSearch");
      Debug.out(0,"PiconDB returned:" + bufstr.length() + "\n" + bufstr);
      return null;
    } else {

      // Now parse the data and try to extract the URL of the picon
      // first locate reference to gif image or xbm image
      int end;
      int gif = bufstr.indexOf(".gif\">");
      int xbm = bufstr.indexOf(".xbm\">");
      if (gif == -1) // No gif image.
        if (xbm == -1) { // No image
          Debug.out("PiconDB returned:" + bufstr.length() + "\n" + bufstr);	
          Debug.out("PiconDBServer.search: No image found");
          return null;
        } else
          end = xbm; // Use the xbm image
      else { // We have a Gif image
        end = gif; // assume that we want the gif
        if (xbm >= 0 && xbm < gif) // if xbm image comes first
          end = xbm; // Use the xmb image
      }

      end += 4; // advance past the ".xbm" or ".gif"
      //Debug.out("end=" + end);

      int start = bufstr.lastIndexOf("href=\"", end);
      start += 6; // advance past 'href="'
      //Debug.out("start=" + start);

      String gifurl = bufstr.substring(start, end);

      if (Debug.ON) Debug.out("PiconDB returned: gifurl=" + gifurl);
      return gifurl;
    }

  }

  /**
   * getInputStream will return an input stream for the specified URL.
   *
   * <p>
   * This method will take care of making a connection to the specified URL
   * (taking care to use a Proxy if required) and return an 
   * InputStream that will access the contents of the URL
   * <p>
   *@param  urlstr  Specifies the URL that should be connected to.
   *@return InputStream
   *@exception IOException MalformedURLException
   */
  public InputStream getInputStream(String urlstr)
    throws IOException, MalformedURLException {

    URL url = null;
    URLConnection urlcon = null;
    InputStream istream = null;

    try {
      if (_DbProxy != null) {
        if (Debug.ON) Debug.out("Access via " + _DbProxy + ":" + _DbPort);
        Properties prop = System.getProperties();
        prop.put("http.proxyHost",_DbProxy);
        prop.put("http.proxyPort",Integer.toString(_DbPort));
      }
      
      //
      // set default Timeouts.
      //  This will only be used in JDK 1.4
      //
      try {
        System.setProperty( "sun.net.client.defaultConnectTimeout", _TIMEOUT);
        System.setProperty( "sun.net.client.defaultReadTimeout", _TIMEOUT );
      } catch (Throwable thrown) {
      	Debug.out("PIconDBServer.getInputStream: System.setProperty not supported on Java 1.1");
      }
  
      url = new URL(urlstr);
      
      if (Debug.ON) Debug.out("PiconDBServer: url=" + url);
      urlcon = url.openConnection();
      if (Debug.ON) Debug.out("PiconDBServer: urlcon=" + urlcon);
      if (_DbProxyUser != null) { 
        String authString = _DbProxyUser+":"+_DbProxyPassword;
        String encodedPassword = "Basic " +new sun.misc.BASE64Encoder().encode(authString.getBytes());
        
        urlcon.setRequestProperty( "Proxy-Authorization", encodedPassword );
      }
      if (urlcon != null) {
        urlcon.connect();
        istream = urlcon.getInputStream();
      }
    } catch (MalformedURLException e) {
      Debug.out("PiconDBServer.MalformedURLException: " + e.getMessage());
      Debug.out("PiconDBServer.getInputStream: MalformedURL - " + urlstr);
      throw e;
    } catch (IOException e) {
      Debug.out("PiconDBServer.getInputStream: " + e.getMessage());
      Debug.out("PiconDBServer.getInputStream: Can not connect to server - " + urlstr);
      throw e;
    }

    return istream;
  }

} // End of class PiconDBServer