package com.jthomas.mailfaces;

import java.awt.*;
import java.awt.event.*;
import java.util.*;

import com.jthomas.util.Debug;
import com.jthomas.util.MultiLineLabel;

/**
 * Prompter.java 
 *   Object that will prompt the user for some input. 
 *   This implementation is primarily done to get around a bug in  
 *   Java 1.0.2 modal Dialogs that does not wait for the User input before 
 *   continuing.  
 *   
 *   This Object is used as follows to display a message and wait for the user
 *   to press OK or Cancel:
 *      Prompter p = Prompter.create();    do this only once
 *      p.setMsgPrompt("Multiline \n msg");
 *      p.showPrompt();
 *      if (p.getStatus() == Prompter.STATUS_OK) 
 *           -----
 *      else
 *           System.exit(0);
 *
 *   This Object is used as follows to display a prompt for input text and wait for the user
 *   to press OK or Cancel:
 *      Prompter p = Prompter.create();    do this only once
 *      p.setTextInputPrompt("Prompt msg","default value","Title line");
 *      p.showPrompt();
 *      if (p.getStatus() == Prompter.STATUS_OK) 
 *           String result = p.getTextInput();
 *      else
 *           System.exit(0);
 *
 *
 *   If used from within an action method, then the
 *   call to p.getStatus() should be done outside of
 *   of the action thread.  
 *
 *   Uses the MultiLineDialog from
 *       David Flanagan
 *      "Java in a Netshell"
 */

public class Prompter extends Dialog {

  // Class Variables
  static protected Frame _DialogFrame; // Frame to attach dialog  to 
  static protected Prompter _Prompter; // The current Prompt dialog

  static final int STATUS_NOTSET = 0;
  static final int STATUS_PENDING = 1;
  static final int STATUS_OK = 2;
  static final int STATUS_CANCEL = 3;

  // Instance Variables
  public int _Status = STATUS_NOTSET;

  private boolean _FlagTextInput = false;
  private String _TextInput;
  private TextField _TextField;
  boolean _Modal;

  /**
  ** Constructor for this class
  ** It calls the constructor for the Dialog super class
  ** This constructor is private because only the
  ** create method needs to call it.
  */

  Prompter(Frame frame, String prompt, boolean modal) {
    super(frame, prompt, modal);
    //------------
    // Handle Window closing event
    //------------
    addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e) {
        Debug.out("Window Closing Event");
        _Status = STATUS_CANCEL;
        Prompter.this.setVisible(false);
        Prompter.this.dispose();
      }
    });

  }

  //---------------------------------------------------
  /**
  ** create(String Prompt,String default)  -> reference to Prompter object
  **   This is a Class method that the user of Prompter will
  **   invoke to build a Prompter dialog object
  **
  **   This method should be invoked only once to build a reuseable
  **   Prompter object
  */
  //------------------------------------------------------

  public static Prompter create(Frame frame, boolean modal) {

    Debug.out("Prompter.create(frame)");
    if (_DialogFrame != null && _DialogFrame != frame) {
      _Prompter = null; // Free any existing reference to Prompt Dialog
    }
    _DialogFrame = frame; // Use our Frame for the Dialog
    _Prompter = new Prompter(_DialogFrame, "Prompt", modal);
    _Prompter._Modal = modal;
    return _Prompter;
  }

  public static Prompter create() {
    Debug.out("Prompter.create()");
    if (_DialogFrame == null)
      _DialogFrame = new Frame("Dialog");

    //if (_Prompter == null)	
    _Prompter = new Prompter(_DialogFrame, "Prompt", true);

    // Return a reference to the Prompter Dialog instance
    _Prompter._Modal = true;
    return _Prompter;
  }

  /**
  ** Method to setup a Msg Dialog 
  ** The message can have multiple lines (using "\n")
  **
  ** @param Message to be displayed in the dialog.  
  */
  public void setMsgPrompt(String msg) {
    Debug.out("Prompter.setMsgPrompt(" + msg + ")");

    removeAll(); // Remove components from last time.
    this.setLayout(new BorderLayout(5, 5));

    Panel main = new Panel();
    main.setLayout(new BorderLayout(5, 5));

    // Setup the line(s) at the top of the window
    this.setFont(new Font("TimesRoman", Font.BOLD, 15));

    MultiLineLabel msgline = new MultiLineLabel(msg, 5, 5);

    main.add("Center", msgline);

    this.add("Center", main);

    this.pack(); // This pack is needed to format correctly with 1.0.2 on Win95

    // Now setup a Panel with the OK and Cancel buttons
    Panel panel = this.setupButtons();

    this.add("South", panel);

    this.pack();

    //_PromptText.move(200,200);       // Where to put it on Users screen
    Dimension size = this.getSize();
    this.setSize(size.width + 20, size.height + 5);
    //this.list(System.out,5);		// <<<<<<<<<<<<<<<<<<<<

  }

  //---------------------------------------------------
  // setTextInputPrompt(String Prompt,String default,String title)  
  //   This will set the Prompt string and Default value
  //   and format the contents of the window.  
  //------------------------------------------------------

  public void setTextInputPrompt(String prompt, String deflt, String title) {
    Debug.out(
      "Prompter.setTextInputPrompt("
        + prompt
        + ","
        + deflt
        + ","
        + title
        + ")");

    _FlagTextInput = true;
    _TextInput = deflt;
    removeAll(); // Remove components from last time.
    this.setLayout(new BorderLayout(5, 5));

    Panel main = new Panel();
    main.setLayout(new BorderLayout(5, 5));

    // Setup the line(s) at the top of the window
    setFont(new Font("TimesRoman", Font.BOLD, 15));
    Panel toppanel = new Panel();
    MultiLineLabel topline = new MultiLineLabel(title, 5, 5);
    toppanel.add(topline);
    main.add("North", toppanel);

    // Setup the prompt line
    setFont(new Font("Dialog", Font.BOLD, 15));
    Label promptLabel = new Label(prompt);
    main.add("West", promptLabel);

    // Setup the text input area
    int fieldLen = deflt.length();
    fieldLen = (fieldLen < 15) ? 15 : fieldLen;
    _TextField = new TextField(deflt.length());
    _TextField.setText(deflt);
    _TextField.requestFocus();
    main.add("Center", _TextField);

    this.add("Center", main);

    this.pack(); // This pack is needed to format correctly with 1.0.2 on Win95

    // Now setup a Panel with the OK and Cancel buttons
    Panel panel = setupButtons();

    this.add("South", panel);

    this.pack();
    this.validate();
    Debug.out(this.getSize().toString());

    //_PromptText.move(200,200);       // Where to put it on Users screen
    Dimension size = this.getSize();
    this.setSize(size.width + 50, size.height + 5);
    //this.list(System.out,5);		// <<<<<<<<<<<<<<<<<<<<

  }

  //-------------------------------------------------------------
  /**
  ** setup a panel with the OK and Cancel buttons
  */
  //----------------------------------------------------------------
  protected Panel setupButtons() {

    Button OK = new Button("  OK  ");
    Button Cancel = new Button("Cancel");
    Panel panel = new Panel();
    panel.setLayout(new FlowLayout(FlowLayout.CENTER, 15, 15));
    setFont(new Font("Dialog", Font.BOLD, 15));
    panel.add(OK);
    panel.add(Cancel);
    OK.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        // handle OK button 
        Debug.out("OK Pressed");
        if (_FlagTextInput)
          _TextInput = _TextField.getText();
        _Status = STATUS_OK;
        setVisible(false);
        //dispose();
      }
    });
    Cancel.addActionListener(new ActionListener() {
      public void actionPerformed(ActionEvent e) {
        // handle OK button 
        Debug.out("Cancel Pressed");
        _Status = STATUS_CANCEL;
        setVisible(false);
        //dispose();			

      }
    });

    return panel;
  }

  //-----------------------------------------------------
  // getTextInput() -> String _TextInput
  //
  //     Return the value that was entered into the TextField
  //------------------------------------------------------------- 
  public String getTextInput() {

    if (_Status == STATUS_OK)
      return _TextInput;
    else // Return null if Cancel button pressed
      return null;
  }

  //-----------------------------------------------------
  /**
   * 
   * showPrompt() -> Put the prompt window up on the screen
   *
   * This is the method that is called by the program using Prompter when
   * it wants the Prompt dialog to be displayed for the user.   
   * At the time it is called, everything is ready, it just needs to be 
   * displayed
   */
  //*------------------------------------------------------------- 
  void showPrompt() {

    Debug.out("Prompter.showPrompt() ");
    _Status = STATUS_PENDING;
    super.setVisible(true);

  }
  
    //-----------------------------------------------------
  /**
   * 
   * setVisible() -> show or hide the window
   *
   * This is the method that is called by the program using Prompter when
   * it wants the Prompt dialog to be displayed for the user.   
   * At the time it is called, everything is ready, it just needs to be 
   * displayed
   */
  //*------------------------------------------------------------- 
  public void setVisible(boolean show) {

    //Debug.out(0,"Prompter.setVisible() ");
    
    super.setVisible(show);

  }

  //-----------------------------------------------------
  /**
  * getStatus() -> int _Status
  *
  * This is the method that is called by the program using Prompter when
  * it wants the Users input.
  * This will poll for the _Status flag to be set.  The polling loop contains
  * a 1 second sleep.
  * Because of the sleep, we cannot invoke this dialog out of an Event thread because
  * would hang if the event thread goes to sleep
  */
  //*------------------------------------------------------------- 
  public int getStatus() {

    Debug.out("Prompter.getStatus()");

    if (!_Modal)
      return _Status;

    if (_Status != STATUS_PENDING) {
      // The Modal Dialog worked! must not be Win95
      Debug.out("getStatus() _Status = " + _Status);
    } else {
      Debug.out("Prompter.getStatus() will poll for status");
      while (_Status == STATUS_PENDING) {

        try {
          Thread.currentThread().sleep(1000);
        } catch (InterruptedException e) {
          Debug.out("Interrupted");
        }
      }
    }
    Debug.out("Prompter.getStatus() - returned" + _Status);

    return _Status;

  }

  //*-----------------------------------------------------
  //* Adding this makes this class self-testing
  //*-----------------------------------------------------
  public static void main(String args[]) {
    String result;
    int status;
    Debug.setDebugOn(true);
    if (args.length > 0) { //* 1st argument is debug output file name
      Debug.setDebugFile(args[0]);
    }
    Prompter p = Prompter.create(); //*do this only once

    p.setMsgPrompt(
      "This is the first line\n"
        + "This is the 2nd line\n"
        + "and this is the last line");
    //p.showPrompt();
    p.showPrompt();
    Debug.out("This line should not show until promt satisfied.");
    status = p.getStatus();

    Debug.out("Status=" + status);
    p = Prompter.create();
    p.setTextInputPrompt("Enter text value", "default", "Prompt for testing");
    p.showPrompt();
    if ((p.getStatus()) == STATUS_OK)
      Debug.out(p.getTextInput());
    
  }

} //End of Prompter class