package com.jthomas.util;

/* --------------------------------------------------------
 * Debug.java
 * Copyright (c) 1996-1998 John Thomas  jthomas@cruzio.com
 *      All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for any purpose is hereby granted provided 
 * that this copyright notice appears in all copies.
 */

import java.io.*;

//--------------------------------------------------------------------
/** 
 * This static class is a substitute for the programmer's favorite
 * debuging tool; inserting 'System.out.println("a="+a);'. 
 * 
 * <pre>
 * "Debug.out()" compared to "System.out.println()"
 * 
 *   - 9 fewer keystrokes!
 *   - It can be turned on and off at either compile time or runtime.
 *   - Any Object can be passed to Debug.out 
 *   - By default, it will display the Threadname as the 1st output parameter
 *   - Output can be written to STDERR and/or to a file.
 *   - Can be used inside a Catch block to both print a message
 *     and produce a stacktrace.
 *   - Can be used to somewhat simulate the C/C++ assert statement.
 *   - Can be used to specify different Debug levels.
 *   - Can print a stacktrace outside of a try/catch block.
 *   - Can print to either stderr and/or file  (controlled with set method)
 *   - The Thread name can optionally be included in each output.
 * --------------------------------------------------------------
 * 
 * John Thomas
 * October 96
 *
 * Usage:
 *
 * To have Debug output displayed (by default to stderr) 
 *     Debug.setDebugOn(true)
 * 
 * To have Debug output also written to a file
 *     Debug.setDebugFile("debug.txt");
 *
 * To suppres the output to STDERR  (defaults to on)
 *     Debug.setDebugOut(false);
 *
 * To control the display of the Thread name(Default=true: 
 *     Debug.setDebugThreadOn(true/false)
 *
 * To produce debug output just specify a string 
 *   as a parameter to the Class method "Debug.out(String x)"
 * Instead of a String, you can specify any Java Object 
 * and it will print the results of Object.toString();
 *
 * 		For example:
 *     Debug.out("value=" + value);
 *
 *     Vector myvector = new Vector();
 *     ....
 *     Debug.out(myvector);
 *
 * To produce a stacktrace both on STDOUT and on the file
 * if being used, when an exception occurs
 * then use the Debug.out(String msg, Exception e) 
 * or Debug.out(Exception e) method.  
 * If you don't want the stacktrace then use:
 *    Debug.out(e.getMessage());
 * 
 * For Example:
 *     try {
 *        something
 *     } 
 *     catch (Exception e) {
 *       Debug.out("Exeception doing something",e);
 *     }
 *
 *
 * To produce a stacktrace both on STDOUT and on the file
 * if being used, at anytime without stoping execution
 * then use the Debug.printStackTrace(String description) method.
 * 
 * For Example:
 *     ...
 *     Debug.printStackTrace("entered wierd state");
 *     ...
 *     
 * ** ASSERT **
 * To produce a RuntimeException when certain conditions are 
 * not true, then use the Debug.assert(boolean asserted) method or 
 * use the Debug.assert(boolean asserted,String debugmsg) method.
 * 
 * For Example:
 *     ...
 *    Debug.assert(thisObj != null);
 *  or
 *     ...
 *    Debug.assert(value < max, "value="+value);
 *
 *  Normally assert processing is only done if Debug.ON is true but
 *  You can call "Debug.setAssert(true) so that assert is always 
 *  processed even if setDebugOn(false) is set or defaulted.
 *
 *
 * To use Debug Levels.
 *    Debug.setDebugOn(5,true);    // turn on level 5.
 *    Debug.out(5,"special message");  
 *    
 *    Levels 3-9 are available for special use
 *    Level 0 (or Debug.ALWAYS) is by default true.
 *    Level 1 (or Debug.MIN) is the same as the default level
 *    Level 2 (or Debug.MAX) will display both MIN and MAX levels
 *          setDebugOn(Debug.MAX,true) will turn on all levels.
 *    
 *    Debug.out(Debug.ALWAYS,"always print");
 *    Debug.out(Debug.MAX,"Print if Debug or Level MIN or MAX is on);
 *    Debug.out(5," print only if Debug.setDebug(5,true)" );
 *
 * To display the Java and OS environment information on the Debug file
 *    
 *    Debug.environment()
 *   
 *       This will display the Java Runtime and OS environment.
 *
 * To completely shutoff all Debug output, even exceptions:
 *
 *    Debug.quiet(true);
 *
 * Performance considerations
 *    Debug when turned on is very expensive because it results in 
 *    I/O to a file or console..But usually in a Debug mode, one is 
 *    not so concerned with performance. 
 *     
 *    If the Debug option is false ( setDebugOn(false) ) then a 
 *    call to Debug.out("msg") will only execute a single if 
 *    statement inside of the Debug.out method.  This is very minimal
 *    and a JIT compiler will make it even more minimal.  However 
 *    note that any calls to generate the debug parameter will also 
 *    be invoked.  
 *    i.e. 
 *      Debug.out("name="+userobj.getName());
 *    will invoke the call to the getName() method regardless of 
 *    of whether Debug is on.  
 *    However since the Debug ON variable is public, you could code
 *      if (Debug.ON) Debug.out("name="+userobj.getName());
 * </pre>
 *
 */
//---------------------------------------------------------------------- 

public class Debug {

  /**
  ** If true then the Thread name will be included in output
  */
  static boolean debugThread = true; // Include Thread name
  /**
  ** If true, then Format and write the info to stderr, file or both.
    ** <p>
    ** One can optimize the efficiency when Debug is not on by coding<br> 
    **   if (Debug.ON) Debug(something.toString()); <br>
    ** This will avoid calling the toString method of something unless 
    ** Debug.ON is set.
    ** <p>
    ** One could optimize away even more overhead by making the following 
  ** "static final"; commenting out where it is set and then recommpiling the code.
  */
  public static boolean ON = false; // Debug On switch

  /**
  ** One could have Debug statements that generate no runtime code
  ** by using the following:
  **   if (Debug.DEBUG) Debug.out("some debug output");
  ** Then these statements could be enabled by modifing the following
  ** value to "true"
  */
  static final boolean DEBUG = true;

  /** 
  ** Setting _quiet to true (with setQuiet(true) method) will ensure that 
  ** Debug will not produce any output even for exceptions.
  */
  static boolean _quiet = false;

  /** 
  ** if True then assert methods will be processed even if Debug.ON is false.
  ** Use setAssert(true) to change.
  */
  static boolean ASSERT = false;

  /**
  ** If true then write to StdErr
  */
  static boolean debugOut = true; // Write to StdErr

  /**
  ** if not null this is the file that we write to .
  */
  static String debugFileName = null; // Write to FileName

  // Let the public access the PrintStream directly 
  //static public		PrintStream debugOut=null;  // Java 1.0

  /**
  **  Stream that is used to write the debugFile
  */
  static public PrintWriter debugFile = null; // Java 1.1

  /**
  ** Array of Debug levels from 0 to 9)
  */
  static boolean[] debugLevel = new boolean[10];

  /**
  ** specifies that Debug out should normally be displayed.
  */
  public static final int ALWAYS = 0;
  /** 
  ** Default level[ALWAYS] to true
  */
  static {
    debugLevel[ALWAYS] = true;
  }

  /**
  ** specifies the default level of Debug output.
  */

  public static final int MIN = 1;

  /** 
  ** specifies that all levels of debug output be displayed.
  */
  public static final int MAX = 2;

  //------------------------------------------------------
  /**
   ** Method to let caller turn the debug switch on or off.
   **
   ** @param OnOff  boolean value to turn Debug output on or off.
   **/
  //------------------------------------------------------
  public static void setDebugOn(boolean OnOff) {
    ON = OnOff;
    debugLevel[MIN] = OnOff;
  }

  //------------------------------------------------------
  /**
    ** Method to let caller control output to STDERR.
    **
    ** @param OnOff  boolean value to turn STDERR output on or off.
    **/
  //------------------------------------------------------
  public static void setDebugOut(boolean OnOff) {
    debugOut = OnOff;

  }
  //------------------------------------------------------
  /**
    ** Method to let caller control output to STDERR.
    **
    ** @param OnOff  boolean value to turn STDERR output on or off.
    **/
  //------------------------------------------------------
  public static boolean getDebugOut() {
    return debugOut;

  }
  //------------------------------------------------------
  /**
    ** Method to let caller turn off all output      
    **
    ** @param OnOff  boolean value to turn all output on or off.
    **/
  //------------------------------------------------------
  public static void setQuiet(boolean OnOff) {
    _quiet = OnOff;
  }

  //------------------------------------------------------
  /**
   ** Method to let caller turn the debug switch on or off.
   ** for specified levels
   ** @param level This is the level that is being turned on or off.
   ** @param OnOff new boolean setting.
   **/
  //------------------------------------------------------
  public static void setDebugOn(int level, boolean OnOff) {
    if (level == MAX)
      for (int i = MAX; i < 10; i++)
        debugLevel[i] = OnOff;
    else
      debugLevel[level] = OnOff;

    if (OnOff == true)
      ON = OnOff; // for convenience
  }

  //------------------------------------------------------
  /**
   ** Method to let caller test the value of the debug switch.
   ** 
   ** @return Returns the boolean value of the Debug setting.
   **/
  //------------------------------------------------------
  public static boolean isDebugOn() {
    return ON;
  }

  //------------------------------------------------------
  /**
    ** Method to let caller turn the thread debug switch on or off.
    **
    **/
  //------------------------------------------------------
  public static void setDebugThreadOn(boolean OnOff) {
    debugThread = OnOff;
  }

  //------------------------------------------------------
  /**
    ** Method to let caller turn the assert processing on or off.
    **
    **/
  //------------------------------------------------------
  public static void setAssert(boolean OnOff) {
    ASSERT = OnOff;
  }

  //------------------------------------------------------
  /**
    ** Method to let caller specify a file to write debug info to.
    ** It will also set the Debug  switch to true.
    **
    ** @param file File name to be used to write the debug output to.
    **/
  //------------------------------------------------------
  public static void setDebugFile(String filename) {
    if (debugFile != null && debugFileName.equals(filename))
      return; // Already in use
    if (filename == null) {
      debugFileName = null;
      debugFile = null;

    } else {
      debugFileName = filename;
      try {
        BufferedOutputStream bos = new BufferedOutputStream(new FileOutputStream(filename));
        debugFile = new PrintWriter(bos); //Java 1.1      
      } catch (IOException e) {
        debugFile = null;
        System.err.println("Unable to setup Debug output file");
      }
    }
    //ON = (debugOut || (debugFile != null));

  }

  //------------------------------------------------------
  /**
    **  If ON switch set, display the specified debug info.
    **  If the specified Object is an Exception then it will be 
    **  treated as if Debug ON switch was set.  
    **  
    **  @param debugobj The parameter is usually a string but it can 
    **  also be any other object, in which case it will display 
    **  the result of the object.toString() method. 
    **  
    **/
  //------------------------------------------------------
  public static void out(Object debugobj) {
    // special processing for exceptions
    if (debugobj instanceof Throwable) {
      out(ALWAYS, "Exception", (Throwable) debugobj);
      return;
    }

    if (!ON)
      return; // Make a quick exit
    else
      out(MIN, debugobj);

  }

  //------------------------------------------------------
  /**
    **   display the specified String unmodified.
   
    **  
    **  @param line The parameter is a String 
    
    **/
  //------------------------------------------------------
  public static void println(String line) {
      System.err.println(line); // always write to stderr
      if (debugFile != null) { // Also make a copy to a file
        // output to file
        debugFile.println(line);
        debugFile.flush();
      }
      
  }
  //------------------------------------------------------
  /**
   *  If debugOn switch set for specified level, display the specified debug info.
   * 
   *  If debugOut is true - use Ssystem.er.println
   *  If debugFile is set, write to the specified file and then 
   *     flush the buffers. 
   *
   *  @param level Only display if this level is set true.
   *  @param debugobj - The parameter is usually a string but it can 
   *  also be any other object, in which case it will display 
   *  the result of the object.toString() method. 
   */
  public static void out(int level, Object debugobj) {

    if (((!ON) && level != ALWAYS) | (_quiet))
      return; // make a quick exit 

    if (level == ALWAYS || debugLevel[level]) {
      // special processing for exceptions
      if (debugobj instanceof Throwable) {
        out(level, "Exception", (Throwable) debugobj);
        return;
      }
      String strobj;
      if (debugobj instanceof String)
        strobj = (String) debugobj;
      else if (debugobj == null)
        strobj = "Object=null";
      else {
        String type = debugobj.getClass().getName();
        strobj = type + "=" + debugobj.toString();
      }
      String t;
      if (debugThread)
        t = Thread.currentThread().getName() + ": " + strobj;
      else
        t = strobj;
      if (debugOut)
        System.err.println(t); // always write to stderr
      if (debugFile != null) { // Also make a copy to a file
        // output to file
        debugFile.println(t);
        debugFile.flush();
      }

    }
  }

  //------------------------------------------------------
  /**
   *  If ON switch set, display the specified debug info for 
   *  one of the Java primitive types.
   *  (boolean,char,byte,short,int,long float,double)
   *	
   *  @param A java primitive type that is to be displayed.
    */
  public static void out(boolean primitive) {
    Debug.out("" + primitive);
  }

  public static void out(char primitive) {
    Debug.out("" + primitive);
  }

  public static void out(byte primitive) {
    Debug.out("" + primitive);
  }

  public static void out(short primitive) {
    Debug.out("" + primitive);
  }

  public static void out(int primitive) {
    Debug.out("" + primitive);
  }

  public static void out(long primitive) {
    Debug.out("" + primitive);
  }

  public static void out(float primitive) {
    Debug.out("" + primitive);
  }

  public static void out(double primitive) {
    Debug.out("" + primitive);
  }

  //------------------------------------------------------
  /**
   *  Display the stacktrace info after an exception
   *  
   *  Display the exception info + stacktrace info after an exception 
   *  This form of Debug.out is always displayed.  
   *
   * @param debugmsg  String to be displayed.
   * @param e         Throwable object that is to be displayed.
   */
  public static void out(String debugmsg, Throwable e) {

    if (_quiet) // No output if in quiet mode
      return;

    String t;

    if (debugThread)
      t = Thread.currentThread().getName() + ": " + debugmsg;
    else
      t = debugmsg;

    System.err.println(t);
    e.printStackTrace();

    if (debugFile != null) { // Also make a copy to a file
      debugFile.println(t);
      e.printStackTrace(debugFile); // output to file
      debugFile.flush();
    }
    Log.out("Exception", e);

    //}
  }

  //------------------------------------------------------
  /**
   *  Display the stacktrace info after an exception based on level
   *
   *  @param level Only display if this level is set true.
   * @param debugmsg  String to be displayed.
   * @param e         Throwable object that is to be displayed.
   */
  public static void out(int level, String debugmsg, Throwable e) {
    if (ON && debugLevel[level])
      out(debugmsg, e);
    else if (level == ALWAYS)
      out(debugmsg, e);

  }

  //------------------------------------------------------
  /**
   *  Display the stacktrace info at any time. 
   *
   * @param debugmsg String to be displayed along with the stacktrace
   */
  public static void printStackTrace(String debugmsg) {
    if (ON) {
      try {
        throw new Exception("Debug");
      } catch (Exception e) {
        Debug.out(debugmsg, e);
      }
    }

  }

  //------------------------------------------------------
  /**
   *  Display the System environment information. 
   *
   */
  public static void environment() {
    if (ON) {
      String version = System.getProperty("java.version");
      Debug.out("java.version=" + version);

      String os = System.getProperty("os.name");
      Debug.out("os.name=" + os);

      String userdir = System.getProperty("user.dir");
      Debug.out("user.dir=" + userdir);

      String compiler = System.getProperty("java.compiler");
      Debug.out("java.compiler=" + compiler);

    }

  }

  //-----------------------------------------------------
  /**
  ** Throw a RuntimeException if the argument is false.
  **
  **
  ** @param expression that will resolve to a boolean value.
  */
  public static void asserted(boolean asserted) {
    if ((!ON && !ASSERT) || asserted)
      return; // Go right back, everything is OK

    // The assertion was not true so throw a runtime Exception
    // Normally one would never throw a runtime exception but 
    // we want the programmer that uses assert to be able 
    // to easily slip the assert in without a try-catch block.
    throw new RuntimeException("Assertion Failed");
  }

  //-----------------------------------------------------
  /**
  ** Display a message if the argument is false.
  **
  **
  **
  ** @param asserted expression that will resolve to a boolean value.
  ** @param debugmsg String to be displayed.
  */
  public static void asserted(boolean asserted, String debugmsg) {
    if ((!ON && !ASSERT) || asserted)
      return; // Go right back, everything is OK
    Debug.out(0, debugmsg);

  }
  //-----------------------------------------------------
  /**
  ** Display a message if the argument is false
  **
  ** @param asserted expression that will resolve to a boolean value.
  ** @param debugmsg String to be displayed.
  ** @param fail    If true then throw a Runtime Exception if assertion is false.
  */
  public static void asserted(boolean asserted, String debugmsg, boolean fail) {
    if ((!ON && !ASSERT) || asserted)
      return; // Go right back, everything is OK
    Debug.out(0, debugmsg);
    if (fail) {
      // The assertion was not true so throw a runtime Exception
      // Normally one would never throw a runtime exception but 
      // we want the programmer that uses assert to be able 
      // to easily slip the assert in without a try-catch block.
      throw new RuntimeException("Assertion Failed");
    }
  }
  /**
  ** Verify that it works.
  */

  public static void main(String[] argv) {

    Debug.setDebugOn(true);
    if (Debug.ON)
      Debug.out("Testing the Debug class");
    Debug.setDebugFile("debug.txt");
    Debug.out("Test Debug output to File");
    Debug.printStackTrace("test StackTrace");
    Debug.environment();
    Debug.out("Generate IndexOutOfBounds Exception");
    try {
      String test = "test";
      test = test.substring(5); // will cause exception
    } catch (Exception e) {
      Debug.out("Catch expected StringIndexOutOfBounds");
      Debug.out(e);
      Debug.out("Caught an exception", e);
    }
    // verify that it handles objects even when they are null
    Thread tobj = null;
    Debug.out(tobj);
    tobj = new Thread(); // Useless thread  
    Debug.out(tobj);
    Debug.out("test OK");
    Debug.setDebugOn(false);
    Debug.setDebugFile(null);
    Debug.out("Should not show up");
    Debug.setDebugOn(false);
    Debug.out(Debug.ALWAYS, "ALWAYS test should be displayed.");
    Debug.setDebugOn(true);

    Debug.setDebugOn(Debug.MIN, true);
    Debug.out(Debug.MIN, "min test1");
    Debug.out(Debug.MAX, "MAX - Should not show up");
    Debug.out(5, "Level 5 - Should not show up");
    Debug.setDebugOn(Debug.MIN, false);
    Debug.setDebugOn(Debug.MAX, true);

    Debug.out(Debug.MIN, "min test2");
    Debug.out(Debug.MAX, "max test");
    Debug.out(5, "level 5 test OK");
    Debug.setDebugOn(false);
    Debug.out("Should not appear");
    Debug.out(Debug.ALWAYS, "This should appear");
    if (Debug.ON)
      Debug.out(0, "This should not appear");
    Debug.setDebugOn(true);

    Debug.setQuiet(true);
    Debug.out(0, "This message should not appear");
    Debug.out("Generate IndexOutOfBounds Exception. Should not appear");
    try {
      String test = "test";
      test = test.substring(5); // will cause exception
    } catch (Exception e) {
      Debug.out(e);
      Debug.out("Caught an exception", e);
    }
    Debug.setQuiet(false);

    Debug.out(0, "Test the assert method");
    int testit = 5;
    Debug.asserted(testit == 5);
    Debug.asserted(testit == 6, "Assert triggered. testit=" + testit);

    Debug.out("We should fail with an \"Assert Failure\" on the next stmt");
    Debug.asserted(testit == 6, "Assert Failure: testit=" + testit, true);
  }

} // End of Debug class
