package com.jthomas.util;

import java.awt.*;
import java.net.*;
import java.awt.image.*;           // For ImageFilter stuff

//============================================================================
// TO DO: Make button resizable and stretch the image to fit.

/**
 * A button class that uses an image instead of a textual label.
 * Unlike many of the ImageButton classes available, it triggers
 * an ACTION_EVENT, so you can add behavior in the same two ways
 * as you can with normal Buttons:
 * <OL>
 *  <LI>Make a subclass and put the behavior in the <B><TT>action</TT></B>
 *      method of that subclass.
 *  <LI>Use the main class but then catch the event in the
 *      <B><TT>action</TT></B> method of the Container.
 * </OL>
 * <P>
 * By default, with FlowLayout the ImageButton takes its minimum size
 * (just enclosing the image). The default with BorderLayout is
 * to expand to fill the region in width (North/South), height
 * (East/West) or both (Center). This is the same behavior as the builtin
 * Button class. If you give an explicit <TT>resize</TT> or
 * <TT>reshape</TT> call <B>before</B> adding the ImageButton to the
 * Container, this size will override the defaults.
 * <P>
 * The original of the source code can be found at 
 * <A HREF="http://www.apl.jhu.edu/~hall/java/ImageLabel/ImageButton.java">
 * http://www.apl.jhu.edu/~hall/java/ImageLabel/ImageButton.java</A>,
 * the documentation is at
 * <A HREF="http://www.apl.jhu.edu/~hall/java/ImageLabel/ImageButton.html">
 * .../ImageButton.html</A>, and a small example program can be found at
 * <A HREF="http://www.apl.jhu.edu/~hall/java/ImageLabel/ImageButtonTest.html">
 * .../ImageButtonTest.html</A>. 
 * <P>
 * No warranty of any kind is provided.
 * Permission is granted to use and/or modify for any purpose.
 * <P>
 * 7/96 Marty Hall:
 * <UL>
 *    <LI><A href="http://www.apl.jhu.edu/~hall/java">
 *        Java Resource Page.</A> 
 *    <LI><A href="mailto:hall@apl.jhu.edu">
 *        Email.</A>
 *    <LI><A href="http://www.apl.jhu.edu/~hall">
 *        Home Page.</A> 
 * </UL>
 *
 * @see Icon
 * @see GrayFilter
 * @version 0.9 (August 1996)
 */

public class ImageButton extends ImageLabel {
  //-----------------------------------------------------------------

  /** Default width of 3D border around image. Currently 4.
   * @see #setBorderWidth
   * @see #getBorderWidth
   */
  protected static final int defaultBorderWidth = 4;

  /** Default color of 3D border around image. Currently a gray
   *  with R/G/B of 160/160/160. Light grays look best.
   * @see #setBorderColor
   * @see #getBorderColor
   */
  protected static final Color defaultBorderColor = new Color(160, 160, 160);

  /** Gray image created automatically from regular image via
   *  an image filter to use when button is depressed. You won't normally
   *  use this directly. 
   */
  protected Image grayImage = null;

  // Darker is consistent with regular buttons
  /** An int whose bits are combined via "and" ("&") with the alpha,
   *  red, green, and blue bits of the pixels of the image to produce
   *  the grayed-out image to use when button is depressed.
   *  Default is 0xffafafaf: af combines with r/g/b to darken image.
   */
  public int darkness = 0xffafafaf;
  
  private boolean mouseIsDown = false;
  
  //-------------------------------------------------------------------------
  // Constructors
  
  /** Create an ImageButton with the default image.
   * @see ImageLabel#defaultImageString
   */
  public ImageButton() {
    super();
    setDefaults();
  }

  /** Create an ImageButton using the image at URL specified by the string.
   * @param imageURLString A String specifying the URL of the image.
   */
  public ImageButton(String imageURLString) {
    super(imageURLString);
    setDefaults();
  }

  /** Create an ImageButton using the image at URL specified.
   * @param imageURL The URL of the image.
   */
  public ImageButton(URL imageURL) {
    super(imageURL);
    setDefaults();
  }

  /** Create an ImageButton using the image specified. You would only want to
   *  call it directly if you already have an image (e.g. created via
   *  createImage).
   * @param image The image.
   */
  public ImageButton(Image image) {
    super(image);
    setDefaults();
  }

  //-----------------------------------------------------------------
  /** Sets the width around the outside of the image.
   * @see #getBorderWidth
   */
  public void setBorderWidth(int borderWidth) {
    margin = borderWidth;
  }

  //-----------------------------------------------------------------
  /** Returns the width around the outside of the image.
   * @see #setBorderWidth
   */
  public int getBorderWidth() {
    return(margin);
  }

  //-----------------------------------------------------------------
  /** Sets the Color for the border around the outside of the image.
   * @see #getBorderColor
   */
  public void setBorderColor(Color borderColor) {
    marginColor = borderColor;
  }

  //-----------------------------------------------------------------
  /** Returns the Color for the border around the outside of the image.
   * @see #setBorderColor
   */
  public Color getBorderColor() {
    return(marginColor);
  }
  
  //-----------------------------------------------------------------
  /** Draws the image with the border around it. If you override
   *  this in a subclass, call super.paint().
   */
  public void paint(Graphics g) {
    super.paint(g);          // Forces main image to get loaded
    if (grayImage == null)   // Creates gray image from main image
      createGrayImage(g);
    drawBorder(g,true);
  }

  //-----------------------------------------------------------------
  // You only want mouseExit to repaint when mouse is down, so
  // you have to set that flag here.

  /** When the mouse is clicked, reverse the 3D border and
   *  draw a dark-gray version of the image. The action is not
   *  triggered until mouseUp.
   */
  public boolean mouseDown(Event event, int x, int y) {
    mouseIsDown = true;
    Graphics g = getGraphics();
    if (explicitSize)
      g.drawImage(grayImage, margin, margin,
		  width-2*margin, height-2*margin,
		  this);
    else
      g.drawImage(grayImage, margin, margin, this);
    drawBorder(g,false);
    return(true);
  }

  //-----------------------------------------------------------------
  /** If cursor is still inside, trigger the action event and redraw
   *  the image (non-gray, button "out"). Otherwise
   *  ignore this.
   */
  public boolean mouseUp(Event event, int x, int y) {
    mouseIsDown = false;
    if (contains(x,y)) {
      paint(getGraphics());
      event.id = Event.ACTION_EVENT;
      event.arg = (Object)image;
      return(action(event, event.arg));
    } else
      return(false);
  }

  //-----------------------------------------------------------------
  /** Generated when the button is clicked and released. Override this
   *  in subclasses to give behavior to the button. Alternatively, since
   *  the default behavior is to pass the ACTION_EVENT along to the
   *  Container, you can catch events for a bunch of buttons there.
   * @see Component#action
   */
  
  public boolean action(Event event, Object arg) {
    debug("Clicked on button for " + imageString + ".");
    return(false);
  }
  
  //-----------------------------------------------------------------
  /** If you move the mouse off the button while the mouse is down,
   *  abort and do NOT trigger the action. Ignore this if
   *  button was not already down.
   */
  public boolean mouseExit(Event event, int x, int y) {
    if (mouseIsDown) 
      paint(getGraphics());
    return(true);
  }
  
  //-----------------------------------------------------------------

  private void drawBorder(Graphics g,boolean isUp) {
    
    g.setColor(marginColor);
    int left = 0;
    int top = 0;
    int width = this.width;
    int height = this.height;
    for(int i=0; i<margin; i++) {
      g.draw3DRect(left, top, width, height, isUp);
      left++;
      top++;
      width = width - 2;
      height = height - 2;
    }
  }

  //-----------------------------------------------------------------

  private void setDefaults() {
    margin = defaultBorderWidth;
    marginColor = defaultBorderColor;
  }

  //-----------------------------------------------------------------
  // The first time the image is drawn, update() is called, and
  // the printout does not come out correctly. So this forces a
  // brief draw on loadup, replaced by real, non-gray image.
  
  private void createGrayImage(Graphics g) {
    ImageFilter filter = new GrayFilter(darkness);
    ImageProducer producer = new FilteredImageSource(image.getSource(),
						     filter);
    grayImage = createImage(producer);
    if (explicitSize)
      g.drawImage(grayImage, margin, margin,
		  width-2*margin, height-2*margin,
		  this);
    else
      g.drawImage(grayImage, margin, margin, this);
    super.paint(g);
  }

  //-----------------------------------------------------------------
}

//============================================================================
/** Builds an image filter that can be used to gray-out the image.
 * @see ImageButton
 */

class GrayFilter extends RGBImageFilter {

  //-----------------------------------------------------------------

  private int darkness = 0xff808080;
  
  //-----------------------------------------------------------------

  public GrayFilter() {
    canFilterIndexColorModel = true;
  }

  public GrayFilter(int darkness) {
    this();
    this.darkness = darkness;
  }

  //-----------------------------------------------------------------

  public int filterRGB(int x, int y, int rgb) {
    return(rgb & darkness);
  }
  
  //-----------------------------------------------------------------
}

//============================================================================