package com.jthomas.util;
/*
 * JEditor.java
 * Copyright (c) 1996-1998 John Thomas  jthomas@cruzio.com
 *      All Rights Reserved.
 *
 * Permission to use, copy, modify, and distribute this software and
 * its documentation for non-commercial purposes
 * is hereby granted provided that this copyright notice
 * appears in all copies.
 *
 */
import java.awt.*;
import java.awt.event.*;
import java.util.Hashtable;
import java.io.*;
import javax.swing.*;
import javax.swing.text.*;
import javax.swing.event.*;
import javax.swing.undo.*;
/**
 * Simple editor to edit a single file. 
 * 
 * This is a simple text editor that will edit a single file. 
 * It supports saving the edited file with the original name
 *      
 * 
 * @author jthomas
 *
 */

public class JEditor extends JFrame {
  
  static final Dimension DEFAULT_DIMENSION = new Dimension(800, 400);
  /**
   * main editor window
   */
  JTextPane _textPane;

  MutableAttributeSet _normalStyle;
  
  /**
   * the Document that we are editing.
   */
  Document _doc = null;

  //String newline = "\n";
  
  /**
   * Hashtable of Action names.  
   * used by the undo menu support
   */
  Hashtable _actions;

  /**
   * The filename that we are editing
   * It is set by the setFileName() method.
   */
  String _fileName = null;

  //undo helpers
  protected UndoAction _undoAction;
  protected RedoAction _redoAction;
  protected UndoManager _undo = new UndoManager();

  protected SaveAction _saveAction;
  protected ExitAction _exitAction;
  protected boolean _fileModified = false;
  /**
   * constructor for JEditor.
   * <pre>
   * It will do the folloing:
   *   
   *   - initialize the JFrame
   * 
   *   - setup the JTextPane
   * 
   *   - get a Document object 
   * 
   *   -  setup the menubar
   *   
   *   - start listening for important events
   *  
   * Window hierarchy
   *   JFrame (this)  
   *      JMenuBar   mb
   *      JPanel  (contentPane)
   *        JScrollPane (scrollPane)
   *           JTextPane  (_textPane)
   *         
   */
  public JEditor() {
    //some initial setup
    super("JEditor");


  }

  //--------------------------------------------------------------
  /**
   * set the FileName for the file that we want to edit.
   *
   * @param fileName is the file we want to edit.
   */
  public void setFileName(String fileName) {
    _fileName = fileName;
    File test = new File(_fileName);
    this.setTitle("JEditor: "+test.getAbsolutePath());
  }
  
  //------------------------------------------------------------------
  /**
   * setup for the JTextPane.
   * 
   * This is the JTextpane processing that occurs after the 
   * caller has set the file name.
   * 
    <pre>
   * It will do the folloing:
   * 
   *   - setup the JTextPane
   * 
   *   - get a Document object 
   * 
   *   - Read the file into the Document
   * 
   *   -  setup the menubar
   *   
   *   -  build the window hierarchy
   * 
   *   - start listening for important events
   *  
   *   -  Make the window visible
   *   </pre>
   */  
  public void setup() {
    //
        //Create the text pane and configure it.
    _textPane = new JTextPane();
    //Create the document for the text area.
    _doc = _textPane.getDocument();
    _normalStyle = new SimpleAttributeSet();
    //
    // set monospace 
    //
    //StyleConstants.setFontFamily(_normalStyle,"SanSerif");
    StyleConstants.setFontFamily(_normalStyle,"Monospaced");
    StyleConstants.setFontSize(_normalStyle,14);
    //StyleConstants.setBold(_normalStyle,true); 
    
    _textPane.setCharacterAttributes(_normalStyle,true);
    
    //
    // read the file into the Document object 
    fileInput(_doc);
    
    // set to start editing at the top
    _textPane.setCaretPosition(0);
    
    _textPane.setMargin(new Insets(5, 5, 5, 5));
    JScrollPane scrollPane = new JScrollPane(_textPane);
    scrollPane.setPreferredSize(DEFAULT_DIMENSION);
    
    //Add the components to the frame.
    JPanel contentPane = new JPanel(new BorderLayout());
    contentPane.add(scrollPane, BorderLayout.CENTER);

    this.setContentPane(contentPane);
    //
    //Set up the menu bar.
    //
    createActionTable(_textPane);

    JMenu fileMenu = createFileMenu();
    JMenu editMenu = createEditMenu();
    
    JMenuBar mb = new JMenuBar();
    mb.add(fileMenu);
    mb.add(editMenu);
    //mb.add(styleMenu);
    setJMenuBar(mb);

 
    //Start watching for undoable edits .
    _doc.addUndoableEditListener(new MyUndoableEditListener());

    _doc.addDocumentListener(new MyDocumentListener());

   
    this.addWindowListener(new WindowAdapter() {
      public void windowClosing(WindowEvent e) {
        //System.exit(0);
        JEditor.this.dispose();
      }
      public void windowActivated(WindowEvent e) {
        JEditor.this._textPane.requestFocus();
      }
    });
 
    // let the user see the window now.   
    this.pack();
    this.setVisible(true);
  }
  
  
  
  
  //-----------------------------------------------------------
  /**
   * access the file that we are going to edit.
   * 
   * @param doc is the Document objet that will 
   * manage the editing.
   */
  private void fileInput(Document doc) {
    try {
      //EditorKit editKit = new DefaultEditorKit();
      
      if (_fileName == null) {
        // this shouldn't happen
        Debug.out(0,"JEditor.fileOutput: No fileName specified");
        doc.insertString(0, "JEditor.fileOutput: No fileName specified", _normalStyle);
      } else {
        // setup a connection to the file
        FileReader fileReader = new FileReader(_fileName);
        BufferedReader bufferReader = new BufferedReader(fileReader);
        
        int pos = 0;          
        String line = null;
        
        //
        // read the file and insert into the Document object
        //
        while (true) {
          try {
          	line = bufferReader.readLine();
          	if (line == null) 
          	  break;
          	line = line.trim(); 
          	if (line.length() == 0)
          	  continue;  // ignore blank lines 
          	line  += "\n";
          	if (Debug.ON) 
          	  Debug.out(0,"> "+line.length()+":"+line);
          	doc.insertString(pos, line, _normalStyle);
          	pos += line.length();
          } catch (IOException io) {
          	Debug.out(io);
          	break;
          } 
        }
          

        // show that it has not been modified.
        setModified(false);
        
      }
    } catch (Exception e) {
      Debug.out(e);
    }
    return;
  }
  
  //-----------------------------------------------------------
  /**
   * write the edited file.
   * 
   * This will write the edited file back to the File.
   * 
   * If it has not bbeen modified then we will display 
   * a warning message and return.
   */
  private void fileOutput() {
    try {
      EditorKit editKit = new DefaultEditorKit();
      
      if (! isModified()) {
        Debug.out(0, "JEditor.fileOutput: File has not been modified.");
        return;
      }
      if (_fileName == null) {

        Debug.out(0, "JEditor.fileOutput: No fileName specified");
      } else {
        FileWriter fileWriter = new FileWriter(_fileName);
        int len = _doc.getLength();
        editKit.write((Writer) fileWriter, _doc, 0, len);
        Debug.out("JEditor.fileOutput: Wrote "+len+" bytes to "+_fileName);
        setModified(false);
      }
    } catch (Exception e) {
      Debug.out(e);
    }
    return;
  }



  //-----------------------------------------------------------
  /**
   * Create the file menu.
   */
  protected JMenu createFileMenu() {
    JMenu menu = new JMenu("File");

    _saveAction = new SaveAction();
    menu.add(_saveAction);

    menu.addSeparator();

    _exitAction = new ExitAction();
    menu.add(_exitAction);
    return menu;
  }

  //-----------------------------------------------------------
  /**
   * Create the edit menu.
   * 
   */
  protected JMenu createEditMenu() {
    JMenu menu = new JMenu("Edit");

    //Undo and redo are actions of our own creation.
    _undoAction = new UndoAction();
    menu.add(_undoAction);

    _redoAction = new RedoAction();
    menu.add(_redoAction);

    menu.addSeparator();

    //These actions come from the default editor kit.
    //Get the ones we want and stick them in the menu.
    menu.add(getActionByName(DefaultEditorKit.cutAction));
    menu.add(getActionByName(DefaultEditorKit.copyAction));
    menu.add(getActionByName(DefaultEditorKit.pasteAction));

    menu.addSeparator();

    menu.add(getActionByName(DefaultEditorKit.selectAllAction));
    return menu;
  } // createEditMenu
  
  
  
  //-------------------------------------------------
  //The following two methods allow us to find an
  //action provided by the editor kit by its name.
  //-------------------------------------------------  
  private void createActionTable(JTextComponent textComponent) {
    _actions = new Hashtable();
    Action[] actionsArray = textComponent.getActions();
    for (int i = 0; i < actionsArray.length; i++) {
      Action a = actionsArray[i];
      _actions.put(a.getValue(Action.NAME), a);
    }
  } // createActionTable

  private Action getActionByName(String name) {
    return (Action) (_actions.get(name));
  }

  //------------------------------------------------------
  /**
   * set the flag that represents whether the file has been modified.
   * 
   * @param modified  True if file is modified.
   */
  protected void setModified(boolean modified) {
  	
  	_fileModified = modified;
  	String title = this.getTitle();
  	int len = title.length();
  	if (_fileModified) {   
  	  if (! title.endsWith(" *"))
  	    this.setTitle(title+" *");
  	} else {
  	  if (title.endsWith(" *"))
  	    this.setTitle(title.substring(0,len-2));
  	}
  }
  
  //---------------------------------------------------------
  /**
   * check the status of the file modified flag
   */
  protected boolean isModified() {
  	return _fileModified;
  }
  

  //----------------------------------------------
  // class: MyUndoableEditListener
  //----------------------------------------------
  /**
   * This listens for edits that can be undone.
   */
  protected class MyUndoableEditListener implements UndoableEditListener {
    public void undoableEditHappened(UndoableEditEvent e) {
      //Remember the edit and update the menus.
      _undo.addEdit(e.getEdit());
      _undoAction.updateUndoState();
      _redoAction.updateRedoState();
    }
  } // class MyUndoableEditListener


  //----------------------------------------------
  // class: MyDocumentListener
  //----------------------------------------------
  /**
   * this class listens for any changes to the document.
   */
  protected class MyDocumentListener implements DocumentListener {
    public void insertUpdate(DocumentEvent e) {
      //displayEditInfo(e);
      //Debug.out(0, "insert");
      setModified(true);
    }
    public void removeUpdate(DocumentEvent e) {
      //displayEditInfo(e);
      //Debug.out(0, "delete");
      setModified(true);
    }
    public void changedUpdate(DocumentEvent e) {
      //displayEditInfo(e);
      //Debug.out(0, "change");
      setModified(true);
    }

  }  // class: MyDocuemtnListener

  //-------------------------------------------------
  // class UndoAction 
  //-------------------------------------------------
  /**
   * class to support the Undo menubar command.
   * 
   */
  class UndoAction extends AbstractAction {
    public UndoAction() {
      super("Undo");
      setEnabled(false);
    }

    public void actionPerformed(ActionEvent e) {
      try {
        _undo.undo();
      } catch (CannotUndoException ex) {
        Debug.out(0,"JEditor: Unable to undo: ", ex);
        
      }
      updateUndoState();
      _redoAction.updateRedoState();
    }

    protected void updateUndoState() {
      if (_undo.canUndo()) {
        setEnabled(true);
        putValue(Action.NAME, _undo.getUndoPresentationName());
      } else {
        setEnabled(false);
        putValue(Action.NAME, "Undo");
      }
    }
  }
  //-------------------------------------------------
  // class RedoAction 
  //-------------------------------------------------
  /**
   * class to support the Redo menubar command.
   * 
   */
  class RedoAction extends AbstractAction {
    public RedoAction() {
      super("Redo");
      setEnabled(false);
    }

    public void actionPerformed(ActionEvent e) {
      try {
        _undo.redo();
      } catch (CannotRedoException ex) {
        Debug.out(0,"JEditor: Unable to redo: ", ex);
        
      }
      updateRedoState();
      _undoAction.updateUndoState();
    }

    protected void updateRedoState() {
      if (_undo.canRedo()) {
        setEnabled(true);
        putValue(Action.NAME, _undo.getRedoPresentationName());
      } else {
        setEnabled(false);
        putValue(Action.NAME, "Redo");
      }
    }
  }
  //-------------------------------------------------
  // class SaveAction 
  //-------------------------------------------------
  /**
   * class to support the Save menubar command.
   * 
   */

  class SaveAction extends AbstractAction {
    public SaveAction() {
      super("Save");
      setEnabled(true);
    }

    public void actionPerformed(ActionEvent e) {
      try {
        Debug.out( "JEditor: save file");
        fileOutput();
      } catch (Exception ex) {
        Debug.out(ex);

      }

    }

  }
  
    //-------------------------------------------------
  // class ExitAction 
  //-------------------------------------------------
  /**
   * class to support the Exit menubar command.
   * 
   */

  class ExitAction extends AbstractAction {
    public ExitAction() {
      super("Exit");
      setEnabled(true);
    }

    public void actionPerformed(ActionEvent e) {
      try {
        if (JEditor.this.isModified()) {
          Debug.out(0,"JEditor:  exit without saving the modified file.");
        }
        Debug.out( "JEditor: Exit");
        JEditor.this.dispose();
      } catch (Exception ex) {
        Debug.out(ex);

      }

    }

  } // class Exit Action
 
 
  //------------------------------------------------- 
  //The standard main method for JEditor.
  //-------------------------------------------------  
  public static void main(String[] args) {
    String fileName = null;
    Debug.setDebugOn(true);
    try {
      if (args.length > 0) {
        fileName = args[0];
      }
      final JEditor editor = new JEditor();
	  editor.addWindowListener(new WindowAdapter() {
		public void windowClosing(WindowEvent e) {
			Debug.out(0,"windowClosing");
		}
		public void windowClosed(WindowEvent e) {
			Debug.out(0,"windowClosed");
			System.exit(0);
		}
	  });
      editor.setFileName(fileName);

      editor.setup();
      

    } catch (Exception e) {
      Debug.out(e);
    }
  }
 }  // class JEditor 
